//-------------------------------------------------------------------
//-------------------------------------------------------------------
//
// Distance Tree
//   -- Header File
//  
// Primary Author: Adam Bargteil (adamb@cs.berkeley.edu)
//
// The distance tree is simply a balanced octree annotated with
// signed distance values.  This tree has two build functions,
// one takes a triangle mesh and splits the tree near the 
// triangles.  The other takes a function which returns a 
// signed distance at any query point.
//
// The individual octree cell's additionally contain a list
// of indices of triangles which intersect the cell.
//
//-------------------------------------------------------------------
//-------------------------------------------------------------------
//
// Copyright (c) 2003-2005, Regents of the University of California.  All
// rights reserved.
//
// This software is part of the Berkeley Fluid Animation & Simulation
// Toolkit.  The name "Berkeley Fluid Animation & Simulation Toolkit" is
// a trademark of the Regents of the University of California.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are
// met:
//
//   Redistributions of source code must retain the above copyright
//  notice, this list of conditions and the following disclaimer.
//
//  Redistributions in binary form must reproduce the above copyright
//  notice, this list of conditions and the following disclaimer in the
//  documentation and/or other materials provided with the distribution.
//
//  Redistributions in binary form as an executable program, or part of
//  an executable program must, when run for the first time by a given
//  user, prominently display the above copyright notice, this list of
//  conditions and the following disclaimer.
//
//  Neither the name of the University of California, Berkeley nor the
//  names of its contributors may be used to endorse or promote products
//  derived from this software without specific prior written
//  permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
//-------------------------------------------------------------------
//-------------------------------------------------------------------


#ifndef distTree_H
#define distTree_H

#include <fstream>
#include <iostream>
#include "slVector.H"
#include "slcUtil.H"
#include <vector>

class DtTree;
class SplitCriterion;

////////////////////////////////////////////////////////////////

class DtCell {
public:
  int level;          // depth in the octree of this cell
  int vertices[8];    // indices into the octree's pts array of the vertices of this cell
  int parent;         // index of the parent of this cell
  int child[8];       // indices of the children of this cell
  std::vector<int> triangles; // list of triangles intersecting this cell
  DtCell() {};
  ~DtCell() {};
  DtCell &operator=(const DtCell &that);
  DtCell(const DtCell &that);
  bool contains(DtTree *t, const double *x); // does x lie in this cell?
	int clamp(DtTree *t, double *x);           // make x lie in this cell.
	int neighbors[26];

	void read(std::ifstream &in);
	void write(std::ofstream &out) const;
};
std::istream &operator>>(std::istream &strm, DtCell &c);
std::ostream &operator<<(std::ostream &strm, const DtCell &c);

std::istream &operator>>(std::istream &strm, std::vector<DtCell> &p);
std::ostream &operator<<(std::ostream &strm, const std::vector<DtCell> &p);

namespace slc {
	void read(std::ifstream &in, std::vector<DtCell> &l);
	void write(std::ofstream &out, const std::vector<DtCell> &l);
}
  
/////////////////////////////////////////////////////////////////

class DtTree {
public:
  SlVector3 lc, uc; // upper and lower coordinates of the tree 
	               // (the domain is (lc,lc,lc)x(uc,uc,uc)
	SlInt3 ilc, iuc;
	int dim, dim2; // integer dimension and dimension squared
  int max_level; // how deep can the tree go?
  std::vector<DtCell> cells; // list of cells
  std::vector<SlVector3> pts; // octree vertices
  std::vector<SlInt3> ipts; // integer coordinates of octree vertices
  std::vector<double> phi; // phi values at oct-tree points
	std::vector<SlInt6> neighbors;
	
  DtTree() {max_level=9;};
  ~DtTree() {neighbors.clear();};

  // function to build the tree.  
  void buildTree(SlVector3 input_lc, SlVector3 input_uc, int max_level,
		 SplitCriterion &splitCriterion);
  void balanceTree();
	void cacheCellNeighbors();

  friend std::istream &operator>>(std::istream &strm, DtTree &t);
  friend std::ostream &operator<<(std::ostream &strm, const DtTree &t);

  // force x into the domain
  int clamp(double *x);
  int clamp(SlVector3 &x);
  
  // find the cell containing x (maybe cell at level level).
  DtCell* findCell(const SlVector3 &x);
  DtCell* findCell(const SlVector3 &x, int level);
  int findCellIndex(const SlVector3 &x);
  int findCellIndex(const SlVector3 &x, int level);
  
  void read(std::ifstream &in);
  void write(std::ofstream &out) const;
  
private:
  int nvertices;
  inline void addVertex(int p, double x, double y, double z, int ix, int iy, int iz);
  inline void addCell(DtCell *cell, int parent, int level, int v0, int v1, int v2, int v3, int v4, int v5, int v6, int v7);
  inline void divideCell(int c, DtTree *tree);
};
std::istream &operator>>(std::istream &strm, DtTree &t);
std::ostream &operator<<(std::ostream &strm,const DtTree &t);

class SplitCriterion {
public:
	SplitCriterion() {};
	virtual ~SplitCriterion() {};
	virtual bool doSplit(DtTree *t, DtCell *c) = 0;
};

class TriangleSplitCriterion : public SplitCriterion {
public:
	TriangleSplitCriterion() {};
	~TriangleSplitCriterion() {};
	bool doSplit(DtTree *t, DtCell *c);
	std::vector<SlVector3> *meshPts;
	std::vector<SlTri> *triangles;
	std::vector<SlVector3> *faceNormals;
};

class PhiSplitCriterion : public SplitCriterion {
public:
	PhiSplitCriterion() {};
	~PhiSplitCriterion() {};
	PhiSplitCriterion(const PhiSplitCriterion &that):SplitCriterion() {this->phiFunc = that.phiFunc; this->phiFuncData=that.phiFuncData;};
	bool doSplit(DtTree *t, DtCell *c);
	double (*phiFunc)(const SlVector3 &x, void *data);
	void *phiFuncData;
};

class RegularSplitCriterion : public SplitCriterion {
public:
	RegularSplitCriterion() {};
	~RegularSplitCriterion() {};
	bool doSplit(DtTree *, DtCell *) {return true;};
};

class TriangleCentroidSplitCriterion : public SplitCriterion {
public:
	TriangleCentroidSplitCriterion() {};
	~TriangleCentroidSplitCriterion() {};
	bool doSplit(DtTree *t, DtCell *c);
	std::vector<SlVector3> *meshPts;
	std::vector<SlTri> *triangles;
#ifdef ROBUST
	std::vector<SlVector3> *faceNormals;
	std::vector<std::vector<int> > cttriangles;
#endif
};

class ConcentricTripleSplitCriterion : public SplitCriterion {
public:
	ConcentricTripleSplitCriterion() {};
	~ConcentricTripleSplitCriterion() {};
	bool doSplit(DtTree *t, DtCell *c);
	DtTree *tree;
};



void buildTree (DtTree *tree, SlVector3 input_lc, SlVector3 input_uc, int max_level,
								std::vector<SlVector3> *meshPts, std::vector<SlTri> *triangles, 
								std::vector<SlVector3> *faceNormals);
void buildTree (DtTree *tree, SlVector3 input_lc, SlVector3 input_uc, int max_level,
								double (*evalPhi)(const SlVector3 &x, void *data), void *evalPhiData=NULL);

inline void DtTree::addVertex(int , double x, double y, double z, int ix, int iy, int iz) {
	pts.push_back(SlVector3(x,y,z));
	ipts.push_back(SlInt3(ix,iy,iz));
	neighbors.push_back(SlInt6(-1));
};

inline void DtTree::addCell(DtCell *cell, int parent, int level, int v0, int v1, int v2, int v3, int v4, int v5, int v6, int v7) {
	cell->parent = parent;
	cell->level = level;
	cell->vertices[0] = v0;
	cell->vertices[1] = v1;
	cell->vertices[2] = v2;
	cell->vertices[3] = v3;
	cell->vertices[4] = v4;
	cell->vertices[5] = v5;
	cell->vertices[6] = v6;
	cell->vertices[7] = v7;

	cell->child[0] = cell->child[1] = cell->child[2] = cell->child[3] =
		cell->child[4] = cell->child[5] = cell->child[6] = cell->child[7] = -1;

	neighbors[v0][0] = v4;
	neighbors[v0][2] = v2;
	neighbors[v0][4] = v1;
	
	neighbors[v1][0] = v5;
	neighbors[v1][2] = v3;
	neighbors[v1][5] = v0;

	neighbors[v2][0] = v6;
	neighbors[v2][3] = v0;
	neighbors[v2][4] = v3;

	neighbors[v3][0] = v7;
	neighbors[v3][3] = v1;
	neighbors[v3][5] = v2;

	neighbors[v4][1] = v0;
	neighbors[v4][2] = v6;
	neighbors[v4][4] = v5;
	
	neighbors[v5][1] = v1;
	neighbors[v5][2] = v7;
	neighbors[v5][5] = v4;

	neighbors[v6][1] = v2;
	neighbors[v6][3] = v4;
	neighbors[v6][4] = v7;

	neighbors[v7][1] = v3;
	neighbors[v7][3] = v5;
	neighbors[v7][5] = v6;

	cell->neighbors[0] = cell->neighbors[1] = cell->neighbors[2] = cell->neighbors[3] = cell->neighbors[4] = 
		cell->neighbors[5] = cell->neighbors[6] = cell->neighbors[7] = cell->neighbors[8] = cell->neighbors[9] = 
		cell->neighbors[10] = cell->neighbors[11] = cell->neighbors[12] = cell->neighbors[13] = cell->neighbors[14] = 
		cell->neighbors[15] = cell->neighbors[16] = cell->neighbors[17] = cell->neighbors[18] = cell->neighbors[19] = 
		cell->neighbors[20] = cell->neighbors[21] = cell->neighbors[22] = cell->neighbors[23] = cell->neighbors[24] = 
		cell->neighbors[25] = -1;		
};


#endif

