//-------------------------------------------------------------------
//-------------------------------------------------------------------
//
// Distance Tree
//   -- Code File
//  
// Primary Author: Adam Bargteil (adamb@cs.berkeley.edu)
// 
//-------------------------------------------------------------------
//-------------------------------------------------------------------
//
// Copyright (c) 2003-2005, Regents of the University of California.  All
// rights reserved.
//
// This software is part of the Berkeley Fluid Animation & Simulation
// Toolkit.  The name "Berkeley Fluid Animation & Simulation Toolkit" is
// a trademark of the Regents of the University of California.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are
// met:
//
//   Redistributions of source code must retain the above copyright
//  notice, this list of conditions and the following disclaimer.
//
//  Redistributions in binary form must reproduce the above copyright
//  notice, this list of conditions and the following disclaimer in the
//  documentation and/or other materials provided with the distribution.
//
//  Redistributions in binary form as an executable program, or part of
//  an executable program must, when run for the first time by a given
//  user, prominently display the above copyright notice, this list of
//  conditions and the following disclaimer.
//
//  Neither the name of the University of California, Berkeley nor the
//  names of its contributors may be used to endorse or promote products
//  derived from this software without specific prior written
//  permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
//-------------------------------------------------------------------
//-------------------------------------------------------------------

#include "distTree.H"
#include "float.h"
#include <cstdlib>
#include "slIO.H"

#ifdef PCUBE
extern "C" {
 int
fast_polygon_intersects_cube(int nverts, const double verts[][3],
			const double polynormal[3],
			int already_know_verts_are_outside_cube,
			int already_know_edges_are_outside_cube);
}
#endif

#define EPS 1e-15
using namespace std;

const double width = 4;

//////////////////////////////////////////////////////////////////
// DtCell
//////////////////////////////////////////////////////////////////
bool DtCell::contains(DtTree *t, const double *x) {
	SlVector3 &v1 = t->pts[vertices[0]];
	SlVector3 &v2 = t->pts[vertices[7]];
  if (x[0] < v1[0]-EPS || x[0] > v2[0]+EPS) return false;
  if (x[1] < v1[1]-EPS || x[1] > v2[1]+EPS) return false;
  if (x[2] < v1[2]-EPS || x[2] > v2[2]+EPS) return false;
  return true;
}

int DtCell::clamp(DtTree *t, double *x) {
	int ret = 0;
	SlVector3 &v1 = t->pts[vertices[0]];
	SlVector3 &v2 = t->pts[vertices[7]];
	
  if (x[0] < v1[0]) { x[0] = v1[0]; ret += 1;}
  if (x[0] > v2[0]) { x[0] = v2[0]; ret += 2;}
  if (x[1] < v1[1]) { x[1] = v1[1]; ret += 4;}
  if (x[1] > v2[1]) { x[1] = v2[1]; ret += 8;}
  if (x[2] < v1[2]) { x[2] = v1[2]; ret += 16;}
  if (x[2] > v2[2]) { x[2] = v2[2]; ret += 32;}
  return ret;
}

std::istream &operator>>(std::istream &strm, DtCell &c) {
  ios::fmtflags orgFlags = strm.setf(ios::skipws);
  eatStr("[",strm);
  strm >> c.level;
  eatChar(',',strm);
  strm >> c.vertices[0];
  eatChar(',',strm);
  strm >> c.vertices[1];
  eatChar(',',strm);
  strm >> c.vertices[2];
  eatChar(',',strm);
  strm >> c.vertices[3];
  eatChar(',',strm);
  strm >> c.vertices[4];
  eatChar(',',strm);
  strm >> c.vertices[5];
  eatChar(',',strm);
  strm >> c.vertices[6];
  eatChar(',',strm);
  strm >> c.vertices[7];
  eatChar(',',strm);
  strm >> c.parent;
  eatChar(',',strm);
  strm >> c.child[0];
  eatChar(',',strm);
  strm >> c.child[1];
  eatChar(',',strm);
  strm >> c.child[2];
  eatChar(',',strm);
  strm >> c.child[3];
  eatChar(',',strm);
  strm >> c.child[4];
  eatChar(',',strm);
  strm >> c.child[5];
  eatChar(',',strm);
  strm >> c.child[6];
  eatChar(',',strm);
  strm >> c.child[7];
  eatChar(',',strm);
  strm >> c.triangles;
  eatChar(']',strm);
  strm.flags(orgFlags);
  return strm;
}

std::ostream &operator<<(std::ostream &strm, const DtCell &c) {
  strm << "[";
  strm << c.level;
  strm << ",";
  strm << c.vertices[0];
  strm << ",";
  strm << c.vertices[1];
  strm << ",";
  strm << c.vertices[2];
  strm << ",";
  strm << c.vertices[3];
  strm << ",";
  strm << c.vertices[4];
  strm << ",";
  strm << c.vertices[5];
  strm << ",";
  strm << c.vertices[6];
  strm << ",";
  strm << c.vertices[7];
  strm << ",";
  strm << c.parent;
  strm << ",";
  strm << c.child[0];
  strm << ",";
  strm << c.child[1];
  strm << ",";
  strm << c.child[2];
  strm << ",";
  strm << c.child[3];
  strm << ",";
  strm << c.child[4];
  strm << ",";
  strm << c.child[5];
  strm << ",";
  strm << c.child[6];
  strm << ",";
  strm << c.child[7];
  strm << ",";
  strm << c.triangles;
  strm << "]";
  return strm;
}  

void DtCell::read(ifstream &in) {
	in.read((char*) &level, sizeof(int));
  in.read((char*) &vertices[0], sizeof(int));
  in.read((char*) &vertices[1], sizeof(int));
  in.read((char*) &vertices[2], sizeof(int));
  in.read((char*) &vertices[3], sizeof(int));
  in.read((char*) &vertices[4], sizeof(int));
  in.read((char*) &vertices[5], sizeof(int));
  in.read((char*) &vertices[6], sizeof(int));
  in.read((char*) &vertices[7], sizeof(int));
  in.read((char*) &parent, sizeof(int));
  in.read((char*) &child[0], sizeof(int));
  in.read((char*) &child[1], sizeof(int));
  in.read((char*) &child[2], sizeof(int));
  in.read((char*) &child[3], sizeof(int));
  in.read((char*) &child[4], sizeof(int));
  in.read((char*) &child[5], sizeof(int));
  in.read((char*) &child[6], sizeof(int));
  in.read((char*) &child[7], sizeof(int));
  slc::read(in, triangles);
	for (unsigned int i=0; i<26; i++) {
		in.read((char*) &neighbors[i], sizeof(int));
	}
}

void DtCell::write(std::ofstream &out) const {
	out.write((char*) &level, sizeof(int));
  out.write((char*) &vertices[0], sizeof(int));
  out.write((char*) &vertices[1], sizeof(int));
  out.write((char*) &vertices[2], sizeof(int));
  out.write((char*) &vertices[3], sizeof(int));
  out.write((char*) &vertices[4], sizeof(int));
  out.write((char*) &vertices[5], sizeof(int));
  out.write((char*) &vertices[6], sizeof(int));
  out.write((char*) &vertices[7], sizeof(int));
  out.write((char*) &parent, sizeof(int));
  out.write((char*) &child[0], sizeof(int));
  out.write((char*) &child[1], sizeof(int));
  out.write((char*) &child[2], sizeof(int));
  out.write((char*) &child[3], sizeof(int));
  out.write((char*) &child[4], sizeof(int));
  out.write((char*) &child[5], sizeof(int));
  out.write((char*) &child[6], sizeof(int));
  out.write((char*) &child[7], sizeof(int));
  slc::write(out, triangles);
	for (unsigned int i=0; i<26; i++) {
		out.write((char*) &neighbors[i], sizeof(int));
	}
}

DtCell &DtCell::operator=(const DtCell &that) {
  level = that.level;
  vertices[0] = that.vertices[0];
  vertices[1] = that.vertices[1];
  vertices[2] = that.vertices[2];
  vertices[3] = that.vertices[3];
  vertices[4] = that.vertices[4];
  vertices[5] = that.vertices[5];
  vertices[6] = that.vertices[6];
  vertices[7] = that.vertices[7];
  parent = that.parent;
  child[0] = that.child[0];
  child[1] = that.child[1];
  child[2] = that.child[2];
  child[3] = that.child[3];
  child[4] = that.child[4];
  child[5] = that.child[5];
  child[6] = that.child[6];
  child[7] = that.child[7];
	triangles = that.triangles;
	neighbors[0] = that.neighbors[0];
	neighbors[1] = that.neighbors[1];
	neighbors[2] = that.neighbors[2];
	neighbors[3] = that.neighbors[3];
	neighbors[4] = that.neighbors[4];
	neighbors[5] = that.neighbors[5];
	neighbors[6] = that.neighbors[6];
	neighbors[7] = that.neighbors[7];
	neighbors[8] = that.neighbors[8];
	neighbors[9] = that.neighbors[9];
	neighbors[10] = that.neighbors[10];
	neighbors[11] = that.neighbors[11];
	neighbors[12] = that.neighbors[12];
	neighbors[13] = that.neighbors[13];
	neighbors[14] = that.neighbors[14];
	neighbors[15] = that.neighbors[15];
	neighbors[16] = that.neighbors[16];
	neighbors[17] = that.neighbors[17];
	neighbors[18] = that.neighbors[18];
	neighbors[19] = that.neighbors[19];
	neighbors[20] = that.neighbors[20];
	neighbors[21] = that.neighbors[21];
	neighbors[22] = that.neighbors[22];
	neighbors[23] = that.neighbors[23];
	neighbors[24] = that.neighbors[24];
	neighbors[25] = that.neighbors[25];

	return (*this);
}

DtCell::DtCell(const DtCell &that) {
  level = that.level;
  vertices[0] = that.vertices[0];
  vertices[1] = that.vertices[1];
  vertices[2] = that.vertices[2];
  vertices[3] = that.vertices[3];
  vertices[4] = that.vertices[4];
  vertices[5] = that.vertices[5];
  vertices[6] = that.vertices[6];
  vertices[7] = that.vertices[7];
  parent = that.parent;
  child[0] = that.child[0];
  child[1] = that.child[1];
  child[2] = that.child[2];
  child[3] = that.child[3];
  child[4] = that.child[4];
  child[5] = that.child[5];
  child[6] = that.child[6];
  child[7] = that.child[7];
	triangles = that.triangles;
	neighbors[0] = that.neighbors[0];
	neighbors[1] = that.neighbors[1];
	neighbors[2] = that.neighbors[2];
	neighbors[3] = that.neighbors[3];
	neighbors[4] = that.neighbors[4];
	neighbors[5] = that.neighbors[5];
	neighbors[6] = that.neighbors[6];
	neighbors[7] = that.neighbors[7];
	neighbors[8] = that.neighbors[8];
	neighbors[9] = that.neighbors[9];
	neighbors[10] = that.neighbors[10];
	neighbors[11] = that.neighbors[11];
	neighbors[12] = that.neighbors[12];
	neighbors[13] = that.neighbors[13];
	neighbors[14] = that.neighbors[14];
	neighbors[15] = that.neighbors[15];
	neighbors[16] = that.neighbors[16];
	neighbors[17] = that.neighbors[17];
	neighbors[18] = that.neighbors[18];
	neighbors[19] = that.neighbors[19];
	neighbors[20] = that.neighbors[20];
	neighbors[21] = that.neighbors[21];
	neighbors[22] = that.neighbors[22];
	neighbors[23] = that.neighbors[23];
	neighbors[24] = that.neighbors[24];
	neighbors[25] = that.neighbors[25];
}
  
  
istream &operator>>(std::istream &strm, std::vector<DtCell> &l) {
  unsigned int i,s = 0;
  ios::fmtflags orgFlags = strm.setf(ios::skipws);
	std::vector<DtCell>::iterator c;

  eatStr("[",strm);
  strm >> s;
  eatStr(":",strm);
  
  if (strm.good()) {
    l.resize(s);
    for(c=l.begin(),i=0;c!=l.end();c++,i++) {
      strm >> (*c);
      if (i != (s-1)) eatStr(",",strm);
    }
  }
  eatChar(']',strm);
  strm.flags(orgFlags);
  return strm;
}

ostream &operator<<(ostream &strm,const std::vector<DtCell> &l) {
	std::vector<DtCell>::const_iterator c;
  strm << "[";
  strm << l.size();
  strm << ":";
  unsigned int i;
  for(c=l.begin(),i=0;c!=l.end();c++,i++) {
    strm << (*c);
    if (i != (l.size()-1)) strm << ",";
  }
  strm << "]";
  return strm;
}

void slc::read(ifstream &in, std::vector<DtCell> &l) {
	int s;
	std::vector<DtCell>::iterator c;
	in.read((char *)&s, sizeof(int));
	l.resize(s);
	for (c=l.begin(); c!=l.end(); c++) (*c).read(in);
}

void slc::write(std::ofstream &out, const std::vector<DtCell> &l) {
	int s = l.size();
	std::vector<DtCell>::const_iterator c;
	out.write((char *)&s, sizeof(int));
	for (c=l.begin(); c<l.end(); c++) (*c).write(out);
}

//////////////////////////////////////////////////////////////////
// DtTree
//////////////////////////////////////////////////////////////////

int DtTree::clamp(SlVector3 &x) {
	return clamp(&(x[0]));
}

int DtTree::clamp(double *x) {
	int ret = 0;
  if (x[0] < lc[0]) { x[0] = lc[0]; ret += 1;}
  if (x[0] > uc[0]) { x[0] = uc[0]; ret += 2;}
  if (x[1] < lc[1]) { x[1] = lc[1]; ret += 4;}
  if (x[1] > uc[1]) { x[1] = uc[1]; ret += 8;}
  if (x[2] < lc[2]) { x[2] = lc[2]; ret += 16;}
  if (x[2] > uc[2]) { x[2] = uc[2]; ret += 32;}
  return ret;
}

DtCell* DtTree::findCell(const SlVector3 &x) {
  int i;
  SlVector3 mid;
	DtCell *c=&(cells[0]);

  while (c->child[0] != -1) {
		mid = (pts[c->vertices[0]] + pts[c->vertices[7]]) / 2.0;
    i = 0;
    if (x[0] >= mid[0]) i|=4;
    if (x[1] >= mid[1]) i|=2;
    if (x[2] >= mid[2]) i|=1;
    c = &(cells[c->child[i]]);
  }
	return c;
}

int DtTree::findCellIndex(const SlVector3 &x) {
  int i,index=0;
  SlVector3 mid;
	DtCell *c=&(cells[0]);

  while (c->child[0] != -1) {
		mid = (pts[c->vertices[0]] + pts[c->vertices[7]]) / 2.0;
    i = 0;
    if (x[0] >= mid[0]) i+=4;
    if (x[1] >= mid[1]) i+=2;
    if (x[2] >= mid[2]) i+=1;
    c = &(cells[index=(c->child[i])]);
  }
	return index;
}

DtCell* DtTree::findCell(const SlVector3 &x, int level) {
  int i;
  SlVector3 mid;
	DtCell *c=&(cells[0]);

  while (c->child[0] != -1 && c->level < level) {
    mid = (pts[c->vertices[0]] + pts[c->vertices[7]]) / 2.0;
    i = 0;
    if (x[0] >= mid[0]) i+=4;
    if (x[1] >= mid[1]) i+=2;
    if (x[2] >= mid[2]) i+=1;
    c = &(cells[c->child[i]]);
  }
	return c;
}

int DtTree::findCellIndex(const SlVector3 &x, int level) {
  int i,index=0;
  SlVector3 mid;
	DtCell *c=&(cells[0]);

  while (c->child[0] != -1 && c->level < level) {
    mid = (pts[c->vertices[0]] + pts[c->vertices[7]]) / 2.0;
    i = 0;
    if (x[0] >= mid[0]) i+=4;
    if (x[1] >= mid[1]) i+=2;
    if (x[2] >= mid[2]) i+=1;
    c = &(cells[index=(c->child[i])]);
  }
	return index;
}

istream &operator>>(std::istream &strm, DtTree &t) {
  ios::fmtflags orgFlags = strm.setf(ios::skipws);
  eatStr("[",strm);
  strm >> t.max_level;
  eatChar(',',strm);
  strm >> t.lc;
  eatChar(',',strm);
  strm >> t.uc;
  eatChar(',',strm);
  strm >> t.cells;
  eatChar(',',strm);
  strm >> t.nvertices;
  eatChar(',',strm);
  strm >> t.pts;
  eatChar(',',strm);
  strm >> t.ipts;
  eatChar(',',strm);
  strm >> t.phi;
  eatChar(']',strm);
  strm.flags(orgFlags);
  return strm;
}

ostream &operator<<(ostream &strm,const DtTree &t) {
  strm << "[";
  strm << t.max_level;
  strm << ",";
  strm << t.lc;
  strm << ",";
  strm << t.uc;
  strm << ",";
  strm << t.cells;
  strm << ",";
  strm << t.nvertices;
  strm << ",";
  strm << t.pts;
  strm << ",";
  strm << t.ipts;
  strm << ",";
  strm << t.phi;
  strm << "]";
  return strm;
}

void DtTree::read(std::ifstream &in) {
	int version;
	in.read((char*)&version, sizeof(int));
	if (in.fail()) {
		cerr<<"can't read from file"<<endl;
		exit(-1);
	}
	if (version == 1) {
		in.read((char*)&max_level, sizeof(int));
		in.read((char*)&(lc[0]), sizeof(double));
		in.read((char*)&(lc[1]), sizeof(double));
		in.read((char*)&(lc[2]), sizeof(double));
		in.read((char*)&(uc[0]), sizeof(double));
		in.read((char*)&(uc[1]), sizeof(double));
		in.read((char*)&(uc[2]), sizeof(double));
		slc::read(in, cells);
		in.read((char*)&nvertices, sizeof(int));
		slc::read(in,pts);
		slc::read(in, ipts);
		slc::read(in, phi);
	} else if (version == 3) {
		in.read((char*)&max_level, sizeof(int));
		in.read((char*)&(lc[0]), sizeof(double));
		in.read((char*)&(lc[1]), sizeof(double));
		in.read((char*)&(lc[2]), sizeof(double));
		in.read((char*)&(uc[0]), sizeof(double));
		in.read((char*)&(uc[1]), sizeof(double));
		in.read((char*)&(uc[2]), sizeof(double));
		slc::read(in, cells);
		in.read((char*)&nvertices, sizeof(int));
		slc::read(in,pts);
		slc::read(in, ipts);
		slc::read(in, phi);
		slc::read(in, neighbors);
	} else {		
		cerr<<"incompatible bdt file version="<<version<<endl;
	}
}

void DtTree::write(std::ofstream &out) const {
	int version = 3;
	int i;
	SlVector3 v;
	i = version;
	out.write((char*)&i, sizeof(int));
	i = max_level;
  out.write((char*)&i, sizeof(int));
	v = lc;
  out.write((char*)&(v[0]), sizeof(double));
  out.write((char*)&(v[1]), sizeof(double));
  out.write((char*)&(v[2]), sizeof(double));
	v = uc;
  out.write((char*)&(v[0]), sizeof(double));
  out.write((char*)&(v[1]), sizeof(double));
  out.write((char*)&(v[2]), sizeof(double));
	slc::write(out, cells);
	i = nvertices;
  out.write((char*)&i, sizeof(int));
	slc::write(out, pts);
	slc::write(out, ipts);
	slc::write(out, phi);
	slc::write(out, neighbors);
}

bool TriangleSplitCriterion::doSplit(DtTree *t, DtCell *c) {
#ifndef PCUBE
  cerr<<"Distance trees cannot be built from a triangle mesh without the pcube library"<<endl;
  exit(-1);
#else

	std::vector<int>::const_iterator j;
	std::vector<int>::iterator k;
	unsigned int i;

  if (c->level == 0) {
    c->triangles.resize(triangles->size());
    for (k=c->triangles.begin(),i=0; k!=c->triangles.end(); i++,k++) {
      (*k) = i;
    }
    return (c->triangles.size() != 0);
  }
  
  DtCell &parent = t->cells[c->parent];
  for (j=parent.triangles.begin(); j!=parent.triangles.end(); j++) {
    SlVector3 center=(t->pts[c->vertices[0]]+t->pts[c->vertices[7]])/2.0;
    double lh = (t->pts[c->vertices[7]][0]-t->pts[c->vertices[0]][0])*3.00;
    double tri[3][3];
    double n[3];
    tri[0][0]=((*meshPts)[(*triangles)[*j][0]][0]-center[0])/(lh);
    tri[0][1]=((*meshPts)[(*triangles)[*j][0]][1]-center[1])/(lh);
    tri[0][2]=((*meshPts)[(*triangles)[*j][0]][2]-center[2])/(lh);
    tri[1][0]=((*meshPts)[(*triangles)[*j][1]][0]-center[0])/(lh);
    tri[1][1]=((*meshPts)[(*triangles)[*j][1]][1]-center[1])/(lh);
    tri[1][2]=((*meshPts)[(*triangles)[*j][1]][2]-center[2])/(lh);
    tri[2][0]=((*meshPts)[(*triangles)[*j][2]][0]-center[0])/(lh);
    tri[2][1]=((*meshPts)[(*triangles)[*j][2]][1]-center[1])/(lh);
    tri[2][2]=((*meshPts)[(*triangles)[*j][2]][2]-center[2])/(lh);
    n[0] = (*faceNormals)[*j][0];
    n[1] = (*faceNormals)[*j][1];
    n[2] = (*faceNormals)[*j][2];

    double m = sqrt(n[0]*n[0]+n[1]*n[1]+n[2]*n[2]);
    n[0] /= m;
    n[1] /= m;
    n[2] /= m;
    
    if (fast_polygon_intersects_cube(3,tri,n,0,0))
      c->triangles.push_back(*j);
  }
  return (c->triangles.size() != 0);
#endif
}

bool TriangleCentroidSplitCriterion::doSplit(DtTree *t, DtCell *c) {
	int i;
	std::vector<int> lcttriangles;
	std::vector<int>::iterator j;
	std::vector<int>::const_iterator k;
	if (c->level == 0) {
		c->triangles.resize(triangles->size());
		lcttriangles.resize(triangles->size());
		for (j=c->triangles.begin(),i=0; j!=c->triangles.end(); j++,i++) {
			(*j) = i;
		}
		return (c->triangles.size() != 0);
	}
	
	DtCell &parent = t->cells[c->parent];
	for (k=parent.triangles.begin(); k!=parent.triangles.end(); k++) {
		SlVector3 centroid=((*meshPts)[(*triangles)[(*k)][0]]+
													 (*meshPts)[(*triangles)[(*k)][1]]+
													 (*meshPts)[(*triangles)[(*k)][2]])/3.0;
		if (c->contains(t, &(centroid[0]))) c->triangles.push_back(*k);
	}
	return (c->triangles.size() != 0);
}


bool PhiSplitCriterion::doSplit(DtTree *t, DtCell *c) {
	SlVector3 x = 0.5*(t->pts[c->vertices[0]]+t->pts[c->vertices[7]]);
	double h = pow2(-c->level-1)*(t->uc[0]-t->lc[0]);
	return (width*h > fabs(phiFunc(x,phiFuncData)));
}

bool ConcentricTripleSplitCriterion::doSplit(DtTree *t, DtCell *c) {
	SlVector3 z;
	double h;
	SlVector3 y = 0.5*(t->pts[c->vertices[0]]+t->pts[c->vertices[7]]);
	DtCell *cell;

	h = (t->pts[c->vertices[7]][0]-t->pts[c->vertices[0]][0]);

	cell = tree->findCell(y, c->level);
	if (cell->triangles.size() != 0) {
		c->triangles = cell->triangles;
		return true;
	}
	
	z[0]=y[0]-h; z[1]=y[1]; z[2]=y[2];
	cell = tree->findCell(z, c->level);
	if (cell->triangles.size() != 0) return true;

	z[0]=y[0]; z[1]=y[1]-h; z[2]=y[2];
	cell = tree->findCell(z, c->level);
	if (cell->triangles.size() != 0) return true;

	z[0]=y[0]; z[1]=y[1]; z[2]=y[2]-h;
	cell = tree->findCell(z, c->level);
	if (cell->triangles.size() != 0) return true;

	z[0]=y[0]; z[1]=y[1]; z[2]=y[2]+h;
	cell = tree->findCell(z, c->level);
	if (cell->triangles.size() != 0) return true;

	z[0]=y[0]; z[1]=y[1]+h; z[2]=y[2];
	cell = tree->findCell(z, c->level);
	if (cell->triangles.size() != 0) return true;

	z[0]=y[0]+h; z[1]=y[1]; z[2]=y[2];
	cell = tree->findCell(z, c->level);
	if (cell->triangles.size() != 0) return true;
	
	z[0]=y[0]-h; z[1]=y[1]-h; z[2]=y[2];
	cell = tree->findCell(z, c->level);
	if (cell->triangles.size() != 0) return true;

	z[0]=y[0]-h; z[1]=y[1]; z[2]=y[2]-h;
	cell = tree->findCell(z, c->level);
	if (cell->triangles.size() != 0) return true;

	z[0]=y[0]-h; z[1]=y[1]; z[2]=y[2]+h;
	cell = tree->findCell(z, c->level);
	if (cell->triangles.size() != 0) return true;

	z[0]=y[0]-h; z[1]=y[1]+h; z[2]=y[2];
	cell = tree->findCell(z, c->level);
	if (cell->triangles.size() != 0) return true;

	z[0]=y[0]; z[1]=y[1]-h; z[2]=y[2]-h;
	cell = tree->findCell(z, c->level);
	if (cell->triangles.size() != 0) return true;

	z[0]=y[0]; z[1]=y[1]-h; z[2]=y[2]+h;
	cell = tree->findCell(z, c->level);
	if (cell->triangles.size() != 0) return true;

	z[0]=y[0]; z[1]=y[1]+h; z[2]=y[2]-h;
	cell = tree->findCell(z, c->level);
	if (cell->triangles.size() != 0) return true;

	z[0]=y[0]; z[1]=y[1]+h; z[2]=y[2]+h;
	cell = tree->findCell(z, c->level);
	if (cell->triangles.size() != 0) return true;

	z[0]=y[0]+h; z[1]=y[1]-h; z[2]=y[2];
	cell = tree->findCell(z, c->level);
	if (cell->triangles.size() != 0) return true;

	z[0]=y[0]+h; z[1]=y[1]; z[2]=y[2]-h;
	cell = tree->findCell(z, c->level);
	if (cell->triangles.size() != 0) return true;

	z[0]=y[0]+h; z[1]=y[1]; z[2]=y[2]+h;
	cell = tree->findCell(z, c->level);
	if (cell->triangles.size() != 0) return true;

	z[0]=y[0]+h; z[1]=y[1]+h; z[2]=y[2];
	cell = tree->findCell(z, c->level);
	if (cell->triangles.size() != 0) return true;

	z[0]=y[0]-h; z[1]=y[1]-h; z[2]=y[2]-h;
	cell = tree->findCell(z, c->level);
	if (cell->triangles.size() != 0) return true;

	z[0]=y[0]-h; z[1]=y[1]-h; z[2]=y[2]+h;
	cell = tree->findCell(z, c->level);
	if (cell->triangles.size() != 0) return true;

	z[0]=y[0]-h; z[1]=y[1]+h; z[2]=y[2]-h;
	cell = tree->findCell(z, c->level);
	if (cell->triangles.size() != 0) return true;

	z[0]=y[0]-h; z[1]=y[1]+h; z[2]=y[2]+h;
	cell = tree->findCell(z, c->level);
	if (cell->triangles.size() != 0) return true;

	z[0]=y[0]+h; z[1]=y[1]-h; z[2]=y[2]-h;
	cell = tree->findCell(z, c->level);
	if (cell->triangles.size() != 0) return true;

	z[0]=y[0]+h; z[1]=y[1]-h; z[2]=y[2]+h;
	cell = tree->findCell(z, c->level);
	if (cell->triangles.size() != 0) return true;
	
	z[0]=y[0]+h; z[1]=y[1]+h; z[2]=y[2]-h;
	cell = tree->findCell(z, c->level);
	if (cell->triangles.size() != 0) return true;

	z[0]=y[0]+h; z[1]=y[1]+h; z[2]=y[2]+h;
	cell = tree->findCell(z, c->level);
	if (cell->triangles.size() != 0) return true;
	
	return false;
}

static int found[19];

inline void DtTree::divideCell(int c, DtTree *tree) {
	int p,vert[19]={-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1};
	SlInt3 i;

	// the following line fixes a problem with the apple compiler.
	cells.size();

	DtCell *cell = &(cells[c]);
	int l = cell->level+1;
	SlVector3 lc = pts[cell->vertices[0]];
	SlVector3 uc = pts[cell->vertices[7]];
	double h = (uc[0]-lc[0])/2.0;
	SlInt3 ilc = ipts[cell->vertices[0]];
	SlInt3 iuc = ipts[cell->vertices[7]];
	int ih = (iuc[0]-ilc[0])/2;

	p = pts.size();
			
	double halfh = h/2.0;
	SlVector3 x;
	DtCell *d;

	i.set(ilc[0], ilc[1], ilc[2]+ih);
	if (ilc[0] > 0 && ilc[1] > 0) {
		x.set(lc[0]-halfh, lc[1]-halfh, lc[2]+halfh);
		d = findCell(x);
		if (d->level >= l) vert[0] = d->vertices[7];
	}
	if (vert[0] == -1 && ilc[1] > 0) {
		x.set(lc[0]+halfh, lc[1]-halfh, lc[2]+halfh);
		d = findCell(x);
		if (d->level >= l) vert[0] = d->vertices[3];
	}
	if (vert[0] == -1 && ilc[0] > 0) {
		x.set(lc[0]-halfh, lc[1]+halfh, lc[2]+halfh);
		d = findCell(x);
		if (d->level >= l) vert[0] = d->vertices[5];
	}
	if (vert[0] == -1) {
		vert[0] = p;
		addVertex(p,lc[0],lc[1],lc[2]+h,i[0],i[1],i[2]);
		p++;
		found[0]++;
	} 

	i.set(ilc[0],ilc[1]+ih,ilc[2]);
	if (ilc[0] > 0 && ilc[2] > 0) {
		x.set(lc[0]-halfh, lc[1]+halfh, lc[2]-halfh);
		d = findCell(x);
		if (d->level >= l) vert[1] = d->vertices[7];
	}
	if (vert[1] == -1 && ilc[2] > 0) {
		x.set(lc[0]+halfh, lc[1]+halfh, lc[2]-halfh);
		d = findCell(x);
		if (d->level >= l) vert[1] = d->vertices[3];
	}
	if (vert[1] == -1 && ilc[0] > 0) {
		x.set(lc[0]-halfh, lc[1]+halfh, lc[2]+halfh);
		d = findCell(x);
		if (d->level >= l) vert[1] = d->vertices[6];
	}
	if (vert[1] == -1) {
		vert[1] = p;
		addVertex(p,lc[0],lc[1]+h,lc[2],i[0],i[1],i[2]);
		p++;
		found[1]++;
	} 
	
	i.set(ilc[0],ilc[1]+ih,ilc[2]+ih);
	if (ilc[0] > 0) {
		x.set(lc[0]-halfh, lc[1]+halfh, lc[2]+halfh);
		d = findCell(x);
		if (d->level >= l) vert[2] = d->vertices[7];
	}
	if (vert[2] == -1) {
		vert[2] = p;
		addVertex(p,lc[0],lc[1]+h,lc[2]+h,i[0],i[1],i[2]);
		p++;
		found[2]++;
	} 


	i.set(ilc[0]+ih,ilc[1],ilc[2]);
	if (ilc[1] > 0 && ilc[2] > 0) {
		x.set(lc[0]+halfh, lc[1]-halfh, lc[2]-halfh);
		d = findCell(x);
		if (d->level >= l) vert[3] = d->vertices[7];
	}
	if (vert[3] == -1 && ilc[2] > 0) {
		x.set(lc[0]+halfh, lc[1]+halfh, lc[2]-halfh);
		d = findCell(x);
		if (d->level >= l) vert[3] = d->vertices[5];
	}
	if (vert[3] == -1 && ilc[1] > 0) {
		x.set(lc[0]+halfh, lc[1]-halfh, lc[2]+halfh);
		d = findCell(x);
		if (d->level >= l) vert[3] = d->vertices[6];
	}
	if (vert[3] == -1) {
		vert[3] = p;
		addVertex(p,lc[0]+h,lc[1],lc[2],i[0],i[1],i[2]);
		p++;
		found[3]++;
	} 
	
	i.set(ilc[0]+ih,ilc[1],ilc[2]+ih);
	if (ilc[1] > 0) {
		x.set(lc[0]+halfh, lc[1]-halfh, lc[2]+halfh);
		d = findCell(x);
		if (d->level >= l) vert[4] = d->vertices[7];
	}
	if (vert[4] == -1) {
		vert[4] = p;
		addVertex(p,lc[0]+h,lc[1],lc[2]+h,i[0],i[1],i[2]);
		p++;
		found[4]++;
	} 
	
	i.set(ilc[0]+ih,ilc[1]+ih,ilc[2]);
	if (ilc[2] > 0) {
		x.set(lc[0]+halfh, lc[1]+halfh, lc[2]-halfh);
		d = findCell(x);
		if (d->level >= l) vert[5] = d->vertices[7];
	}

	if (vert[5] == -1) {
		vert[5] = p;
		addVertex(p,lc[0]+h,lc[1]+h,lc[2],i[0],i[1],i[2]);
		p++;
		found[5]++;
	} 
	
	i.set(ilc[0]+ih,ilc[1]+ih,ilc[2]+ih);
	vert[6] = p;
	addVertex(p,lc[0]+h,lc[1]+h,lc[2]+h,i[0],i[1],i[2]);
	p++;
	found[6]++;
	
	i.set(ilc[0],ilc[1]+ih,iuc[2]);
	if (ilc[0] > 0) {
		x.set(lc[0]-halfh, lc[1]+halfh, uc[2]-halfh);
		d = findCell(x);
		if (d->level >= l) vert[7] = d->vertices[7];
	}
	if (vert[7] == - 1 && iuc[2] < this->iuc[2]) {
		x.set(lc[0]+halfh, lc[1]+halfh, uc[2]+halfh);
		d = findCell(x);
		if (d->level >= l) vert[7] = d->vertices[2];
	}
	if (vert[7] == -1 && ilc[0] > 0 && iuc[2] < this->iuc[2]) {
		x.set(lc[0]-halfh, lc[1]+halfh, uc[2]+halfh);
		d = findCell(x);
		if (d->level >= l) vert[7] = d->vertices[6];
	}
	if (vert[7] == -1) {
		vert[7] = p;
		addVertex(p,lc[0],lc[1]+h,uc[2],i[0],i[1],i[2]);
		p++;
		found[7]++;
	} 
	
	i.set(ilc[0]+ih,ilc[1],iuc[2]);
	if (ilc[1] > 0) {
		x.set(lc[0]+halfh, lc[1]-halfh, uc[2]-halfh);
		d = findCell(x);
		if (d->level >= l) vert[8] = d->vertices[7];
	}
	if (vert[8] == -1 && ilc[1] > 0 && uc[2] < this->iuc[2]) {
		x.set(lc[0]+halfh, lc[1]-halfh, uc[2]+halfh);
		d = findCell(x);
		if (d->level >= l) vert[8] = d->vertices[6];
	}
	if (vert[8] == -1) {
		vert[8] = p;
		addVertex(p,lc[0]+h,lc[1],uc[2],i[0],i[1],i[2]);
		p++;
		found[8]++;
	} 
	
	i.set(ilc[0]+ih,ilc[1]+ih,iuc[2]);
	vert[9] = p;
	addVertex(p,lc[0]+h,lc[1]+h,uc[2],i[0],i[1],i[2]);
	p++;
	found[9]++;
	
	i.set(ilc[0],iuc[1],ilc[2]+ih);
	if (ilc[0]) {
		x.set(lc[0]-halfh, uc[1]-halfh, lc[2]+halfh);
		d = findCell(x);
		if (d->level >= l) vert[10] = d->vertices[7];
	}
	if (vert[10] == -1 && ilc[0] > 0 && iuc[1] < this->iuc[1]) {
		x.set(lc[0]-halfh, uc[1]+halfh, lc[2]+halfh);
		d = findCell(x);
		if (d->level >= l) vert[10] = d->vertices[5];
	}

	if (vert[10] == -1) {
		vert[10] = p;
		addVertex(p,lc[0],uc[1],lc[2]+h,i[0],i[1],i[2]);
		p++;
		found[10]++;
	} 
	
	i.set(ilc[0]+ih,iuc[1],ilc[2]);
	if (ilc[2] > 0) {
		x.set(lc[0]+halfh, uc[1]-halfh, lc[2]-halfh);
		d = findCell(x);
		if (d->level >= l) vert[11] = d->vertices[7];
	}
	if (vert[11] == -1 && iuc[1] < this->iuc[1] && ilc[2] > 0) {
		x.set(lc[0]+halfh, uc[1]+halfh, lc[2]-halfh);
		d = findCell(x);
		if (d->level >= l) vert[11] = d->vertices[5];
	}
	if (vert[11] == -1 && iuc[1] < this->iuc[1]) {
		x.set(lc[0]+halfh, uc[1]+halfh, lc[2]+halfh);
		d = findCell(x);
		if (d->level >= l) vert[11] = d->vertices[4];
	}
	if (vert[11] == -1) {
		vert[11] = p;
		addVertex(p,lc[0]+h,uc[1],lc[2],i[0],i[1],i[2]);
		p++;
		found[11]++;
	} 
	
	i.set(ilc[0]+ih,iuc[1],ilc[2]+ih);
	vert[12] = p;
	addVertex(p,lc[0]+h,uc[1],lc[2]+h,i[0],i[1],i[2]);
	p++;
	found[12]++;
	
	i.set(ilc[0]+ih,iuc[1],iuc[2]);
	vert[13] = p;
	addVertex(p,lc[0]+h,uc[1],uc[2],i[0],i[1],i[2]);
	p++;
	found[13]++;
	
	i.set(iuc[0],ilc[1],ilc[2]+ih);
	if (ilc[1] > 0) {
		x.set(uc[0]-halfh, lc[1]-halfh, lc[2]+halfh);
		d = findCell(x);
		if (d->level >= l) vert[14] = d->vertices[7];
	}
	if (vert[14] == -1 && iuc[0] < this->iuc[0] && ilc[1] > 0) {
		x.set(uc[0]+halfh, lc[1]-halfh, lc[2]+halfh);
		d = findCell(x);
		if (d->level >= l) vert[14] = d->vertices[3];
	}
	if (vert[14] == -1 && iuc[0] < this->iuc[0]) {
		x.set(lc[0]+halfh, lc[1]+halfh, lc[2]+halfh);
		d = findCell(x);
		if (d->level >= l) vert[14] = d->vertices[1];
	}

	if (vert[14] == -1) {
		vert[14] = p;
		addVertex(p,uc[0],lc[1],lc[2]+h,i[0],i[1],i[2]);
		p++;
		found[14]++;
	} 
	
	i.set(iuc[0],ilc[1]+ih,ilc[2]);
	if (ilc[2] > 0) {
		x.set(uc[0]-halfh, lc[1]+halfh, lc[2]-halfh);
		d = findCell(x);
		if (d->level >= l) vert[15] = d->vertices[7];
	}
	if (vert[15] == -1 && iuc[0] < this->iuc[0] && ilc[2] > 0) {
		x.set(uc[0]+halfh, lc[1]+halfh, lc[2]-halfh);
		d = findCell(x);
		if (d->level >= l) vert[15] = d->vertices[3];
	}
	if (vert[15] == -1 && iuc[0] < this->iuc[0]) {
		x.set(uc[0]+halfh, lc[1]+halfh, lc[2]+halfh);
		d = findCell(x);
		if (d->level >= l) vert[15] = d->vertices[2];
	}
	if (vert[15] == -1) {
		vert[15] = p;
		addVertex(p,uc[0],lc[1]+h,lc[2],i[0],i[1],i[2]);
		p++;
		found[15]++;
	} 
	
	i.set(iuc[0],ilc[1]+ih,ilc[2]+ih);
	vert[16] = p;
	addVertex(p,uc[0],lc[1]+h,lc[2]+h,i[0],i[1],i[2]);
	p++;
	found[16]++;
	
	i.set(iuc[0],ilc[1]+ih,iuc[2]);
	vert[17] = p;
	addVertex(p,uc[0],lc[1]+h,uc[2],i[0],i[1],i[2]);
	p++;
	found[17]++;
	
	i.set(iuc[0],iuc[1],ilc[2]+ih);
	vert[18] = p;
	addVertex(p,uc[0],uc[1],lc[2]+h,i[0],i[1],i[2]);
	p++;
	found[18]++;

	int n = cells.size();
	cells.resize(n+8);
	cell = &(cells[c]);
	cell->child[0] = n;
	cell->child[1] = n+1;
	cell->child[2] = n+2;
	cell->child[3] = n+3;
	cell->child[4] = n+4;
	cell->child[5] = n+5;
	cell->child[6] = n+6;
	cell->child[7] = n+7;

	cell = &(cells[n]);
	addCell(cell,c,l,cells[c].vertices[0], vert[0], vert[1], vert[2], vert[3],
					vert[4],vert[5],vert[6]);

	cell = &(cells[n+1]);
	addCell(cell,c,l, vert[0], cells[c].vertices[1], vert[2], vert[7], vert[4],
					vert[8],vert[6],vert[9]);

	cell = &(cells[n+2]);
	addCell(cell,c,l,vert[1], vert[2], cells[c].vertices[2], vert[10], vert[5],
					vert[6],vert[11],vert[12]);

	cell = &(cells[n+3]);
	addCell(cell,c,l,vert[2], vert[7], vert[10], cells[c].vertices[3], vert[6],
					vert[9],vert[12],vert[13]);

	cell = &(cells[n+4]);
	addCell(cell,c,l,vert[3], vert[4], vert[5], vert[6], cells[c].vertices[4],
					vert[14],vert[15],vert[16]);

	cell = &(cells[n+5]);
	addCell(cell,c,l,vert[4], vert[8], vert[6], vert[9], vert[14],
					cells[c].vertices[5],vert[16],vert[17]);

	cell = &(cells[n+6]);
	addCell(cell,c,l,vert[5], vert[6], vert[11], vert[12], vert[15],
					vert[16],cells[c].vertices[6],vert[18]);

	cell = &(cells[n+7]);
	addCell(cell,c,l,vert[6], vert[9], vert[12], vert[13], vert[16],
					vert[17],vert[18],cells[c].vertices[7]);

	cells[n+0].neighbors[3] = n+1;
	cells[n+0].neighbors[4] = n+2;
	cells[n+0].neighbors[13] = n+3;
	cells[n+0].neighbors[5] = n+4;
	cells[n+0].neighbors[16] = n+5;
	cells[n+0].neighbors[17] = n+6;
	cells[n+0].neighbors[25] = n+7;
	cells[n+1].neighbors[2] = n+0;
	cells[n+1].neighbors[12] = n+2;
	cells[n+1].neighbors[4] = n+3;
	cells[n+1].neighbors[15] = n+4;
	cells[n+1].neighbors[5] = n+5;
	cells[n+1].neighbors[24] = n+6;
	cells[n+1].neighbors[17] = n+7;
	cells[n+2].neighbors[1] = n+0;
	cells[n+2].neighbors[11] = n+1;
	cells[n+2].neighbors[3] = n+3;
	cells[n+2].neighbors[14] = n+4;
	cells[n+2].neighbors[23] = n+5;
	cells[n+2].neighbors[5] = n+6;
	cells[n+2].neighbors[16] = n+7;
	cells[n+3].neighbors[10] = n+0;
	cells[n+3].neighbors[1] = n+1;
	cells[n+3].neighbors[2] = n+2;
	cells[n+3].neighbors[22] = n+4;
	cells[n+3].neighbors[14] = n+5;
	cells[n+3].neighbors[15] = n+6;
	cells[n+3].neighbors[5] = n+7;
	cells[n+4].neighbors[0] = n+0;
	cells[n+4].neighbors[8] = n+1;
	cells[n+4].neighbors[9] = n+2;
	cells[n+4].neighbors[21] = n+3;
	cells[n+4].neighbors[3] = n+5;
	cells[n+4].neighbors[4] = n+6;
	cells[n+4].neighbors[13] = n+7;
	cells[n+5].neighbors[7] = n+0;
	cells[n+5].neighbors[0] = n+1;
	cells[n+5].neighbors[20] = n+2;
	cells[n+5].neighbors[9] = n+3;
	cells[n+5].neighbors[2] = n+4;
	cells[n+5].neighbors[12] = n+6;
	cells[n+5].neighbors[4] = n+7;
	cells[n+6].neighbors[6] = n+0;
	cells[n+6].neighbors[19] = n+1;
	cells[n+6].neighbors[0] = n+2;
	cells[n+6].neighbors[8] = n+3;
	cells[n+6].neighbors[1] = n+4;
	cells[n+6].neighbors[11] = n+5;
	cells[n+6].neighbors[3] = n+7;
	cells[n+7].neighbors[18] = n+0;
	cells[n+7].neighbors[6] = n+1;
	cells[n+7].neighbors[7] = n+2;
	cells[n+7].neighbors[0] = n+3;
	cells[n+7].neighbors[10] = n+4;
	cells[n+7].neighbors[1] = n+5;
	cells[n+7].neighbors[2] = n+6;

	if (cells[c].neighbors[0] != -1) {
		DtCell &neighbor = cells[cells[c].neighbors[0]];
		if (neighbor.child[0] != -1) {
			cells[n+0].neighbors[0] = neighbor.child[4];
			cells[neighbor.child[4]].neighbors[5] = n+0;
			cells[n+0].neighbors[8] = neighbor.child[5];
			cells[neighbor.child[5]].neighbors[15] = n+0;
			cells[n+0].neighbors[9] = neighbor.child[6];
			cells[neighbor.child[6]].neighbors[14] = n+0;
			cells[n+0].neighbors[21] = neighbor.child[7];
			cells[neighbor.child[7]].neighbors[22] = n+0;

			cells[n+1].neighbors[7] = neighbor.child[4];
			cells[neighbor.child[4]].neighbors[16] = n+1;
			cells[n+1].neighbors[0] = neighbor.child[5];
			cells[neighbor.child[5]].neighbors[5] = n+1;
			cells[n+1].neighbors[20] = neighbor.child[6];
			cells[neighbor.child[6]].neighbors[23] = n+1;
			cells[n+1].neighbors[9] = neighbor.child[7];
			cells[neighbor.child[7]].neighbors[14] = n+1;

			cells[n+2].neighbors[6] = neighbor.child[4];
			cells[neighbor.child[4]].neighbors[17] = n+2;
			cells[n+2].neighbors[19] = neighbor.child[5];
			cells[neighbor.child[5]].neighbors[24] = n+2;
			cells[n+2].neighbors[0] = neighbor.child[6];
			cells[neighbor.child[6]].neighbors[5] = n+2;
			cells[n+2].neighbors[8] = neighbor.child[7];
			cells[neighbor.child[7]].neighbors[15] = n+2;

			cells[n+3].neighbors[18] = neighbor.child[4];
			cells[neighbor.child[4]].neighbors[25] = n+3;
			cells[n+3].neighbors[6] = neighbor.child[5];
			cells[neighbor.child[5]].neighbors[17] = n+3;
			cells[n+3].neighbors[7] = neighbor.child[6];
			cells[neighbor.child[6]].neighbors[16] = n+3;
			cells[n+3].neighbors[0] = neighbor.child[7];
			cells[neighbor.child[7]].neighbors[5] = n+3;
		}
	}
	if (cells[c].neighbors[1] != -1) {
		DtCell &neighbor = cells[cells[c].neighbors[1]];
		if (neighbor.child[0] != -1) {
			cells[n+0].neighbors[1] = neighbor.child[2];
			cells[neighbor.child[2]].neighbors[4] = n+0;
			cells[n+0].neighbors[11] = neighbor.child[3];
			cells[neighbor.child[3]].neighbors[12] = n+0;
			cells[n+0].neighbors[14] = neighbor.child[6];
			cells[neighbor.child[6]].neighbors[9] = n+0;
			cells[n+0].neighbors[23] = neighbor.child[7];
			cells[neighbor.child[7]].neighbors[20] = n+0;

			cells[n+1].neighbors[10] = neighbor.child[2];
			cells[neighbor.child[2]].neighbors[13] = n+1;
			cells[n+1].neighbors[1] = neighbor.child[3];
			cells[neighbor.child[3]].neighbors[4] = n+1;
			cells[n+1].neighbors[22] = neighbor.child[6];
			cells[neighbor.child[6]].neighbors[21] = n+1;
			cells[n+1].neighbors[14] = neighbor.child[7];
			cells[neighbor.child[7]].neighbors[9] = n+1;

			cells[n+4].neighbors[6] = neighbor.child[2];
			cells[neighbor.child[2]].neighbors[17] = n+4;
			cells[n+4].neighbors[19] = neighbor.child[3];
			cells[neighbor.child[3]].neighbors[24] = n+4;
			cells[n+4].neighbors[1] = neighbor.child[6];
			cells[neighbor.child[6]].neighbors[4] = n+4;
			cells[n+4].neighbors[11] = neighbor.child[7];
			cells[neighbor.child[7]].neighbors[12] = n+4;

			cells[n+5].neighbors[18] = neighbor.child[2];
			cells[neighbor.child[2]].neighbors[25] = n+5;
			cells[n+5].neighbors[6] = neighbor.child[3];
			cells[neighbor.child[3]].neighbors[17] = n+5;
			cells[n+5].neighbors[10] = neighbor.child[6];
			cells[neighbor.child[6]].neighbors[13] = n+5;
			cells[n+5].neighbors[1] = neighbor.child[7];
			cells[neighbor.child[7]].neighbors[4] = n+5;
		}
	}
	if (cells[c].neighbors[2] != -1) {
		DtCell &neighbor = cells[cells[c].neighbors[2]];
		if (neighbor.child[0] != -1) {
			cells[n+0].neighbors[2] = neighbor.child[1];
			cells[neighbor.child[1]].neighbors[3] = n+0;
			cells[n+0].neighbors[12] = neighbor.child[3];
			cells[neighbor.child[3]].neighbors[11] = n+0;
			cells[n+0].neighbors[15] = neighbor.child[5];
			cells[neighbor.child[5]].neighbors[8] = n+0;
			cells[n+0].neighbors[24] = neighbor.child[7];
			cells[neighbor.child[7]].neighbors[19] = n+0;

			cells[n+2].neighbors[10] = neighbor.child[1];
			cells[neighbor.child[1]].neighbors[13] = n+2;
			cells[n+2].neighbors[2] = neighbor.child[3];
			cells[neighbor.child[3]].neighbors[3] = n+2;
			cells[n+2].neighbors[22] = neighbor.child[5];
			cells[neighbor.child[5]].neighbors[21] = n+2;
			cells[n+2].neighbors[15] = neighbor.child[7];
			cells[neighbor.child[7]].neighbors[8] = n+2;

			cells[n+4].neighbors[7] = neighbor.child[1];
			cells[neighbor.child[1]].neighbors[16] = n+4;
			cells[n+4].neighbors[20] = neighbor.child[3];
			cells[neighbor.child[3]].neighbors[23] = n+4;
			cells[n+4].neighbors[2] = neighbor.child[5];
			cells[neighbor.child[5]].neighbors[3] = n+4;
			cells[n+4].neighbors[12] = neighbor.child[7];
			cells[neighbor.child[7]].neighbors[11] = n+4;

			cells[n+6].neighbors[18] = neighbor.child[1];
			cells[neighbor.child[1]].neighbors[25] = n+6;
			cells[n+6].neighbors[7] = neighbor.child[3];
			cells[neighbor.child[3]].neighbors[16] = n+6;
			cells[n+6].neighbors[10] = neighbor.child[5];
			cells[neighbor.child[5]].neighbors[13] = n+6;
			cells[n+6].neighbors[2] = neighbor.child[7];
			cells[neighbor.child[7]].neighbors[3] = n+6;
		}
	}
	if (cells[c].neighbors[3] != -1) {
		DtCell &neighbor = cells[cells[c].neighbors[3]];
		if (neighbor.child[0] != -1) {
			cells[n+1].neighbors[3] = neighbor.child[0];
			cells[neighbor.child[0]].neighbors[2] = n+1;
			cells[n+1].neighbors[13] = neighbor.child[2];
			cells[neighbor.child[2]].neighbors[10] = n+1;
			cells[n+1].neighbors[16] = neighbor.child[4];
			cells[neighbor.child[4]].neighbors[7] = n+1;
			cells[n+1].neighbors[25] = neighbor.child[6];
			cells[neighbor.child[6]].neighbors[18] = n+1;

			cells[n+3].neighbors[11] = neighbor.child[0];
			cells[neighbor.child[0]].neighbors[12] = n+3;
			cells[n+3].neighbors[3] = neighbor.child[2];
			cells[neighbor.child[2]].neighbors[2] = n+3;
			cells[n+3].neighbors[23] = neighbor.child[4];
			cells[neighbor.child[4]].neighbors[20] = n+3;
			cells[n+3].neighbors[16] = neighbor.child[6];
			cells[neighbor.child[6]].neighbors[7] = n+3;

			cells[n+5].neighbors[8] = neighbor.child[0];
			cells[neighbor.child[0]].neighbors[15] = n+5;
			cells[n+5].neighbors[21] = neighbor.child[2];
			cells[neighbor.child[2]].neighbors[22] = n+5;
			cells[n+5].neighbors[3] = neighbor.child[4];
			cells[neighbor.child[4]].neighbors[2] = n+5;
			cells[n+5].neighbors[13] = neighbor.child[6];
			cells[neighbor.child[6]].neighbors[10] = n+5;

			cells[n+7].neighbors[19] = neighbor.child[0];
			cells[neighbor.child[0]].neighbors[24] = n+7;
			cells[n+7].neighbors[8] = neighbor.child[2];
			cells[neighbor.child[2]].neighbors[15] = n+7;
			cells[n+7].neighbors[11] = neighbor.child[4];
			cells[neighbor.child[4]].neighbors[12] = n+7;
			cells[n+7].neighbors[3] = neighbor.child[6];
			cells[neighbor.child[6]].neighbors[2] = n+7;
		}
	}
	if (cells[c].neighbors[4] != -1) {
		DtCell &neighbor = cells[cells[c].neighbors[4]];
		if (neighbor.child[0] != -1) {
			cells[n+2].neighbors[4] = neighbor.child[0];
			cells[neighbor.child[0]].neighbors[1] = n+2;
			cells[n+2].neighbors[13] = neighbor.child[1];
			cells[neighbor.child[1]].neighbors[10] = n+2;
			cells[n+2].neighbors[17] = neighbor.child[4];
			cells[neighbor.child[4]].neighbors[6] = n+2;
			cells[n+2].neighbors[25] = neighbor.child[5];
			cells[neighbor.child[5]].neighbors[18] = n+2;

			cells[n+3].neighbors[12] = neighbor.child[0];
			cells[neighbor.child[0]].neighbors[11] = n+3;
			cells[n+3].neighbors[4] = neighbor.child[1];
			cells[neighbor.child[1]].neighbors[1] = n+3;
			cells[n+3].neighbors[24] = neighbor.child[4];
			cells[neighbor.child[4]].neighbors[19] = n+3;
			cells[n+3].neighbors[17] = neighbor.child[5];
			cells[neighbor.child[5]].neighbors[6] = n+3;

			cells[n+4].neighbors[9] = neighbor.child[0];
			cells[neighbor.child[0]].neighbors[14] = n+4;
			cells[n+4].neighbors[21] = neighbor.child[1];
			cells[neighbor.child[1]].neighbors[22] = n+4;
			cells[n+4].neighbors[4] = neighbor.child[4];
			cells[neighbor.child[4]].neighbors[1] = n+4;
			cells[n+4].neighbors[13] = neighbor.child[5];
			cells[neighbor.child[5]].neighbors[10] = n+4;

			cells[n+5].neighbors[20] = neighbor.child[0];
			cells[neighbor.child[0]].neighbors[23] = n+5;
			cells[n+5].neighbors[9] = neighbor.child[1];
			cells[neighbor.child[1]].neighbors[14] = n+5;
			cells[n+5].neighbors[12] = neighbor.child[4];
			cells[neighbor.child[4]].neighbors[11] = n+5;
			cells[n+5].neighbors[4] = neighbor.child[5];
			cells[neighbor.child[5]].neighbors[1] = n+5;
		}
	}
	if (cells[c].neighbors[5] != -1) {
		DtCell &neighbor = cells[cells[c].neighbors[5]];
		if (neighbor.child[0] != -1) {
			cells[n+4].neighbors[5] = neighbor.child[0];
			cells[neighbor.child[0]].neighbors[0] = n+4;
			cells[n+4].neighbors[16] = neighbor.child[1];
			cells[neighbor.child[1]].neighbors[7] = n+4;
			cells[n+4].neighbors[17] = neighbor.child[2];
			cells[neighbor.child[2]].neighbors[6] = n+4;
			cells[n+4].neighbors[25] = neighbor.child[3];
			cells[neighbor.child[3]].neighbors[18] = n+4;

			cells[n+5].neighbors[15] = neighbor.child[0];
			cells[neighbor.child[0]].neighbors[8] = n+5;
			cells[n+5].neighbors[5] = neighbor.child[1];
			cells[neighbor.child[1]].neighbors[0] = n+5;
			cells[n+5].neighbors[24] = neighbor.child[2];
			cells[neighbor.child[2]].neighbors[19] = n+5;
			cells[n+5].neighbors[13] = neighbor.child[3];
			cells[neighbor.child[3]].neighbors[10] = n+5;

			cells[n+6].neighbors[14] = neighbor.child[0];
			cells[neighbor.child[0]].neighbors[9] = n+6;
			cells[n+6].neighbors[23] = neighbor.child[1];
			cells[neighbor.child[1]].neighbors[20] = n+6;
			cells[n+6].neighbors[5] = neighbor.child[2];
			cells[neighbor.child[2]].neighbors[0] = n+6;
			cells[n+6].neighbors[16] = neighbor.child[3];
			cells[neighbor.child[3]].neighbors[7] = n+6;

			cells[n+7].neighbors[22] = neighbor.child[0];
			cells[neighbor.child[0]].neighbors[21] = n+7;
			cells[n+7].neighbors[14] = neighbor.child[1];
			cells[neighbor.child[1]].neighbors[9] = n+7;
			cells[n+7].neighbors[15] = neighbor.child[2];
			cells[neighbor.child[2]].neighbors[8] = n+7;
			cells[n+7].neighbors[5] = neighbor.child[3];
			cells[neighbor.child[3]].neighbors[0] = n+7;
		}
	}
	if (cells[c].neighbors[6] != -1) {
		DtCell &neighbor = cells[cells[c].neighbors[6]];
		if (neighbor.child[0] != -1) {
			cells[n+0].neighbors[6] = neighbor.child[6];
			cells[neighbor.child[6]].neighbors[17] = n+0;
			cells[n+0].neighbors[19] = neighbor.child[7];
			cells[neighbor.child[7]].neighbors[24] = n+0;

			cells[n+1].neighbors[18] = neighbor.child[6];
			cells[neighbor.child[6]].neighbors[25] = n+1;
			cells[n+1].neighbors[6] = neighbor.child[7];
			cells[neighbor.child[7]].neighbors[17] = n+1;
		}
	}
	if (cells[c].neighbors[7] != -1) {
		DtCell &neighbor = cells[cells[c].neighbors[7]];
		if (neighbor.child[0] != -1) {
			cells[n+0].neighbors[7] = neighbor.child[5];
			cells[neighbor.child[5]].neighbors[16] = n+0;
			cells[n+0].neighbors[20] = neighbor.child[7];
			cells[neighbor.child[7]].neighbors[23] = n+0;

			cells[n+2].neighbors[18] = neighbor.child[5];
			cells[neighbor.child[5]].neighbors[25] = n+2;
			cells[n+2].neighbors[7] = neighbor.child[7];
			cells[neighbor.child[7]].neighbors[16] = n+2;
		}
	}
	if (cells[c].neighbors[8] != -1) {
		DtCell &neighbor = cells[cells[c].neighbors[8]];
		if (neighbor.child[0] != -1) {
			cells[n+1].neighbors[8] = neighbor.child[4];
			cells[neighbor.child[4]].neighbors[15] = n+1;
			cells[n+1].neighbors[21] = neighbor.child[6];
			cells[neighbor.child[6]].neighbors[22] = n+1;

			cells[n+3].neighbors[19] = neighbor.child[4];
			cells[neighbor.child[4]].neighbors[24] = n+3;
			cells[n+3].neighbors[8] = neighbor.child[6];
			cells[neighbor.child[6]].neighbors[15] = n+3;
		}
	}
	if (cells[c].neighbors[9] != -1) {
		DtCell &neighbor = cells[cells[c].neighbors[9]];
		if (neighbor.child[0] != -1) {
			cells[n+2].neighbors[9] = neighbor.child[4];
			cells[neighbor.child[4]].neighbors[14] = n+2;
			cells[n+2].neighbors[21] = neighbor.child[5];
			cells[neighbor.child[5]].neighbors[22] = n+2;

			cells[n+3].neighbors[20] = neighbor.child[4];
			cells[neighbor.child[4]].neighbors[23] = n+3;
			cells[n+3].neighbors[9] = neighbor.child[5];
			cells[neighbor.child[5]].neighbors[14] = n+3;
		}
	}
	if (cells[c].neighbors[10] != -1) {
		DtCell &neighbor = cells[cells[c].neighbors[10]];
		if (neighbor.child[0] != -1) {
			cells[n+0].neighbors[10] = neighbor.child[3];
			cells[neighbor.child[3]].neighbors[13] = n+0;
			cells[n+0].neighbors[22] = neighbor.child[7];
			cells[neighbor.child[7]].neighbors[21] = n+0;

			cells[n+4].neighbors[18] = neighbor.child[3];
			cells[neighbor.child[3]].neighbors[25] = n+4;
			cells[n+4].neighbors[10] = neighbor.child[7];
			cells[neighbor.child[7]].neighbors[13] = n+4;
		}
	}
	if (cells[c].neighbors[11] != -1) {
		DtCell &neighbor = cells[cells[c].neighbors[11]];
		if (neighbor.child[0] != -1) {
			cells[n+1].neighbors[11] = neighbor.child[2];
			cells[neighbor.child[2]].neighbors[12] = n+1;
			cells[n+1].neighbors[23] = neighbor.child[6];
			cells[neighbor.child[6]].neighbors[20] = n+1;

			cells[n+5].neighbors[19] = neighbor.child[2];
			cells[neighbor.child[2]].neighbors[24] = n+5;
			cells[n+5].neighbors[11] = neighbor.child[6];
			cells[neighbor.child[6]].neighbors[12] = n+5;
		}
	}
	if (cells[c].neighbors[12] != -1) {
		DtCell &neighbor = cells[cells[c].neighbors[12]];
		if (neighbor.child[0] != -1) {
			cells[n+2].neighbors[12] = neighbor.child[1];
			cells[neighbor.child[1]].neighbors[11] = n+2;
			cells[n+2].neighbors[24] = neighbor.child[5];
			cells[neighbor.child[5]].neighbors[19] = n+2;

			cells[n+6].neighbors[20] = neighbor.child[1];
			cells[neighbor.child[1]].neighbors[23] = n+6;
			cells[n+6].neighbors[12] = neighbor.child[5];
			cells[neighbor.child[5]].neighbors[11] = n+6;
		}
	}
	if (cells[c].neighbors[13] != -1) {
		DtCell &neighbor = cells[cells[c].neighbors[13]];
		if (neighbor.child[0] != -1) {
			cells[n+3].neighbors[13] = neighbor.child[0];
			cells[neighbor.child[0]].neighbors[10] = n+3;
			cells[n+3].neighbors[25] = neighbor.child[4];
			cells[neighbor.child[4]].neighbors[18] = n+3;

			cells[n+7].neighbors[21] = neighbor.child[0];
			cells[neighbor.child[0]].neighbors[22] = n+7;
			cells[n+7].neighbors[13] = neighbor.child[4];
			cells[neighbor.child[4]].neighbors[10] = n+7;
		}
	}
	if (cells[c].neighbors[14] != -1) {
		DtCell &neighbor = cells[cells[c].neighbors[14]];
		if (neighbor.child[0] != -1) {
			cells[n+4].neighbors[14] = neighbor.child[2];
			cells[neighbor.child[2]].neighbors[9] = n+4;
			cells[n+4].neighbors[23] = neighbor.child[3];
			cells[neighbor.child[3]].neighbors[20] = n+4;

			cells[n+5].neighbors[22] = neighbor.child[2];
			cells[neighbor.child[2]].neighbors[21] = n+5;
			cells[n+5].neighbors[14] = neighbor.child[3];
			cells[neighbor.child[3]].neighbors[9] = n+5;
		}
	}
	if (cells[c].neighbors[15] != -1) {
		DtCell &neighbor = cells[cells[c].neighbors[15]];
		if (neighbor.child[0] != -1) {
			cells[n+4].neighbors[15] = neighbor.child[1];
			cells[neighbor.child[1]].neighbors[8] = n+4;
			cells[n+4].neighbors[24] = neighbor.child[3];
			cells[neighbor.child[3]].neighbors[19] = n+4;

			cells[n+6].neighbors[22] = neighbor.child[1];
			cells[neighbor.child[1]].neighbors[21] = n+6;
			cells[n+6].neighbors[15] = neighbor.child[3];
			cells[neighbor.child[3]].neighbors[8] = n+6;
		}
	}
	if (cells[c].neighbors[16] != -1) {
		DtCell &neighbor = cells[cells[c].neighbors[16]];
		if (neighbor.child[0] != -1) {
			cells[n+5].neighbors[16] = neighbor.child[0];
			cells[neighbor.child[0]].neighbors[7] = n+5;
			cells[n+5].neighbors[25] = neighbor.child[2];
			cells[neighbor.child[2]].neighbors[18] = n+5;

			cells[n+7].neighbors[23] = neighbor.child[0];
			cells[neighbor.child[0]].neighbors[20] = n+7;
			cells[n+7].neighbors[16] = neighbor.child[2];
			cells[neighbor.child[2]].neighbors[7] = n+7;
		}
	}
	if (cells[c].neighbors[17] != -1) {
		DtCell &neighbor = cells[cells[c].neighbors[17]];
		if (neighbor.child[0] != -1) {
			cells[n+6].neighbors[17] = neighbor.child[0];
			cells[neighbor.child[0]].neighbors[6] = n+6;
			cells[n+6].neighbors[25] = neighbor.child[1];
			cells[neighbor.child[1]].neighbors[18] = n+6;

			cells[n+7].neighbors[24] = neighbor.child[0];
			cells[neighbor.child[0]].neighbors[19] = n+7;
			cells[n+7].neighbors[17] = neighbor.child[1];
			cells[neighbor.child[1]].neighbors[6] = n+7;
		}
	}
	if (cells[c].neighbors[18] != -1) {
		DtCell &neighbor = cells[cells[c].neighbors[18]];
		if (neighbor.child[0] != -1) {
			cells[n+0].neighbors[18] = neighbor.child[7];
			cells[neighbor.child[7]].neighbors[25] = n+0;
		}
	}
	if (cells[c].neighbors[19] != -1) {
		DtCell &neighbor = cells[cells[c].neighbors[19]];
		if (neighbor.child[0] != -1) {
			cells[n+1].neighbors[19] = neighbor.child[6];
			cells[neighbor.child[6]].neighbors[24] = n+1;
		}
	}
	if (cells[c].neighbors[20] != -1) {
		DtCell &neighbor = cells[cells[c].neighbors[20]];
		if (neighbor.child[0] != -1) {
			cells[n+2].neighbors[20] = neighbor.child[5];
			cells[neighbor.child[5]].neighbors[23] = n+2;
		}
	}
	if (cells[c].neighbors[21] != -1) {
		DtCell &neighbor = cells[cells[c].neighbors[21]];
		if (neighbor.child[0] != -1) {
			cells[n+3].neighbors[21] = neighbor.child[4];
			cells[neighbor.child[4]].neighbors[22] = n+3;
		}
	}
	if (cells[c].neighbors[22] != -1) {
		DtCell &neighbor = cells[cells[c].neighbors[22]];
		if (neighbor.child[0] != -1) {
			cells[n+4].neighbors[22] = neighbor.child[3];
			cells[neighbor.child[3]].neighbors[21] = n+4;
		}
	}
	if (cells[c].neighbors[23] != -1) {
		DtCell &neighbor = cells[cells[c].neighbors[23]];
		if (neighbor.child[0] != -1) {
			cells[n+5].neighbors[23] = neighbor.child[2];
			cells[neighbor.child[2]].neighbors[20] = n+5;
		}
	}
	if (cells[c].neighbors[24] != -1) {
		DtCell &neighbor = cells[cells[c].neighbors[24]];
		if (neighbor.child[0] != -1) {
			cells[n+6].neighbors[24] = neighbor.child[1];
			cells[neighbor.child[1]].neighbors[19] = n+6;
		}
	}
	if (cells[c].neighbors[25] != -1) {
		DtCell &neighbor = cells[cells[c].neighbors[25]];
		if (neighbor.child[0] != -1) {
			cells[n+7].neighbors[25] = neighbor.child[0];
			cells[neighbor.child[0]].neighbors[18] = n+7;
		}
	}
}

void DtTree::buildTree(SlVector3 input_lc, SlVector3 input_uc, int max_level,
											 SplitCriterion &splitCriterion) {
  unsigned int c=0;
  this->lc = input_lc;
  this->uc = input_uc;
	this->max_level = max_level;
	this->ilc = 0;
	this->iuc = (int)pow2(max_level);
	dim = (int)pow2(max_level)+1;
	dim2 = dim*dim;
	nvertices = dim2*dim;
	DtCell *cell;
	pts.resize(8);
	ipts.resize(8);
  cells.resize(1);
	neighbors.resize(8);
  cells[0].level = 0;
  cells[0].vertices[0] = 0;
  cells[0].vertices[1] = 1;
  cells[0].vertices[2] = 2;
  cells[0].vertices[3] = 3;
  cells[0].vertices[4] = 4;
  cells[0].vertices[5] = 5;
  cells[0].vertices[6] = 6;
  cells[0].vertices[7] = 7;
  cells[0].parent = -1;
  cells[0].child[0] = cells[0].child[1] = cells[0].child[2] = cells[0].child[3] =
    cells[0].child[4] = cells[0].child[5] = cells[0].child[6] = cells[0].child[7] = -1;

	pts[0].set(this->lc[0], this->lc[1], this->lc[2]);
	ipts[0].set(ilc[0], ilc[1], ilc[2]);

	pts[1].set(this->lc[0], this->lc[1], this->uc[2]);
	ipts[1].set(ilc[0], ilc[1], iuc[2]);
	
	pts[2].set(this->lc[0], this->uc[1], this->lc[2]);
	ipts[2].set(ilc[0], iuc[1], ilc[2]);

	pts[3].set(this->lc[0], this->uc[1], this->uc[2]);
	ipts[3].set(ilc[0], iuc[1], iuc[2]);

	pts[4].set(this->uc[0], this->lc[1], this->lc[2]);
	ipts[4].set(iuc[0], ilc[1], ilc[2]);

	pts[5].set(this->uc[0], this->lc[1], this->uc[2]);
	ipts[5].set(iuc[0], ilc[1], iuc[2]);

	pts[6].set(this->uc[0], this->uc[1], this->lc[2]);
	ipts[6].set(iuc[0], iuc[1], ilc[2]);

	pts[7].set(this->uc[0], this->uc[1], this->uc[2]);
	ipts[7].set(iuc[0], iuc[1], iuc[2]);

	neighbors[0].set(-1);
	neighbors[0][0] = 4;
	neighbors[0][2] = 2;
	neighbors[0][4] = 1;
	
	neighbors[1].set(-1);
	neighbors[1][0] = 5;
	neighbors[1][2] = 3;
	neighbors[1][5] = 0;

	neighbors[2].set(-1);
	neighbors[2][0] = 6;
	neighbors[2][3] = 0;
	neighbors[2][4] = 3;

	neighbors[3].set(-1);
	neighbors[3][0] = 7;
	neighbors[3][3] = 1;
	neighbors[3][5] = 2;

	neighbors[4].set(-1);
	neighbors[4][1] = 0;
	neighbors[4][2] = 6;
	neighbors[4][4] = 5;
	
	neighbors[5].set(-1);
	neighbors[5][1] = 1;
	neighbors[5][2] = 7;
	neighbors[5][5] = 4;

	neighbors[6].set(-1);
	neighbors[6][1] = 2;
	neighbors[6][3] = 4;
	neighbors[6][4] = 7;

	neighbors[7].set(-1);
	neighbors[7][1] = 3;
	neighbors[7][3] = 5;
	neighbors[7][5] = 6;

	cells[0].neighbors[0] = cells[0].neighbors[1] = cells[0].neighbors[2] = cells[0].neighbors[3] = cells[0].neighbors[4] = 
		cells[0].neighbors[5] = cells[0].neighbors[6] = cells[0].neighbors[7] = cells[0].neighbors[8] = cells[0].neighbors[9] = 
		cells[0].neighbors[10] = cells[0].neighbors[11] = cells[0].neighbors[12] = cells[0].neighbors[13] = cells[0].neighbors[14] = 
		cells[0].neighbors[15] = cells[0].neighbors[16] = cells[0].neighbors[17] = cells[0].neighbors[18] = cells[0].neighbors[19] = 
		cells[0].neighbors[20] = cells[0].neighbors[21] = cells[0].neighbors[22] = cells[0].neighbors[23] = cells[0].neighbors[24] = 
		cells[0].neighbors[25] = -1;		

	while (c < cells.size()) {
		cell = &(cells[c]);
		if (splitCriterion.doSplit(this, cell) && cell->level < max_level)
			divideCell(c,this);
		c++;
	} 

	cells.reserve(cells.size());
	pts.reserve(pts.size());
	ipts.reserve(ipts.size());
	for (c=0; c<cells.size(); c++)
		cells[c].triangles.reserve(cells[c].triangles.size());
}

inline int getChildId(std::vector<DtCell> cells, int c) {
	if (c == -1) return -1;
	int p = cells[c].parent;
	if (cells[p].child[0] == c) return 0;
	if (cells[p].child[1] == c) return 1;
	if (cells[p].child[2] == c) return 2;
	if (cells[p].child[3] == c) return 3;
	if (cells[p].child[4] == c) return 4;
	if (cells[p].child[5] == c) return 5;
	if (cells[p].child[6] == c) return 6;
	if (cells[p].child[7] == c) return 7;
	return -1;
}

void DtTree::cacheCellNeighbors() {
	std::vector<DtCell>::iterator cell;
	int c;
	SlVector3 y,z;
	double h;
	int i;

	for (i=0,cell=cells.begin(); cell!=cells.end(); cell++,i++) {
		if (cell->level != max_level) continue;

		y = 0.5*(pts[cell->vertices[0]]+pts[cell->vertices[7]]);
		h = pts[cell->vertices[7]][0] - pts[cell->vertices[0]][0];

		z[0]=y[0]-h; z[1]=y[1]; z[2]=y[2];
		c = findCellIndex(z);
		if (cells[c].contains(this, &(z[0]))) {
			if (cells[c].level == max_level) {
				cell->neighbors[0] = c;
			}
		}
		
		z[0]=y[0]; z[1]=y[1]-h; z[2]=y[2];
		c = findCellIndex(z);
		if (cells[c].contains(this, &(z[0]))) {
			if (cells[c].level == max_level) {
				cell->neighbors[1] = c;
			}
		}
			
		z[0]=y[0]; z[1]=y[1]; z[2]=y[2]-h;
		c = findCellIndex(z);
		if (cells[c].contains(this, &(z[0]))) {
			if (cells[c].level == max_level) {
				cell->neighbors[2] = c;
			}
		}
  
		z[0]=y[0]; z[1]=y[1]; z[2]=y[2]+h;
		c = findCellIndex(z);
		if (cells[c].contains(this, &(z[0]))) {
			if (cells[c].level == max_level) {
				cell->neighbors[3] = c;
			}
		}
  
		z[0]=y[0]; z[1]=y[1]+h; z[2]=y[2];
		c = findCellIndex(z);
		if (cells[c].contains(this, &(z[0]))) {
			if (cells[c].level == max_level) {
				cell->neighbors[4] = c;
			}
		}
		
		z[0]=y[0]+h; z[1]=y[1]; z[2]=y[2];
		c = findCellIndex(z);
		if (cells[c].contains(this, &(z[0]))) {
			if (cells[c].level == max_level) {
				cell->neighbors[5] = c;
			}
		}
  
		z[0]=y[0]-h; z[1]=y[1]-h; z[2]=y[2];
		c = findCellIndex(z);
		if (cells[c].contains(this, &(z[0]))) {
			if (cells[c].level == max_level) {
				cell->neighbors[6] = c;
			}
		}
  
		z[0]=y[0]-h; z[1]=y[1]; z[2]=y[2]-h;
		c = findCellIndex(z);
		if (cells[c].contains(this, &(z[0]))) {
			if (cells[c].level == max_level) {
				cell->neighbors[7] = c;
			}
		}
  
		z[0]=y[0]-h; z[1]=y[1]; z[2]=y[2]+h;
		c = findCellIndex(z);
		if (cells[c].contains(this, &(z[0]))) {
			if (cells[c].level == max_level) {
				cell->neighbors[8] = c;
			}
		}
  
		z[0]=y[0]-h; z[1]=y[1]+h; z[2]=y[2];
		c = findCellIndex(z);
		if (cells[c].contains(this, &(z[0]))) {
			if (cells[c].level == max_level) {
				cell->neighbors[9] = c;
			}
		}
  
		z[0]=y[0]; z[1]=y[1]-h; z[2]=y[2]-h;
		c = findCellIndex(z);
		if (cells[c].contains(this, &(z[0]))) {
			if (cells[c].level == max_level) {
				cell->neighbors[10] = c;
			}
		}
	
		z[0]=y[0]; z[1]=y[1]-h; z[2]=y[2]+h;
		c = findCellIndex(z);
		if (cells[c].contains(this, &(z[0]))) {
			if (cells[c].level == max_level) {
				cell->neighbors[11] = c;
			}
		}
  
		z[0]=y[0]; z[1]=y[1]+h; z[2]=y[2]-h;
		c = findCellIndex(z);
		if (cells[c].contains(this, &(z[0]))) {
			if (cells[c].level == max_level) {
				cell->neighbors[12] = c;
			}
		}

		z[0]=y[0]; z[1]=y[1]+h; z[2]=y[2]+h;
		c = findCellIndex(z);
		if (cells[c].contains(this, &(z[0]))) {
			if (cells[c].level == max_level) {
				cell->neighbors[13] = c;
			}
		}
	
		z[0]=y[0]+h; z[1]=y[1]-h; z[2]=y[2];
		c = findCellIndex(z);
		if (cells[c].contains(this, &(z[0]))) {
			if (cells[c].level == max_level) {
				cell->neighbors[14] = c;
			}
		}
  
		z[0]=y[0]+h; z[1]=y[1]; z[2]=y[2]-h;
		c = findCellIndex(z);
		if (cells[c].contains(this, &(z[0]))) {
			if (cells[c].level == max_level) {
				cell->neighbors[15] = c;
			}
		}
	
		z[0]=y[0]+h; z[1]=y[1]; z[2]=y[2]+h;
		c = findCellIndex(z);
		if (cells[c].contains(this, &(z[0]))) {
			if (cells[c].level == max_level) {
				cell->neighbors[16] = c;
			}
		}
	
		z[0]=y[0]+h; z[1]=y[1]+h; z[2]=y[2];
		c = findCellIndex(z);
		if (cells[c].contains(this, &(z[0]))) {
			if (cells[c].level == max_level) {
				cell->neighbors[17] = c;
			}
		}
	
		z[0]=y[0]-h; z[1]=y[1]-h; z[2]=y[2]-h;
		c = findCellIndex(z);
		if (cells[c].contains(this, &(z[0]))) {
			if (cells[c].level == max_level) {
				cell->neighbors[18] = c;
			}
		}
  
		z[0]=y[0]-h; z[1]=y[1]-h; z[2]=y[2]+h;
		c = findCellIndex(z);
		if (cells[c].contains(this, &(z[0]))) {
			if (cells[c].level == max_level) {
				cell->neighbors[19] = c;
			}
		}

		z[0]=y[0]-h; z[1]=y[1]+h; z[2]=y[2]-h;
		c = findCellIndex(z);
		if (cells[c].contains(this, &(z[0]))) {
			if (cells[c].level == max_level) {
				cell->neighbors[20] = c;
			}
		}

		z[0]=y[0]-h; z[1]=y[1]+h; z[2]=y[2]+h;
		c = findCellIndex(z);
		if (cells[c].contains(this, &(z[0]))) {
			if (cells[c].level == max_level) {
				cell->neighbors[21] = c;
			}
		}

		z[0]=y[0]+h; z[1]=y[1]-h; z[2]=y[2]-h;
		c = findCellIndex(z);
		if (cells[c].contains(this, &(z[0]))) {
			if (cells[c].level == max_level) {
				cell->neighbors[22] = c;
			}
		}

		z[0]=y[0]+h; z[1]=y[1]-h; z[2]=y[2]+h;
		c = findCellIndex(z);
		if (cells[c].contains(this, &(z[0]))) {
			if (cells[c].level == max_level) {
				cell->neighbors[23] = c;
			}
		}

		z[0]=y[0]+h; z[1]=y[1]+h; z[2]=y[2]-h;
		c = findCellIndex(z);
		if (cells[c].contains(this, &(z[0]))) {
			if (cells[c].level == max_level) {
				cell->neighbors[24] = c;
			}
		}
		
		z[0]=y[0]+h; z[1]=y[1]+h; z[2]=y[2]+h;
		c = findCellIndex(z);
		if (cells[c].contains(this, &(z[0]))) {
			if (cells[c].level == max_level) {
				cell->neighbors[25] = c;
			}
		}
	}
}  

void DtTree::balanceTree() {
	unsigned int c=0;
	DtCell *cell1, *cell2;
	double LEPS = 1e-100;
	int done = false;
	SlVector3 x, v[8];

	while (!done) {
		done = true;
		while (c < cells.size()) {
			cell1 = &(cells[c]);
			v[0] = pts[cell1->vertices[0]];
			v[1] = pts[cell1->vertices[1]];
			v[2] = pts[cell1->vertices[2]];
			v[3] = pts[cell1->vertices[3]];
			v[4] = pts[cell1->vertices[4]];
			v[5] = pts[cell1->vertices[5]];
			v[6] = pts[cell1->vertices[6]];
			v[7] = pts[cell1->vertices[7]];
			
			x.set(v[0][0]-LEPS, v[0][1]+LEPS, v[0][2]+LEPS);
			cell2 = findCell(x);
			if (cell2->level > cell1->level+1) { divideCell(c,this); done = false; continue; }

			x.set(v[1][0]-LEPS, v[1][1]+LEPS, v[1][2]-LEPS);
			cell2 = findCell(x);
			if (cell2->level > cell1->level+1) { divideCell(c,this); done = false; continue; }

			x.set(v[2][0]-LEPS, v[2][1]-LEPS, v[2][2]+LEPS);
			cell2 = findCell(x);
			if (cell2->level > cell1->level+1) { divideCell(c,this); done = false; continue; }

			x.set(v[3][0]-LEPS, v[3][1]-LEPS, v[3][2]-LEPS);
			cell2 = findCell(x);
			if (cell2->level > cell1->level+1) { divideCell(c,this); done = false; continue; }

			x.set(v[4][0]+LEPS, v[4][1]+LEPS, v[4][2]+LEPS);
			cell2 = findCell(x);
			if (cell2->level > cell1->level+1) { divideCell(c,this); done = false; continue; }

			x.set(v[5][0]+LEPS, v[5][1]+LEPS, v[5][2]-LEPS);
			cell2 = findCell(x);
			if (cell2->level > cell1->level+1) { divideCell(c,this); done = false; continue; }

			x.set(v[6][0]+LEPS, v[6][1]-LEPS, v[6][2]+LEPS);
			cell2 = findCell(x);
			if (cell2->level > cell1->level+1) { divideCell(c,this); done = false; continue; }

			x.set(v[7][0]+LEPS, v[7][1]-LEPS, v[7][2]-LEPS);
			cell2 = findCell(x);
			if (cell2->level > cell1->level+1) { divideCell(c,this); done = false; continue; }

			x.set(v[0][0]+LEPS, v[0][1]-LEPS, v[0][2]+LEPS);
			cell2 = findCell(x);
			if (cell2->level > cell1->level+1) { divideCell(c,this); done = false; continue; }

			x.set(v[1][0]+LEPS, v[1][1]-LEPS, v[1][2]-LEPS);
			cell2 = findCell(x);
			if (cell2->level > cell1->level+1) { divideCell(c,this); done = false; continue; }

			x.set(v[4][0]-LEPS, v[4][1]-LEPS, v[4][2]+LEPS);
			cell2 = findCell(x);
			if (cell2->level > cell1->level+1) { divideCell(c,this); done = false; continue; }

			x.set(v[5][0]-LEPS, v[5][1]-LEPS, v[5][2]-LEPS);
			cell2 = findCell(x);
			if (cell2->level > cell1->level+1) { divideCell(c,this); done = false; continue; }

			x.set(v[2][0]+LEPS, v[2][1]+LEPS, v[2][2]+LEPS);
			cell2 = findCell(x);
			if (cell2->level > cell1->level+1) { divideCell(c,this); done = false; continue; }

			x.set(v[3][0]+LEPS, v[3][1]+LEPS, v[3][2]-LEPS);
			cell2 = findCell(x);
			if (cell2->level > cell1->level+1) { divideCell(c,this); done = false; continue; }

			x.set(v[6][0]-LEPS, v[6][1]+LEPS, v[6][2]+LEPS);
			cell2 = findCell(x);
			if (cell2->level > cell1->level+1) { divideCell(c,this); done = false; continue; }

			x.set(v[6][0]-LEPS, v[6][1]+LEPS, v[6][2]-LEPS);
			cell2 = findCell(x);
			if (cell2->level > cell1->level+1) { divideCell(c,this); done = false; continue; }

			x.set(v[0][0]+LEPS, v[0][1]+LEPS, v[0][2]-LEPS);
			cell2 = findCell(x);
			if (cell2->level > cell1->level+1) { divideCell(c,this); done = false; continue; }

			x.set(v[2][0]+LEPS, v[2][1]-LEPS, v[2][2]-LEPS);
			cell2 = findCell(x);
			if (cell2->level > cell1->level+1) { divideCell(c,this); done = false; continue; }

			x.set(v[4][0]-LEPS, v[4][1]+LEPS, v[4][2]-LEPS);
			cell2 = findCell(x);
			if (cell2->level > cell1->level+1) { divideCell(c,this); done = false; continue; }

			x.set(v[6][0]-LEPS, v[6][1]+LEPS, v[6][2]-LEPS);
			cell2 = findCell(x);
			if (cell2->level > cell1->level+1) { divideCell(c,this); done = false; continue; }

			x.set(v[1][0]+LEPS, v[1][1]+LEPS, v[1][2]+LEPS);
			cell2 = findCell(x);
			if (cell2->level > cell1->level+1) { divideCell(c,this); done = false; continue; }

			x.set(v[3][0]+LEPS, v[3][1]-LEPS, v[3][2]+LEPS);
			cell2 = findCell(x);
			if (cell2->level > cell1->level+1) { divideCell(c,this); done = false; continue; }

			x.set(v[5][0]-LEPS, v[5][1]+LEPS, v[5][2]+LEPS);
			cell2 = findCell(x);
			if (cell2->level > cell1->level+1) { divideCell(c,this); done = false; continue; }

			x.set(v[7][0]-LEPS, v[7][1]-LEPS, v[7][2]+LEPS);
			cell2 = findCell(x);
			if (cell2->level > cell1->level+1) { divideCell(c,this); done = false; continue; }

			c++;
		}
	}
}

void buildTree (DtTree *tree, SlVector3 input_lc, SlVector3 input_uc, int max_level,
								std::vector<SlVector3> *meshPts, std::vector<SlTri> *triangles, 
								std::vector<SlVector3> *faceNormals) {
	TriangleSplitCriterion splitCriterion;
	splitCriterion.meshPts = meshPts;
	splitCriterion.triangles = triangles;
	splitCriterion.faceNormals = faceNormals;
	tree->buildTree(input_lc, input_uc, max_level, splitCriterion);
}

void buildTree (DtTree *tree, SlVector3 input_lc, SlVector3 input_uc, int max_level,
								double (*evalPhi)(const SlVector3 &x, void *data), void *evalPhiData) {
	std::vector<SlVector3>::const_iterator x;
	std::vector<double>::iterator p;
	PhiSplitCriterion splitCriterion;
	splitCriterion.phiFunc = evalPhi;
	splitCriterion.phiFuncData = evalPhiData;
	tree->buildTree(input_lc, input_uc, max_level, splitCriterion);

	tree->phi.resize(tree->pts.size());
	for (p=tree->phi.begin(),x=tree->pts.begin(); x!=tree->pts.end(); p++, x++) 
		(*p) = evalPhi((*x), evalPhiData);
}

