#include <sys/time.h>
#include <iostream>
#include <slVector.H>
#include <slMatrix.H>
#include <vector>
#include <slcUtil.H>
#include <petscksp.h>
#include <fstream>

#include "staggeredGrid.H"

using std::cout;
using std::cerr;
using std::endl;

class PhiData {
public:
	StaggeredGrid *g;
	SlcSurface *s;
	double dt;
};

static const float EPS = 1e-12;

SlVector3 vel(const SlVector3 &x, void *data) {
	StaggeredGrid *grid = (StaggeredGrid*) data;
	SlVector3 y(x);
	y[0] = std::max<double>(y[0], grid->lc[0]+grid->h);
	y[1] = std::max<double>(y[1], grid->lc[1]+grid->h);
	y[2] = std::max<double>(y[2], grid->lc[2]+grid->h);
	y[0] = std::min<double>(y[0], grid->uc[0]-grid->h);
	y[1] = std::min<double>(y[1], grid->uc[1]-grid->h);
	y[2] = std::min<double>(y[2], grid->uc[2]-grid->h);

	return grid->interpVel(y);
}

double phi(const SlVector3 &x, void *data) {
	PhiData *pd = (PhiData *)data;
	SlVector3 y = slcTraceBack(x,vel,pd->dt,(void*)pd->g);
	return pd->s->eval(y);
}

bool updateSurface(StaggeredGrid *g, SlcSurface* &s, double dt) {
	timeval startTime, endTime;
	PhiData pd;
	pd.g = g;
	pd.s = s;
	pd.dt = dt;
	DtTree *newTree = new DtTree();
	SlcSurface *newSurface = new SlcSurface(newTree);
	double h = 0.0*pow2(-s->tree->max_level)*(s->tree->uc[0]-s->tree->lc[0]);
	SlVector3 offset(((((double)rand())/RAND_MAX)-0.5)*h,
									 ((((double)rand())/RAND_MAX)-0.5)*h,
									 ((((double)rand())/RAND_MAX)-0.5)*h);
	
	gettimeofday(&startTime, NULL);
	buildTree(newSurface->tree, g->lc-h+offset, g->uc+h+offset, s->tree->max_level, phi, (void*)&pd);
	gettimeofday(&endTime, NULL);
	cout<<"\tbuildTree took "<<(endTime.tv_sec-startTime.tv_sec)+
		(endTime.tv_usec-startTime.tv_usec)*1.0e-6<<endl;
	gettimeofday(&startTime, NULL);
	newSurface->contourTree(phi, (void*)&pd);
	gettimeofday(&endTime, NULL);
	cout<<"\tcontourTree took "<<(endTime.tv_sec-startTime.tv_sec)+
			(endTime.tv_usec-startTime.tv_usec)*1.0e-6<<endl;
	delete s->tree;
	delete s;
	s = newSurface;
	gettimeofday(&startTime, NULL);
	s->redistance();
	gettimeofday(&endTime, NULL);
	cout<<"\tredistance took "<<(endTime.tv_sec-startTime.tv_sec)+
		(endTime.tv_usec-startTime.tv_usec)*1.0e-6<<endl;
	return true;
}

double initPhi(const SlVector3 &x, void* data) {
	SlVector3 c(20, 20, 30);
	SlVector3 lc(0.5,0.5,0.5), uc(39.5,39.5,10.1);
  double sphere = mag(x-c)-5; 
	double m,n,o;
	m = absmin2(lc[0]-x[0],x[0]-uc[0]);
	n = absmin2(lc[1]-x[1],x[1]-uc[1]);
	o = absmin2(lc[2]-x[2],x[2]-uc[2]);
	double ret = min2(sphere, max3(m,n,o));
	return ret;
} 

inline static
std::istream &eatChar(char c,std::istream &buf) {
  char r;
  buf >> r;
  if (r!=c) {
    buf.clear(buf.rdstate() | std::ios::failbit);
  }
  return buf;
}

inline static
std::istream &eatStr(const char *s,std::istream &buf) {
  while (*s != '\0') {
    eatChar(*s,buf);
    s++;
  }
  return buf;
}


bool loadSim(char *fname, StaggeredGrid &grid, SlcSurface &s, double &lambda, double &mu_e, 
						 double &yieldStress, double &flowRate, double &plasticModulus, int &step, double &dt) {
	char sfname[80];
	std::ifstream in(fname, std::ios::out);
	in.getline(sfname, 80);
	s.bdtRead(sfname);
	
	eatStr("lambda:", in);
	in >> lambda;
	eatStr("mu_e:", in);
	in >> mu_e;
	eatStr("yieldStress:", in);
	in >> yieldStress;
	eatStr("flowRate:", in);
	in >> flowRate;
	eatStr("plasticModulus:", in);
	in >> lambda;
	eatStr("step:", in);
	in >> step;
	eatStr("dt:", in);
	in >> dt;
	in >> grid;
	in.close();

	return true;
}

bool saveSim(char *fname, const StaggeredGrid &grid, const SlcSurface &s, double lambda, double mu_e, 
						 double yieldStress, double flowRate, double plasticModulus, int step, double dt) {
	char *sfname = strdup(fname);
	int len = strlen(sfname);
	sfname[len-3] = 'b';
	sfname[len-2] = 'd';
	sfname[len-1] = 't';

	std::ofstream out(fname, std::ios::out);
	out << sfname << "\n";
	out << "lambda: " << lambda << "\n";
	out << "mu_e: " << mu_e << "\n";
	out << "yieldStress: " << yieldStress << "\n";
	out << "flowRate: " << flowRate << "\n";
	out << "plasticModulus: " << plasticModulus << "\n";
	out << "step: " << step << "\n";
	out << "dt: " << dt << "\n";
	out << grid <<"\n";

	out.close();

	s.bdtDump(sfname);

	free (sfname);
	return true;
}

int main(int argc, char* argv[]) {
	timeval startTime, endTime;
	double dt = 0.001;
	double lambda = 0;
	double mu_e = 100;
	double yieldStress = 10;
	double plasticModulus = 0;
	double flowRate = 10;
	StaggeredGrid grid;
	SlcSurface *surface;
	int step = 0;
	int numSteps = 50;

	//PetscInitialize(&argc, &argv, NULL, NULL);

	if (argc > 1) {
		surface = new SlcSurface(new DtTree);
		loadSim(argv[1], grid, *surface, lambda, mu_e, yieldStress, flowRate, plasticModulus, step, dt);
	} else {
		unsigned int nx = 40; //120
		unsigned int ny = 40; //120
		unsigned int nz = 40; //120
		double h = 1.0; // 1/3.0
		SlVector3 lc(0.0,0.0,0.0), uc(h*nx, h*ny, h*nz);
		int band = 6;
		grid.allocate(nx, ny, nz, h, lc, uc, band);
		grid.allocateStrain();
		
		grid.u = 0.0;
		grid.v = 0.0;
		grid.w = -10.0;
		grid.exx = 0.0;
		grid.eyy = 0.0;
		grid.ezz = 0.0;
		grid.exy = 0.0;
		grid.exz = 0.0;
		grid.eyz = 0.0;
		
		{ 
			int max_level = 7;
			double h = pow2(-max_level)*(uc[0]-lc[0]);
			DtTree *t = new DtTree();
			surface = new SlcSurface(t);
			cout<<"buildtree"<<endl;
			buildTree(t, lc-h, uc+h, max_level, initPhi);
			cout<<"contour surface"<<endl;
			surface->contourTree(initPhi);
			cout<<"redistance"<<endl;
			surface->redistance();
		}
	}

	cout<<"starting simulation"<<endl;
	while (step < numSteps) {
		cout<<"==========================================================="<<endl;
		cout<<"Step "<<step<<endl;
		if (step % 50 == 0) {
			char fname[80];
			sprintf(fname, "checkPoint.1%04d.sim", step);
			cout<<"saving \""<<fname<<"\"...";
			cout.flush();
			//saveSim(fname, grid, *surface, lambda, mu_e, yieldStress, flowRate, plasticModulus, step, dt);
			cout<<" done."<<endl;
		}

		{
			char fname[80];
			sprintf(fname, "foo.1%04d.bvt", step);
			cout<<"saving \""<<fname<<"\"...";
			cout.flush();
			surface->bvtDump(fname);
			cout<<" done."<<endl;
			if (step % 30 == 0) {
				char command[80];
				sprintf(command, "./render.perl 10 %i %i &", step, step);
				system(command);
			}
		}

		gettimeofday(&startTime, NULL);
		grid.setLabels(*surface);
		gettimeofday(&endTime, NULL);
		cout<<"setLabels took "<<(endTime.tv_sec-startTime.tv_sec)+
			(endTime.tv_usec-startTime.tv_usec)*1.0e-6<<endl;

		grid.advectStrain(dt);

		gettimeofday(&startTime, NULL);
		grid.advect(dt);
		gettimeofday(&endTime, NULL);
		cout<<"advect took "<<(endTime.tv_sec-startTime.tv_sec)+
			(endTime.tv_usec-startTime.tv_usec)*1.0e-6<<endl;

		grid.updateStrain(dt, yieldStress, flowRate);
		grid.applyElasticForces(dt, mu_e);

		for (unsigned int i=0; i<grid.w.nx(); i++) {
			for (unsigned int j=0; j<grid.w.ny(); j++) {
				for (unsigned int k=0; k<grid.w.nz(); k++) {
					grid.w(i,j,k) -= dt*9.8;
				}
			}
		}

		grid.setBoundaryVelocity();

		gettimeofday(&startTime, NULL);
		//grid.computePressureLiquid();
		gettimeofday(&endTime, NULL);
		cout<<"computePressure took "<<(endTime.tv_sec-startTime.tv_sec)+
			(endTime.tv_usec-startTime.tv_usec)*1.0e-6<<endl;

		gettimeofday(&startTime, NULL);
		grid.computePressureLiquidNoPetsc();
		gettimeofday(&endTime, NULL);
		cout<<"computePressure took "<<(endTime.tv_sec-startTime.tv_sec)+
			(endTime.tv_usec-startTime.tv_usec)*1.0e-6<<endl;


		gettimeofday(&startTime, NULL);
		grid.extrapolate(*surface);
		gettimeofday(&endTime, NULL);
		cout<<"extrapolate took "<<(endTime.tv_sec-startTime.tv_sec)+
			(endTime.tv_usec-startTime.tv_usec)*1.0e-6<<endl;

		gettimeofday(&startTime, NULL);
		updateSurface(&grid, surface, dt);
		gettimeofday(&endTime, NULL);
		cout<<"updateSurface took "<<(endTime.tv_sec-startTime.tv_sec)+
			(endTime.tv_usec-startTime.tv_usec)*1.0e-6<<endl;

		step++;
	}

	//PetscFinalize();

	return 0;
}


