//-------------------------------------------------------------------
//-------------------------------------------------------------------
//
// SLC Surface Tracker
//   -- Header File
//  
// Primary Author: Adam Bargteil (adamb@cs.berkeley.edu)
// 
//-------------------------------------------------------------------
//-------------------------------------------------------------------
//
// Copyright (c) 2003-2005, Regents of the University of California.  All
// rights reserved.
//
// This software is part of the Berkeley Fluid Animation & Simulation
// Toolkit.  The name "Berkeley Fluid Animation & Simulation Toolkit" is
// a trademark of the Regents of the University of California.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are
// met:
//
//   Redistributions of source code must retain the above copyright
//  notice, this list of conditions and the following disclaimer.
//
//  Redistributions in binary form must reproduce the above copyright
//  notice, this list of conditions and the following disclaimer in the
//  documentation and/or other materials provided with the distribution.
//
//  Redistributions in binary form as an executable program, or part of
//  an executable program must, when run for the first time by a given
//  user, prominently display the above copyright notice, this list of
//  conditions and the following disclaimer.
//
//  Neither the name of the University of California, Berkeley nor the
//  names of its contributors may be used to endorse or promote products
//  derived from this software without specific prior written
//  permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
//-------------------------------------------------------------------
//-------------------------------------------------------------------

#ifndef slcSurface_H
#define slcSurface_H

#include <fstream>
#include <iostream>
#include "slVector.H"
#include "slcUtil.H"
#include "distTree.H"

class DtFMMHeap;

class SlcSurface {
public:
	SlcSurface(DtTree *tree);
	SlcSurface();
	~SlcSurface();

	DtTree *tree;
	std::vector<SlTri> triangles; // triangles in polygon mesh
	std::vector<SlVector3> meshPts; // points in polygon mesh
	std::vector<SlVector3> normals; // normals (at points) in polygon mesh
	std::vector<SlVector3> psuedoNormals; // angle weighted normals (at points) in polygon mesh
	std::vector<SlVector3> faceNormals; // face normals in polygon mesh

	bool builtFromTriangles; // was the tree constructed from triangles or a phi func.

	std::vector<std::vector<SlInt2> >edgeList;

	// Marching cubes functions.
	// vertid creates new vertices or finds already created ones.
	// dotet creates triangles for a tet through which the surface passes.
	// contourTree drives the contouring.

	int vertid(double (*evalPhi)(const SlVector3 &x, void *data), int a, int b, SlVector3 x, SlVector3 y, 
						 bool apos, bool bpos, double aphi, double bphi, void *evalPhiData=NULL);
	void dotet(double (*evalPhi)(const SlVector3 &x, void *data), DtCell *cell,  
						 int a, int b, int c, int d, SlVector3 w, SlVector3 x, SlVector3 y, SlVector3 z, void *evalPhiData=NULL);
	void contourTree(double (*evalPhi)(const SlVector3 &x, void *data), void *evalPhiData=NULL);

	// end marching cubes functions

	void fastsweep(); // redistance the tree.
	void redistance(); // redistance the tree.

	void computeNormals();
	void computePsuedoNormals();

	// a variety of signed-distance evaluation routines.  x is the
	// evaluation point.  If interpolate is true, the evaluation is
	// gotten from linear interpolation of the values stored at
	// octree vertices.  s & t are barycentric coordinates and face is the 
	// triangle of the nearest point in the mesh.  

	double eval(const double *x, bool interpolate = false) const;
	double eval(const double *x, double &s, double &t, int &face, bool interpolate = false) const;
	double eval(const SlVector3 &x, bool interpolate = false) const;
	double eval(const SlVector3 &x, double &s, double &t, int &face, bool interpolate = false) const;

	// evaluate the gradient at x.
	SlVector3 grad(const SlVector3 &x);
	SlVector3 grad(const double *x);

	// distance from a point to a triangle.  This searches the entire triangle
	// list of cell.  The second version is more
	// complicated because it is meant to be called for each cell in the concentric
	// triple and has to deal with sharp features, etc.
	double pointToTri(DtCell *c, const SlVector3 &x, double &phi) const;
	double pointToTri(DtCell *cell, const SlVector3 &x, double &phi, int &sgn,
											double &s, double &t, int &face, double &dp, int edge[2]) const;

	// i/o functions
	void read(std::ifstream &in);
	void write(std::ofstream &out) const;
	void objDump(const char *fname) const;
	void ribDumpFaceNormals(const char *fname) const;
	void ribDump(const char *fname) const;
	void bvtDump(const char *fname) const;
	void bvtRead(const char *fname);
	void bdtDump(const char *fname) const;
	void bdtRead(const char *fname);

	void setMaxFunc(double (*maxFunc)(const double *x, void *data), void *data=NULL);
	void setMinFunc(double (*minFunc)(const double *x, void *data), void *data=NULL);
	
	void rebuildTree();

#ifdef INVENTOR_DEFINED
	SoSeparator *ivTree();
	void ivDump(char *fname);
#endif

private:	
	void addToFMMHeap(int u, DtFMMHeap &heap,
										bool *accepted, std::vector<double> &newPhi, int foo);
	double exacteval(DtCell *c, const SlVector3 &x, double &s, double &t, 
									 int &face) const;
	double interpolate(DtCell *c, const SlVector3 &x) const;

	// whenever phi is evaluated, the returned result will be min(minFunc, phi)
	double (*minFunc)(const double *x, void *data);
	void *minFuncData;

	// whenever phi is evaluated, the returned result will be max(maxFunc, phi)
	double (*maxFunc)(const double *x, void *data);
	void *maxFuncData;
  bool deleteTree;
};

#ifdef INVENTOR_DEFINED
#include <Inventor/SoOutput.h>
#include <Inventor/actions/SoWriteAction.h>
int bfastWriteIVFile(const char *filename, SoNode *node);
#endif

// Fast marching method code.
class DtFMMHeapEntry {
 public:
  int u;
  double phi;
  DtFMMHeapEntry(){};
  DtFMMHeapEntry(int u, double phi) {this->u = u, this->phi = phi;};
  DtFMMHeapEntry &operator=(const DtFMMHeapEntry &e){u=e.u,phi=e.phi; return *this;};
};
  
class DtFMMHeap {
 private:
  void swap(int i, int j);
  std::vector<DtFMMHeapEntry> entries;
 public:
  DtFMMHeap();
  DtFMMHeap(int size);
  ~DtFMMHeap();
  void push(int u, double phi);
  DtFMMHeapEntry pop();
  inline unsigned int size() const {return entries.size();};
  void resize(unsigned int num) {entries.resize(num);};
};

////////////////////////////////////////////////////////////////

#endif


//-------------------------------------------------------------------
//-------------------------------------------------------------------
//
// RCS Revision History
//
// $Log: slcSurface.H,v $
// Revision 1.9  2007/07/26 00:19:51  adamwb
// major updates to stl and faster hash_map usage
//
// Revision 1.8  2007/06/28 23:23:12  adamwb
// updates
//
// Revision 1.7  2007/06/16 01:05:26  adamwb
// *** empty log message ***
//
// Revision 1.6  2007/01/22 14:04:17  adamwb
// collision object updates
//
// Revision 1.5  2007/01/12 16:15:02  adamwb
// blowup = true when CG doesn't converge
//
// Revision 1.4  2006/01/15 01:43:52  adamb
// added rebuildTree
//
// Revision 1.3  2005/12/29 01:59:38  adamb
// Added void * to all passed in functions
//
// Revision 1.2  2005/11/08 19:45:06  adamb
// forced checkin
//
// Revision 1.1.1.1  2005/09/06 22:45:12  adamb
// imported sources
//
// Revision 1.3  2005/07/20 01:46:58  adamb
// forced checkin of slc code before changes to hash function
//
// Revision 1.2  2005/06/10 20:08:47  adamb
// Various bug fixes, added iuc and ilc to cell output, rebuild vertMap after readining in a surface, fixed argc checking in the drivers.
//
// Revision 1.1.1.1  2005/05/25 05:41:04  adamb
// Initial Revision
//
//
//-------------------------------------------------------------------
//-------------------------------------------------------------------
