//-------------------------------------------------------------------
//-------------------------------------------------------------------
//
// SLC Surface Tracker
//   -- Code File
//  
// Primary Author: Adam Bargteil (adamb@cs.berkeley.edu)
// 
//-------------------------------------------------------------------
//-------------------------------------------------------------------
//
// Copyright (c) 2003-2006, Regents of the University of California.  All
// rights reserved.
//
// This software is part of the Berkeley Fluid Animation & Simulation
// Toolkit.  The name "Berkeley Fluid Animation & Simulation Toolkit" is
// a trademark of the Regents of the University of California.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are
// met:
//
//   Redistributions of source code must retain the above copyright
//  notice, this list of conditions and the following disclaimer.
//
//  Redistributions in binary form must reproduce the above copyright
//  notice, this list of conditions and the following disclaimer in the
//  documentation and/or other materials provided with the distribution.
//
//  Redistributions in binary form as an executable program, or part of
//  an executable program must, when run for the first time by a given
//  user, prominently display the above copyright notice, this list of
//  conditions and the following disclaimer.
//
//  Neither the name of the University of California, Berkeley nor the
//  names of its contributors may be used to endorse or promote products
//  derived from this software without specific prior written
//  permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
//-------------------------------------------------------------------
//-------------------------------------------------------------------


#include "distTree.H"
#include "float.h"
#include <cstdlib>
#include "slcSurface.H"
#include <float.h>
#include "slIO.H"

extern "C" {
#include "pcube/pcube.h"
}

using namespace std;

#define EPS 1e-15

const double width = 3;

double orient3d(double *pa, double *pb, double *pc, double *pd) {
  double adx, bdx, cdx;
  double ady, bdy, cdy;
  double adz, bdz, cdz;

  adx = pa[0] - pd[0];
  bdx = pb[0] - pd[0];
  cdx = pc[0] - pd[0];
  ady = pa[1] - pd[1];
  bdy = pb[1] - pd[1];
  cdy = pc[1] - pd[1];
  adz = pa[2] - pd[2];
  bdz = pb[2] - pd[2];
  cdz = pc[2] - pd[2];

  return adx * (bdy * cdz - bdz * cdy)
       + bdx * (cdy * adz - cdz * ady)
       + cdx * (ady * bdz - adz * bdy);
}


SlcSurface::SlcSurface(DtTree *tree) {
	this->tree = tree;
	deleteTree = false;
	minFunc = NULL;
	maxFunc = NULL;
	minFuncData = NULL;
	maxFuncData = NULL;
	edgeList.clear();
	builtFromTriangles = false;
};

SlcSurface::SlcSurface() {
  deleteTree = false;
  minFunc = NULL;
  maxFunc = NULL;
  minFuncData = NULL;
  maxFuncData = NULL;
  edgeList.clear();
};

SlcSurface::~SlcSurface() {
  if (deleteTree) delete tree;
  edgeList.clear();
}


double SlcSurface::exacteval(DtCell *c, const SlVector3 &x, double &s, double &t, 
														 int &face) const{
  SlVector3 y,z;
  DtCell *oc = c;
  double h;
  double up,down,left,right,front,back;
  double phi = DBL_MAX;
  double dp = 0.0;
  int edge[2] = {-1,-1};
  int sgn = 0;

  right = tree->pts[c->vertices[7]][0]-x[0];
  left = x[0]-tree->pts[c->vertices[0]][0];
  up = tree->pts[c->vertices[7]][1]-x[1];
  down = x[1]-tree->pts[c->vertices[0]][1];
  back = tree->pts[c->vertices[7]][2]-x[2];
  front = x[2]-tree->pts[c->vertices[0]][2];

  y = (tree->pts[c->vertices[7]]+tree->pts[c->vertices[0]])/2.0;
  
  h = (tree->pts[c->vertices[7]][0]-tree->pts[c->vertices[0]][0]);
  if (c->triangles.size() != 0) 
    pointToTri(c,x,phi,sgn,s,t,face,dp,edge);
  
  if (phi > left) {
		if (oc->neighbors[0] != -1) {
			c = &(tree->cells[oc->neighbors[0]]);
			if (c->triangles.size())
				pointToTri(c,x,phi,sgn,s,t,face,dp,edge);
		}
	}
  
  if (phi > down) {
		if (oc->neighbors[1] != -1) {
			c = &(tree->cells[oc->neighbors[1]]);
			if (c->triangles.size()) 
				pointToTri(c,x,phi,sgn,s,t,face,dp,edge);
		}
  }

  if (phi > front) {
		if (oc->neighbors[2] != -1) {
			c = &(tree->cells[oc->neighbors[2]]);
			if (c->triangles.size()) 
				pointToTri(c,x,phi,sgn,s,t,face,dp,edge);
		}
  }
   
  if (phi > back) {
		if (oc->neighbors[3] != -1) {
			c = &(tree->cells[oc->neighbors[3]]);
			if (c->triangles.size()) 
				pointToTri(c,x,phi,sgn,s,t,face,dp,edge);
		}
  }
   
  if (phi > up) {
		if (oc->neighbors[4] != -1) {
			c = &(tree->cells[oc->neighbors[4]]);
			if (c->triangles.size()) 
				pointToTri(c,x,phi,sgn,s,t,face,dp,edge);
		}
  }
   
  if (phi > right) {
		if (oc->neighbors[5] != -1) {
			c = &(tree->cells[oc->neighbors[5]]);
			if (c->triangles.size())
				pointToTri(c,x,phi,sgn,s,t,face,dp,edge);
		}
  }
    
  if (phi > left && phi > down) {
		if (oc->neighbors[6] != -1) {
			c = &(tree->cells[oc->neighbors[6]]);
			if (c->triangles.size()) 
				pointToTri(c,x,phi,sgn,s,t,face,dp,edge);
		}
  }
  
  if (phi > left && phi > front) {
		if (oc->neighbors[7] != -1) {
			c = &(tree->cells[oc->neighbors[7]]);
			if (c->triangles.size()) 
				pointToTri(c,x,phi,sgn,s,t,face,dp,edge);
		}
  }
  
  if (phi > left && phi > back) {
		if (oc->neighbors[8] != -1) {
			c = &(tree->cells[oc->neighbors[8]]);
			if (c->triangles.size()) 
				pointToTri(c,x,phi,sgn,s,t,face,dp,edge);
		}
  }
  
  if (phi > left && phi > up) {
		if (oc->neighbors[9] != -1) {
			c = &(tree->cells[oc->neighbors[9]]);
			if (c->triangles.size()) 
				pointToTri(c,x,phi,sgn,s,t,face,dp,edge);
		}
  }
  
  if (phi > down && phi > front) {
		if (oc->neighbors[10] != -1) {
			c = &(tree->cells[oc->neighbors[10]]);
			if (c->triangles.size()) 
				pointToTri(c,x,phi,sgn,s,t,face,dp,edge);
		}
  }
  
	if (phi > down && phi > back) {
		if (oc->neighbors[11] != -1) {
			c = &(tree->cells[oc->neighbors[11]]);
			if (c->triangles.size()) 
				pointToTri(c,x,phi,sgn,s,t,face,dp,edge);
		}
  }
  
  if (phi > up && phi > front) {
		if (oc->neighbors[12] != -1) {
			c = &(tree->cells[oc->neighbors[12]]);
			if (c->triangles.size()) 
				pointToTri(c,x,phi,sgn,s,t,face,dp,edge);
		}
  }
	
  if (phi > up && phi > back) {
		if (oc->neighbors[13] != -1) {
			c = &(tree->cells[oc->neighbors[13]]);
			if (c->triangles.size()) 
				pointToTri(c,x,phi,sgn,s,t,face,dp,edge);
		}
  }

  if (phi > right && phi > down) {
		if (oc->neighbors[14] != -1) {
			c = &(tree->cells[oc->neighbors[14]]);
			if (c->triangles.size()) 
				pointToTri(c,x,phi,sgn,s,t,face,dp,edge);
		}
  }
  
  if (phi > right && phi > front) {
		if (oc->neighbors[15] != -1) {
			c = &(tree->cells[oc->neighbors[15]]);
			if (c->triangles.size()) 
				pointToTri(c,x,phi,sgn,s,t,face,dp,edge);
		}
  }
	
  if (phi > right && phi > back) {
		if (oc->neighbors[16] != -1) {
			c = &(tree->cells[oc->neighbors[16]]);
			if (c->triangles.size()) 
				pointToTri(c,x,phi,sgn,s,t,face,dp,edge);
		}
  }
	
  if (phi > right && phi > up) {
		if (oc->neighbors[17] != -1) {
			c = &(tree->cells[oc->neighbors[17]]);
			if (c->triangles.size()) 
				pointToTri(c,x,phi,sgn,s,t,face,dp,edge);
		}
  }
	
  if (phi > left && phi > down && phi > front) {
		if (oc->neighbors[18] != -1) {
			c = &(tree->cells[oc->neighbors[18]]);
			if (c->triangles.size()) 
				pointToTri(c,x,phi,sgn,s,t,face,dp,edge);
		}
  }
  	
  if (phi > left && phi > down && phi > back) {
		if (oc->neighbors[19] != -1) {
			c = &(tree->cells[oc->neighbors[19]]);
			if (c->triangles.size()) 
				pointToTri(c,x,phi,sgn,s,t,face,dp,edge);
		}
  }

  if (phi > left && phi > up && phi > front) {
		if (oc->neighbors[20] != -1) {
			c = &(tree->cells[oc->neighbors[20]]);
			if (c->triangles.size()) 
				pointToTri(c,x,phi,sgn,s,t,face,dp,edge);
		}
	}

  if (phi > left && phi > up && phi > back) {
		if (oc->neighbors[21] != -1) {
			c = &(tree->cells[oc->neighbors[21]]);
			if (c->triangles.size()) 
				pointToTri(c,x,phi,sgn,s,t,face,dp,edge);
		}
  }
	
  if (phi > right && phi > down && phi > front) {
		if (oc->neighbors[22] != -1) {
			c = &(tree->cells[oc->neighbors[22]]);
			if (c->triangles.size()) 
				pointToTri(c,x,phi,sgn,s,t,face,dp,edge);
		}
  }
  
  if (phi > right && phi > down && phi > back) {
		if (oc->neighbors[23] != -1) {
			c = &(tree->cells[oc->neighbors[23]]);
			if (c->triangles.size()) 
				pointToTri(c,x,phi,sgn,s,t,face,dp,edge);
		}
  }
  
  if (phi > right && phi > up && phi > front) {
		if (oc->neighbors[24] != -1) {
			c = &(tree->cells[oc->neighbors[24]]);
			if (c->triangles.size()) 
				pointToTri(c,x,phi,sgn,s,t,face,dp,edge);
		}
	}
  
  if (phi > right && phi > up && phi > back) {
		if (oc->neighbors[25] != -1) {
			c = &(tree->cells[oc->neighbors[25]]);
			if (c->triangles.size()) 
				pointToTri(c,x,phi,sgn,s,t,face,dp,edge);
		}
  }
  
  if (phi!=DBL_MAX) {
    if (oc->triangles.size() == 0) 
      sgn = sign(tree->phi[oc->vertices[0]]);
    else if (phi > h) 
      sgn = sign(interpolate(oc,x));
    return sgn*phi;
  } else 
    return interpolate(oc,x);

}

double SlcSurface::pointToTri(DtCell *c, const SlVector3 &x, double &phi) const {
  phi = DBL_MAX;
  double s = 0.0, t = 0.0;
  int face = -1;
  double dp = 0.0;
  int sgn = 0;
  int edge[2]={-1,-1};
  pointToTri(c,x,phi,sgn,s,t,face,dp,edge);
  if (sgn != 0) phi *= sgn;
  return phi;
}

double SlcSurface::pointToTri(DtCell *cell, const SlVector3 &x, double &phi, int &sgn,
				 double &s, double &t, int &face, double &dp, int edge[2]) const {
  SlVector3 P1, P2, P3, D, E0, E1, X(x), C;
  double ls,lt,a,b,c,d,e,f,det,numer,denom,est;
  int onEdge=0; 
  int point=-1;
	std::vector<int>::const_iterator j;
  
  for (j=cell->triangles.begin(); j!=cell->triangles.end(); j++) {
    onEdge = 0;
    point = -1;
    P1 = meshPts[triangles[*j][0]];
    P2 = meshPts[triangles[*j][1]];
    P3 = meshPts[triangles[*j][2]];
    E0 = P2-P1;
    E1 = P3-P1;
    D=P1-X; 
    a = dot(E0,E0);
    b = dot(E0,E1);
    c = dot(E1,E1); 
    d = dot(E0,D);
    e = dot(E1,D); 
    f = dot(D,D);
    det = a*c-b*b;
    ls = b*e-c*d;
    lt = b*d-a*e;
    
    if (ls+lt <= det) {
      if (ls<0) {
				if (lt<0) {
					if (d<0.0) {
						lt = 0.0;
						if (-d>=a) {ls = 1.0; point=triangles[*j][1];}
						else {ls = -d/a; onEdge=1;}
					} else {
						ls = 0.0;
						if (e >= 0.0) {lt = 0.0; point=triangles[*j][0];}
						else if (-e >= c) {lt = 1.0; point=triangles[*j][2];}
						else {lt = -e/c; onEdge=2;}
					}
				} else {
					ls = 0.0;
					if (e >= 0.0) {lt = 0.0; point=triangles[*j][0];}
					else if (-e >= c) {lt = 1.0; point=triangles[*j][2];}
					else {lt = -e/c; onEdge=2;}
				}
      } else if (lt <0) {
				lt = 0;
				if (d >= 0) {ls = 0; point=triangles[*j][0];}
				else if (-d >= a) {ls = 1; point=triangles[*j][1];}
				else {ls = -d/a; onEdge=1;}
      } else {
				if (det != 0) {ls /= det; lt /= det;}
				else {point=triangles[*j][0];}
      }
    } else {
      double tmp0;
      double tmp1;
      if (ls < 0) {
				tmp0 = b+d;
				tmp1 = c+e;
				if (tmp1 > tmp0) {
					numer = tmp1-tmp0;
					denom = a-2.0*b+c;
					if (numer >= denom) {ls = 1.0; lt = 0.0; point=triangles[*j][1];} 
					else {ls = numer/denom; lt = 1.0-ls; onEdge=3;}
				} else {
					ls = 0;
					if (tmp1 <= 0.0) {lt = 1.0; point=triangles[*j][2];}
					else if (e >= 0) {lt = 0.0; point=triangles[*j][0];}
					else {lt = -e/c; onEdge=2;}
				}
      } else if (lt < 0) {
				tmp0 = b+e;
				tmp1 = a+d;
				if (tmp1 > tmp0) {
					numer = tmp1-tmp0;
					denom = a-2.0*b+c;
					if (numer >= denom) {lt = 1.0; ls = 0.0; point=triangles[*j][2];}
					else {lt = numer/denom; ls = 1.0-lt; onEdge=3;}
				} else {
					lt = 0.0;
					if (tmp1 <= 0) {ls = 1.0; point=triangles[*j][1];}
					else if (d >= 0.0) {ls = 0.0; point=triangles[*j][0];}
					else {ls = -d/a; onEdge=1;}
				}
      } else {
				numer = c+e-b-d;
				if (numer <= 0) {ls=0.0; lt=1.0; point=triangles[*j][2];}
				else {
					denom = a-2*b+c;
					if (numer >= denom) {ls = 1.0; lt = 0.0; point=triangles[*j][1];}
					else {ls = numer/denom; lt = 1.0-ls; onEdge=3;}
				}
      }
    }
    
    C = P1+ls*E0+lt*E1;
    D = C-X;
    est = mag(D);
    
    if (onEdge &&
				(((onEdge == 1) &&
					((edge[0] == triangles[*j][0] && edge[1] == triangles[*j][1]) ||
					 (edge[1] == triangles[*j][0] && edge[0] == triangles[*j][1]))) ||
				 ((onEdge == 2) &&
					((edge[0] == triangles[*j][0] && edge[1] == triangles[*j][2]) ||
					 (edge[1] == triangles[*j][0] && edge[0] == triangles[*j][2]))) ||
				 ((onEdge == 3) &&
					((edge[0] == triangles[*j][1] && edge[1] == triangles[*j][2]) ||
					 (edge[1] == triangles[*j][1] && edge[0] == triangles[*j][2]))))) {
      
      SlVector3 fn = faceNormals[*j];
      normalize(fn);
      double ndp = dot(D,fn);
      if (fabs(ndp) > fabs(dp)) {
				face = (*j);
				s = ls;
				t = lt;
				dp = ndp;
				sgn = -sign(dp);
      } 
    } else if (est < phi) {
      if (onEdge) {
				phi = est;
				face = (*j);
				s = ls;
				t = lt;
				SlVector3 fn = faceNormals[*j];
				normalize(fn);
				dp = dot(D,fn);
				sgn = -sign(dp);
				if (onEdge == 1) {edge[0] = triangles[*j][0]; edge[1] = triangles[*j][1];}
				else if (onEdge == 2) {edge[0] = triangles[*j][0]; edge[1] = triangles[*j][2];}
				else if (onEdge == 3) {edge[0] = triangles[*j][1]; edge[1] = triangles[*j][2];}
      } else if (point != -1) {
				phi = est;
				face = (*j);
				s = ls;
				t = lt;
				sgn = -sign(dot(D, psuedoNormals[point]));
				dp = 0.0;
				edge[0] = edge[1] = -1;
      } else {
				phi = est;
				face = (*j);
				s = ls;
				t = lt;
				double x1[3] = {meshPts[triangles[*j][0]][0], meshPts[triangles[*j][0]][1], meshPts[triangles[*j][0]][2]};
				double x2[3] = {meshPts[triangles[*j][1]][0], meshPts[triangles[*j][1]][1], meshPts[triangles[*j][1]][2]};
				double x3[3] = {meshPts[triangles[*j][2]][0], meshPts[triangles[*j][2]][1], meshPts[triangles[*j][2]][2]};
				sgn =  -sign(orient3d(x1, x2, x3, &(X[0])));
				edge[0] = edge[1] = -1;
				dp = 0;
      }
    }
  }
  if (sgn == 0) return phi;
  return sgn*phi;
}



double SlcSurface::interpolate(DtCell *c, const SlVector3 &x) const {
	SlVector3 y = x-tree->pts[c->vertices[0]];
	SlVector3 z = tree->pts[c->vertices[7]]-tree->pts[c->vertices[0]];
	double w[3] = {y[0]/z[0], y[1]/z[1], y[2]/z[2]};
	return (double)
		(1.0-w[0])*(1.0-w[1])*(1.0-w[2])*(tree->phi[c->vertices[0]])+
		(1.0-w[0])*(1.0-w[1])*(w[2])*(tree->phi[c->vertices[1]])+
		(1.0-w[0])*(w[1])*(1.0-w[2])*(tree->phi[c->vertices[2]])+
		(1.0-w[0])*(w[1])*(w[2])*(tree->phi[c->vertices[3]])+
		(w[0])*(1.0-w[1])*(1.0-w[2])*(tree->phi[c->vertices[4]])+
		(w[0])*(1.0-w[1])*(w[2])*(tree->phi[c->vertices[5]])+
		(w[0])*(w[1])*(1.0-w[2])*(tree->phi[c->vertices[6]])+
		(w[0])*(w[1])*(w[2])*(tree->phi[c->vertices[7]]);
}



double SlcSurface::eval(const SlVector3 &x, bool interpolate) const {
	double y[3] = {x[0], x[1], x[2]}, s = 0.0, t = 0.0;
	int face = -1;
	return eval(y,s,t,face,interpolate);
}

double SlcSurface::eval(const SlVector3 &x, double &s, double &t, 
													 int &face, bool interpolate) const {
	double y[3] = {x[0], x[1], x[2]};
	return eval(y,s,t,face,interpolate);
}

double SlcSurface::eval(const double *x, bool interpolate) const {
	double s = 0.0, t = 0.0;
	int face = -1;
	return eval(x,s,t,face,interpolate);
}

double SlcSurface::eval(const double *x, double &s, double &t, 
													 int &face, bool interpolate) const {
	double ret, tmp;
	DtCell *c = tree->findCell(x);
	if (c->level == tree->max_level && !interpolate) ret = exacteval(c,x,s,t,face);
	else ret = this->interpolate(c,x);
	if (minFunc && (tmp = minFunc(x, minFuncData)) < ret) ret = tmp;
	if (maxFunc && (tmp = maxFunc(x, maxFuncData)) > ret) ret = tmp;
	return ret;
}


SlVector3 SlcSurface::grad(const SlVector3 &x) {
	SlVector3 y(x);
	return grad(&y[0]);
}

SlVector3 SlcSurface::grad(const double *x) {
	DtCell *c = tree->findCell(x);
	SlVector3 g;

	SlVector3 y = x-tree->pts[c->vertices[0]];
	SlVector3 z = tree->pts[c->vertices[7]]-tree->pts[c->vertices[0]];
	double w[3] = {y[0]/z[0], y[1]/z[1], y[2]/z[2]};
	
	g[0] = (1.0-w[1])*(1.0-w[2])*(tree->phi[c->vertices[4]]-tree->phi[c->vertices[0]])+
		(1.0-w[1])*(w[2])*(tree->phi[c->vertices[5]]-tree->phi[c->vertices[1]])+
		(w[1])*(1.0-w[2])*(tree->phi[c->vertices[6]]-tree->phi[c->vertices[2]])+
		(w[1])*(w[2])*(tree->phi[c->vertices[7]]-tree->phi[c->vertices[3]]);
	g[1] = (1.0-w[0])*(1.0-w[2])*(tree->phi[c->vertices[2]]-tree->phi[c->vertices[0]])+
		(1.0-w[0])*(w[2])*(tree->phi[c->vertices[3]]-tree->phi[c->vertices[1]])+
		(w[0])*(1.0-w[2])*(tree->phi[c->vertices[6]]-tree->phi[c->vertices[4]])+
		(w[0])*(w[2])*(tree->phi[c->vertices[7]]-tree->phi[c->vertices[5]]);
	g[2] = (1.0-w[0])*(1.0-w[1])*(tree->phi[c->vertices[1]]-tree->phi[c->vertices[0]])+
		(1.0-w[0])*(w[1])*(tree->phi[c->vertices[3]]-tree->phi[c->vertices[2]])+
		(w[0])*(1.0-w[1])*(tree->phi[c->vertices[5]]-tree->phi[c->vertices[4]])+
		(w[0])*(w[1])*(tree->phi[c->vertices[7]]-tree->phi[c->vertices[6]]);
	g/=(z[0]);
	return g;
}


istream &operator>>(std::istream &strm, SlcSurface &s) {
  ios::fmtflags orgFlags = strm.setf(ios::skipws);
  eatStr("[",strm);
  strm >> *(s.tree);
  eatChar(',',strm);
  strm >> s.meshPts;
  eatChar(',',strm);
  strm >> s.triangles;
  eatChar(']',strm);
  strm.flags(orgFlags);
  return strm;
}

ostream &operator<<(ostream &strm,const SlcSurface &s) {
  strm << "[";
  strm << *(s.tree);
  strm << ",";
  strm << s.meshPts;
  strm << ",";
  strm << s.triangles;
  strm << "]";
  return strm;
}

void SlcSurface::read(std::ifstream &in) {
	int version;
	in.read((char*)&version, sizeof(int));
	if (in.fail()) {
		cerr<<"can't read from file"<<endl;
		exit(-1);
	}
	if (version == 2) {
		tree->read(in);
		slc::read(in, meshPts);
		slc::read(in, triangles);
	} else {		
		cerr<<"incompatible bdt file version="<<version<<endl;
	}
	computeNormals();
	computePsuedoNormals();
}

void SlcSurface::write(std::ofstream &out) const {
	int version = 2;
	out.write((char*)&version, sizeof(int));
  tree->write(out);
	slc::write(out, meshPts);
	slc::write(out, triangles);
}

void SlcSurface::bdtDump(const char *fname) const {
	ofstream out(fname, ios::out | ios::binary);
	write(out);
}

void SlcSurface::bdtRead(const char *fname) {
	ifstream in(fname, ios::in | ios::binary);
	read(in);
}

void SlcSurface::objDump(const char *fname) const {
	ofstream out(fname, ios::out);
	std::vector<SlVector3>::const_iterator p;
	std::vector<SlTri>::const_iterator t;
	for (p=meshPts.begin(); p!=meshPts.end(); p++)
		out<<"v "<<(*p)[0]<<" "<<(*p)[1]<<" "<<(*p)[2]<<endl;
	for (t=triangles.begin(); t!=triangles.end(); t++)
		out<<"f "<<(*t)[0]+1<<" "<<(*t)[1]+1<<" "<<(*t)[2]+1<<endl;
}

void SlcSurface::bvtDump(const char *fname) const {	
	ofstream fout (fname, ios::out | ios::binary);
	int version = 1;
	fout.write((char*)&version, sizeof(int));
	slc::write(fout, meshPts);
	slc::write(fout, triangles);
}

void SlcSurface::bvtRead(const char *fname) {
	int version;
	ifstream fin (fname, ios::in | ios::binary);
	fin.read((char*)&version, sizeof(int));
	if (version == 1) {
		slc::read(fin, meshPts);
		slc::read(fin, triangles);
	} else {
		cerr << "Incompatible bvt file: "<<fname<<endl;
		exit(-1);
	}
}

void SlcSurface::ribDumpFaceNormals(const char *fname) const {
	ofstream out(fname, ios::out);
	unsigned int i;
	std::vector<SlVector3>::const_iterator p,n;
	std::vector<SlTri>::const_iterator t;

	out<<"PointsPolygons [";
	for (i=0; i<triangles.size(); i++)
		out<<"3 ";
	out<<"] [";
	for (i=0; i<triangles.size(); i++)
		out<<3*i<<" "<<3*i+1<<" "<<3*i+2<<" ";		
	out<<"]\n";

	out<<"\"P\" [";
	for (t=triangles.begin(); t!=triangles.end(); t++) {
		out<<meshPts[(*t)[0]][0]<<" "<<meshPts[(*t)[0]][1]<<" "<<meshPts[(*t)[0]][2]<<" ";
		out<<meshPts[(*t)[1]][0]<<" "<<meshPts[(*t)[1]][1]<<" "<<meshPts[(*t)[1]][2]<<" ";
		out<<meshPts[(*t)[2]][0]<<" "<<meshPts[(*t)[2]][1]<<" "<<meshPts[(*t)[2]][2]<<" ";
	}
	out<<"]\n\"N\" [";
	for (i=0; i<triangles.size(); i++) {
		out<<faceNormals[i][0]<<" "<<faceNormals[i][1]<<" "<<faceNormals[i][2]<<" ";
		out<<faceNormals[i][0]<<" "<<faceNormals[i][1]<<" "<<faceNormals[i][2]<<" ";
		out<<faceNormals[i][0]<<" "<<faceNormals[i][1]<<" "<<faceNormals[i][2]<<" ";
	}
	out<<"]\n\n";
}

void SlcSurface::ribDump(const char *fname) const {
	ofstream out(fname, ios::out);
	unsigned int i;
	std::vector<SlVector3>::const_iterator p,n;
	std::vector<SlTri>::const_iterator t;

	out<<"PointsPolygons [";
	for (i=0; i<triangles.size(); i++)
		out<<"3 ";
	out<<"] [";
	for (t=triangles.begin(); t!=triangles.end(); t++)
		out<<(*t)[0]<<" "<<(*t)[1]<<" "<<(*t)[2]<<" ";
	out<<"]\n";

	out<<"\"P\" [";
	for (p=meshPts.begin(); p!=meshPts.end(); p++)
		out<<(*p)[0]<<" "<<(*p)[1]<<" "<<(*p)[2]<<" ";
	out<<"]\n\"N\" [";
	for (n=normals.begin(); n!=normals.end(); n++)
		out<<(*n)[0]<<" "<<(*n)[1]<<" "<<(*n)[2]<<" ";
	out<<"]\n\n";
}

#ifdef INVENTOR_DEFINED
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoMaterialBinding.h>
#include <Inventor/nodes/SoNormal.h>
#include <Inventor/nodes/SoNormalBinding.h>
SoSeparator *SlcSurface::ivTree() {
	std::vector<SlVector3>::const_iterator n;
	int i;
	SoSeparator *mesh = new SoSeparator();
	mesh->ref();
	
	SoNormal *sonormal = new SoNormal();
	mesh->addChild(sonormal);
	SoNormalBinding *normalBinding = new SoNormalBinding();
	mesh->addChild(normalBinding);
	normalBinding->value = SoNormalBinding::PER_VERTEX_INDEXED;
	
	for(n=normals.begin(),i=0; n!=normals.end(); n++, i++) 
		sonormal->vector.set1Value(i,(*n)[0],(*n)[1],(*n)[2]);

	mesh->addChild(slc::coord(meshPts));
	mesh->addChild(slc::faceSet(triangles));
	mesh->unrefNoDelete();
	return mesh;
}

void SlcSurface::ivDump(char *fname) const {
	SoSeparator *iv = ivTree();
	iv->ref();
	slcWriteIVFile(fname, iv);
	iv->unref();
}

#endif

void SlcSurface::contourTree(double (*evalPhi)(const SlVector3 &x, void *data), void *evalPhiData) {
	int i;
	std::vector<int> cubes;
	std::vector<DtCell>::const_iterator ci;
	std::vector<int>::const_iterator ii;

	// clear out any old data 
	triangles.clear();
	meshPts.clear();
	normals.clear();
	psuedoNormals.clear();
	faceNormals.clear();
	builtFromTriangles = false;
	edgeList.clear();
	edgeList.resize(tree->pts.size());

	// create list of cubes to look at
	for (ci=tree->cells.begin(),i=0; ci!=tree->cells.end(); i++, ci++) {
		if (ci->level != tree->max_level) continue;
		int foo = sign(tree->phi[ci->vertices[0]]);
		if ((foo != sign(tree->phi[ci->vertices[1]])) ||
				(foo != sign(tree->phi[ci->vertices[2]])) ||
				(foo != sign(tree->phi[ci->vertices[3]])) ||
				(foo != sign(tree->phi[ci->vertices[4]])) ||
				(foo != sign(tree->phi[ci->vertices[5]])) ||
				(foo != sign(tree->phi[ci->vertices[6]])) ||
				(foo != sign(tree->phi[ci->vertices[7]]))) {
			cubes.push_back(i);
		}
	}

	for (ii=cubes.begin(); ii!=cubes.end(); ii++) {
		DtCell *c = &(tree->cells[*ii]);
		SlVector3 &x0 = tree->pts[c->vertices[0]];
		SlVector3 &x1 = tree->pts[c->vertices[1]];
		SlVector3 &x2 = tree->pts[c->vertices[2]];
		SlVector3 &x3 = tree->pts[c->vertices[3]];
		SlVector3 &x4 = tree->pts[c->vertices[4]];
		SlVector3 &x5 = tree->pts[c->vertices[5]];
		SlVector3 &x6 = tree->pts[c->vertices[6]];
		SlVector3 &x7 = tree->pts[c->vertices[7]];

		dotet(evalPhi,c,c->vertices[0],c->vertices[2],c->vertices[4],c->vertices[1],
					x0,x2,x4,x1, evalPhiData);
		dotet(evalPhi,c,c->vertices[6],c->vertices[2],c->vertices[1],c->vertices[4],
					x6,x2,x1,x4, evalPhiData);
		dotet(evalPhi,c,c->vertices[6],c->vertices[2],c->vertices[3],c->vertices[1],
					x6,x2,x3,x1, evalPhiData);
		dotet(evalPhi,c,c->vertices[6],c->vertices[4],c->vertices[1],c->vertices[5],
					x6,x4,x1,x5, evalPhiData);
		dotet(evalPhi,c,c->vertices[6],c->vertices[1],c->vertices[3],c->vertices[5],
					x6,x1,x3,x5, evalPhiData);
		dotet(evalPhi,c,c->vertices[6],c->vertices[3],c->vertices[7],c->vertices[5],
					x6,x3,x7,x5, evalPhiData);
	}

	edgeList.clear();

	triangles.reserve(triangles.size());
	meshPts.resize(meshPts.size());
	computeNormals();
	computePsuedoNormals();
}

int SlcSurface::vertid(double (*evalPhi)(const SlVector3 &x, void *data), int a, int b, SlVector3 x, 
											 SlVector3 y, bool apos, bool bpos, double aphi, double bphi, void *evalPhiData) {
	double TOL = 1e-10;
	int MAXITER = 100;
	double p,sum;
	bool cpos;
	int iter = 0;
	//unsigned long long ei;
	int vi;
	SlVector3 z,v,zprime;

	SlVector3 ox(x), oy(y);
	double odist = sqrMag(y-x);

	if (a > b) { 
		int ti;
		bool tb;
		double td;
		SlVector3 tv;
		ti = a; a = b; b = ti; 
		tv = x; x = y; y = tv;
		tb = apos; apos = bpos; bpos = tb;
		td = aphi; aphi = bphi; bphi = td;
	}
	for (std::vector<SlInt2>::iterator i=edgeList[a].begin(); i!=edgeList[a].end(); i++) {
		if ((*i)[0] == b) return (*i)[1];
	}

	sum = fabs(aphi-bphi);
	if (sum < TOL) z = 0.5*(x+y);
	else z = (fabs(bphi)/sum)*x+(fabs(aphi)/sum)*y;
	while (fabs(p=evalPhi(z,evalPhiData)) > TOL && iter++ < MAXITER) {
		if ((cpos = (p > 0.0)) == apos) {
			x = z;
			aphi = p;
		} else {
			y = z;
			bphi = p;
		}
		sum = fabs(aphi-bphi);
		z = (fabs(bphi)/sum)*x+(fabs(aphi)/sum)*y;
	}

	if (sqrMag(ox-z) < 0.000001*odist)
		z = 0.999*ox+0.001*oy;

	if (sqrMag(oy-z) < 0.000001*odist)
		z = 0.001*ox+0.999*oy;

	vi = meshPts.size();
	edgeList[a].push_back(SlInt2(b,vi));
	meshPts.push_back(z);

	return vi;
}

void SlcSurface::dotet(double (*evalPhi)(const SlVector3 &x, void *data), DtCell *cell, 
											 int a, int b, int c, int d,
											 SlVector3 w, SlVector3 x, SlVector3 y, SlVector3 z, void *evalPhiData) {
	bool apos, bpos, cpos, dpos;
	double aphi, bphi, cphi, dphi;
	int index(0), e1(0), e2(0), e3(0), e4(0), e5(0), e6(0);
	unsigned int l = triangles.size();
	SlVector3 foo;
	if ((apos = ((aphi=tree->phi[a]) > 0.0))) index += 8;
	if ((bpos = ((bphi=tree->phi[b]) > 0.0))) index += 4;
	if ((cpos = ((cphi=tree->phi[c]) > 0.0))) index += 2;
	if ((dpos = ((dphi=tree->phi[d]) > 0.0))) index += 1;

	if (apos != bpos) e1 = vertid(evalPhi,a,b,w,x,apos,bpos,aphi,bphi,evalPhiData);
	if (apos != cpos) e2 = vertid(evalPhi,a,c,w,y,apos,cpos,aphi,cphi,evalPhiData);
	if (apos != dpos) e3 = vertid(evalPhi,a,d,w,z,apos,dpos,aphi,dphi,evalPhiData);
	if (bpos != cpos) e4 = vertid(evalPhi,b,c,x,y,bpos,cpos,bphi,cphi,evalPhiData);
	if (bpos != dpos) e5 = vertid(evalPhi,b,d,x,z,bpos,dpos,bphi,dphi,evalPhiData);
	if (cpos != dpos) e6 = vertid(evalPhi,c,d,y,z,cpos,dpos,cphi,dphi,evalPhiData);
	
	switch (index) {
	case 1: 
		triangles.push_back(SlTri(e5,e3,e6)); 
		cell->triangles.push_back(l);
		break;
	case 2: 
		triangles.push_back(SlTri(e2,e4,e6)); 
		cell->triangles.push_back(l);
		break;
	case 3: 
		triangles.push_back(SlTri(e3,e4,e5));
		triangles.push_back(SlTri(e3,e2,e4));
		cell->triangles.push_back(l);
		cell->triangles.push_back(l+1);
		break;
	case 4: 
		triangles.push_back(SlTri(e1,e5,e4)); 
		cell->triangles.push_back(l);
		break;
	case 5: 
		triangles.push_back(SlTri(e3,e4,e1));
		triangles.push_back(SlTri(e3,e6,e4));
		cell->triangles.push_back(l);
		cell->triangles.push_back(l+1);
		break;
	case 6: 
		triangles.push_back(SlTri(e1,e6,e2));
		triangles.push_back(SlTri(e1,e5,e6));
		cell->triangles.push_back(l);
		cell->triangles.push_back(l+1);
		break;
	case 7: 
		triangles.push_back(SlTri(e1,e3,e2)); 
		cell->triangles.push_back(l);
		break;
	case 8: 
		triangles.push_back(SlTri(e1,e2,e3)); 
		cell->triangles.push_back(l);
		break;
	case 9: 
		triangles.push_back(SlTri(e1,e6,e5));
		triangles.push_back(SlTri(e1,e2,e6));
		cell->triangles.push_back(l);
		cell->triangles.push_back(l+1);
		break;
	case 10: 
		triangles.push_back(SlTri(e1,e6,e3));
		triangles.push_back(SlTri(e1,e4,e6));
		cell->triangles.push_back(l);
		cell->triangles.push_back(l+1);
		break;
	case 11: 
		triangles.push_back(SlTri(e1,e4,e5)); 
		cell->triangles.push_back(l);
		break;
	case 12: 
		triangles.push_back(SlTri(e3,e4,e2));
		triangles.push_back(SlTri(e3,e5,e4));
		cell->triangles.push_back(l);
		cell->triangles.push_back(l+1);
		break;
	case 13: 
		triangles.push_back(SlTri(e6,e4,e2)); 
		cell->triangles.push_back(l);
		break;
	case 14: 
		triangles.push_back(SlTri(e5,e6,e3)); 
		cell->triangles.push_back(l);
		break;
	}
}

void SlcSurface::computeNormals() {
	faceNormals.resize(triangles.size());
	normals.resize(meshPts.size()); 
	std::vector<SlTri>::const_iterator t;
	std::vector<SlVector3>::iterator n;
	for (n=normals.begin(); n!=normals.end(); n++)
		(*n) = 0.0;

	for (n=faceNormals.begin(),t=triangles.begin(); n!=faceNormals.end(); n++,t++) {
		(*n) = cross(meshPts[(*t)[1]]-meshPts[(*t)[0]], meshPts[(*t)[2]]-meshPts[(*t)[0]]);
		normals[(*t)[0]]+=(*n);
		normals[(*t)[1]]+=(*n);
		normals[(*t)[2]]+=(*n);
	}

	for (n=normals.begin(); n!=normals.end(); n++)
		normalize(*n);
}

void SlcSurface::computePsuedoNormals() {
	std::vector<SlVector3>::iterator n;
	std::vector<SlTri>::const_iterator t;
	if (psuedoNormals.size()) return;
	SlVector3 V1, V2, V3, fn;
	double tmp;
	psuedoNormals.resize(meshPts.size());
	for (n=psuedoNormals.begin(); n!=psuedoNormals.end(); n++) {
		(*n) = 0.0;
	}
	for (n=faceNormals.begin(),t=triangles.begin(); n!=faceNormals.end(); n++,t++) {
		fn = (*n);
		V1 = meshPts[(*t)[1]]-meshPts[(*t)[0]];
		V2 = meshPts[(*t)[2]]-meshPts[(*t)[1]];
		V3 = meshPts[(*t)[0]]-meshPts[(*t)[2]];
		normalize(fn);
		normalize(V1);
		normalize(V2);
		normalize(V3);

		tmp = dot(V1,-V3);
		if (tmp < 1.0)
			psuedoNormals[(*t)[0]]+=acos(tmp)*fn;
		tmp = dot(V2,-V1);
		if (tmp < 1.0)
			psuedoNormals[(*t)[1]]+=acos(tmp)*fn;
		tmp = dot(V3,-V2);
		if (tmp < 1.0)
			psuedoNormals[(*t)[2]]+=acos(tmp)*fn;
	}

	for (n=psuedoNormals.begin(); n!=psuedoNormals.end(); n++) {
		normalize(*n);
	}
}

#ifdef INVENTOR_DEFINED
int slcWriteIVFile(const char *filename, SoNode *node) {
  // Open the output file
  SoOutput mySceneOutput;
  if (strcmp(filename,"-")==0) {
    mySceneOutput.setFilePointer(stdout);
  } else if (strcmp(filename,"--")==0) {
    mySceneOutput.setFilePointer(stderr);
  } else if (!mySceneOutput.openFile(filename)) {
    return false;
  }

	mySceneOutput.setBinary(true);

  // Read the whole file into the database
  SoWriteAction myWriter(&mySceneOutput);
  myWriter.apply(node);

  if ((strcmp(filename,"-")!=0) && (strcmp(filename,"--")!=0)) {
    mySceneOutput.closeFile();
  }

  return true;
}
#endif

///////////////////////////////////////////////////////////////////
// Code for fast marching method
///////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////
// DtFMMHeap (used for fast marching method)
////////////////////////////////////////////////////////////////////////
DtFMMHeap::DtFMMHeap() {
  entries.clear();
}

DtFMMHeap::DtFMMHeap(int size) {
  entries.clear();
	entries.reserve(size);
}

DtFMMHeap::~DtFMMHeap() {
	entries.clear();
}

void DtFMMHeap::swap(int i, int j) {
  DtFMMHeapEntry tmp = entries[i];
  entries[i] = entries[j];
  entries[j] = tmp;
}

void DtFMMHeap::push(int u, double phi) {
	int i = entries.size();
	entries.push_back(DtFMMHeapEntry(u,phi));
  while(i>0 && fabs(phi)<fabs(entries[(i-1)/2].phi)) {
    swap(i,(i-1)/2);
    i=(i-1)/2;
  }
}

DtFMMHeapEntry DtFMMHeap::pop(){
  DtFMMHeapEntry ret = entries.front();
  DtFMMHeapEntry e = entries.back();
  swap(0,entries.size()-1);
  entries.pop_back();
  unsigned int i=0;
  while (i*2+1 < size()) {
    if (fabs(entries[i*2+1].phi) < fabs(e.phi))
      if (i*2+2 < size() && fabs(entries[i*2+2].phi) < fabs(e.phi))
				if (fabs(entries[i*2+1].phi) < fabs(entries[i*2+2].phi)) {
					swap(i,i*2+1);
					i = i*2+1;
				} else {
					swap(i,i*2+2);
					i = i*2+2;
				}
      else {
				swap(i,i*2+1);
				i = i*2+1;
      }
    else if (i*2+2 < size() && fabs(entries[i*2+2].phi) < fabs(e.phi)) {
      swap(i,i*2+2);
      i=i*2+2;
    } else break;
  }
  return ret;
}

void SlcSurface::addToFMMHeap(int u, DtFMMHeap &heap, 
															bool *accepted, std::vector<double> &newPhi,
															int foo) {
  double p, phi1, phi2, phi3, q, a(0.0), b(0.0), c(0.0), x1, x2;
	int v;
  bool d1, d2, d3;
  phi1=phi2=phi3=DBL_MAX;
  d1=d2=d3=false;
	double h; 
	if (tree->neighbors[u][0] == -1) h = tree->pts[u][0]-tree->pts[tree->neighbors[u][1]][0];
	else if (tree->neighbors[u][1] == -1) h = tree->pts[tree->neighbors[u][0]][0] - tree->pts[u][0];
	else h = std::min(tree->pts[tree->neighbors[u][0]][0] - tree->pts[u][0], 
										tree->pts[u][0]-tree->pts[tree->neighbors[u][1]][0]);
	double h2 = h*h;

	SlInt6 &neighbors = tree->neighbors[u];

	v = neighbors[0];
  if ((v!=-1) && accepted[v]) {
    phi1 = newPhi[v];
    d1 = true;
  }
  
	v = neighbors[1];
  if ((v!=-1) && accepted[v]) {
    phi1 = absmin2(phi1,newPhi[v]);
    d1 = true;
  }

	v = neighbors[2];
  if ((v!=-1) && accepted[v]) {
    phi2 = newPhi[v];
    d2 = true;
  } 

	v = neighbors[3];
  if ((v!=-1) && accepted[v]) {
    phi2 = absmin2(phi2,newPhi[v]);
    d2 = true;
  }

	v = neighbors[4];
  if ((v!=-1) && accepted[v]) {
    phi3 = newPhi[v];
    d3 = true;
  } 
  
	v = neighbors[5];
  if ((v!=-1) && accepted[v]) {
    phi3 = absmin2(phi3,newPhi[v]);
    d3 = true;
  }

  int done = 0;
  if (d1 && d2 && d3) {
    if ((((phi1-phi2)/h)*((phi1-phi2)/h)+((phi1-phi3)/h)*((phi1-phi3)/h) > 1.0) ||
				(((phi2-phi1)/h)*((phi2-phi1)/h)+((phi2-phi3)/h)*((phi2-phi3)/h) > 1.0) ||
				(((phi3-phi1)/h)*((phi3-phi1)/h)+((phi3-phi2)/h)*((phi3-phi2)/h) > 1.0)) {
      if (fabs(phi1) > fabs(phi2) && fabs(phi1) > fabs(phi3))
				d1 = 0;
      else if (phi2 > phi3)
				d2 = 0;
      else
				d3 = 0;
    } else {
      a = 1.0/(h2)+1.0/(h2)+1.0/(h2);
      b = -2.0*(phi1/h2+phi2/h2+phi3/h2);
      c = (phi1*phi1/h2)+(phi2*phi2/h2)+(phi3*phi3/h2)-1.0;
      done = 1;
    }
  }
  if (!done && d1 && d2) {
    if ((((phi1-phi2)/h)*((phi1-phi2)/h) > 1.0) ||
				(((phi2-phi1)/h)*((phi2-phi1)/h) > 1.0)) {
      if (fabs(phi1) > fabs(phi2))
				d1 = 0;
      else 
				d2 = 0;
    } else {
      a = 1.0/(h2)+1.0/(h2);
      b = -2.0*(phi1/h2+phi2/h2);
      c = (phi1*phi1/h2)+(phi2*phi2/h2)-1.0;
      done = 1;
    }
  }
  if (!done && d1 && d3) {
    if ((((phi1-phi3)/h)*((phi1-phi3)/h) > 1.0) ||
				(((phi3-phi1)/h)*((phi3-phi1)/h) > 1.0)) {
      if (fabs(phi1) > fabs(phi3))
				d1 = 0;
      else 
				d3 = 0;
    } else {
      a = 1.0/(h2)+1.0/(h2);
      b = -2.0*(phi1/h2+phi3/h2);
      c = (phi1*phi1/h2)+(phi3*phi3/h2)-1.0;
      done = 1;
    }
  }
  if (!done && d2 && d3) {
    if ((((phi2-phi3)/h)*((phi2-phi3)/h) > 1.0) ||
				(((phi3-phi2)/h)*((phi3-phi2)/h) > 1.0)) {
      if (fabs(phi2) > fabs(phi3))
				d2 = 0;
      else 
				d3 = 0;
    } else {
      a = 1.0/(h2)+1.0/(h2);
      b = -2.0*(phi2/h2+phi3/h2);
      c = (phi2*phi2/h2)+(phi3*phi3/h2)-1.0;
      done = 1;
    }
  }
  if (!done && d1) {
    a = 1.0/(h2);
    b = -2.0*(phi1/h2);
    c = (phi1*phi1/h2)-1.0;
    done = 1;
  }
  if (!done && d2) {
    a = 1.0/(h2);
    b = -2.0*(phi2/h2);
    c = (phi2*phi2/h2)-1.0;
    done = 1;
  }
  if (!done && d3) {
    a = 1.0/(h2);
    b = -2.0*(phi3/h2);
    c = (phi3*phi3/h2)-1.0;
    done = 1;
  }

  if (b >= 0) 
    q = -0.5*(b+sqrt(b*b-4*a*c));
  else
    q = 0.5*(-b+sqrt(b*b-4*a*c));
  x1 = q/a;
  x2 = c/q;

  double oldPhi = tree->phi[u];
	if (oldPhi == 0.0) 	p = absmax2(x1,x2);
	else {
		if (sign(oldPhi) != sign(x1))
			if (sign(oldPhi) != sign(x2)) 
				p = absmax2(x1,x2);
			else p = x2;
		else if (sign(oldPhi) != sign(x2))
			p = x1;
		else
			p = absmax2(x1,x2);
	}
	p = absmax2(x1,x2);

	if (fabs(p) < fabs(newPhi[u])) {
		newPhi[u] = p;
		heap.push(u, p);
	}
}

void SlcSurface::redistance() {
	int i,j,u,v;
	SlInt3 x,y;
  DtFMMHeap heap(tree->pts.size());
  DtFMMHeapEntry e;
	
	std::vector<double> newPhi;
	tree->phi.resize(tree->pts.size());
	newPhi.resize(tree->phi.size());
    
	DtCell *c;
	double d;
	std::vector<double>::iterator p;
	bool *acc, *accepted = new bool[tree->pts.size()];

	for (p=newPhi.begin(), acc=accepted; p!=newPhi.end(); p++, acc++) {
		*p = DBL_MAX;
		*acc = false;
	}
	
	for (i=tree->cells.size()-1; i >= 0; i--) {
		c = &(tree->cells[i]);
		if (c->child[0] != -1 && !c->triangles.size()) {
			for (j=0; j<8; j++) {
				unsigned int k = c->triangles.size();
				c->triangles.resize(k+tree->cells[c->child[j]].triangles.size());
				for (unsigned int l=0; k<c->triangles.size(); k++,l++) {
					c->triangles[k] = tree->cells[c->child[j]].triangles[l];
				}
			}
		}
		if (c->level != tree->max_level) continue;
 		if (c->triangles.size()) {
			double h = tree->pts[c->vertices[7]][0]-tree->pts[c->vertices[0]][0];
			for (j=0; j<8; j++) {
				v = c->vertices[j];
				pointToTri(c,&(tree->pts[v][0]),d);
				if (fabs(d) > h) continue;
				if (fabs(d) < fabs(newPhi[v]) && d!=DBL_MAX) {
					newPhi[v] = d;
					accepted[v] = true;
				}
			}
		}
	}

  bool doDelete = true;
  bool done = false;
  int iter = 0;

	std::vector<std::vector<int> > addLists;

  while (!done && iter < 2) {
		addLists.resize(tree->pts.size());
		for (i=0; i<(int)tree->ipts.size(); i++) {
			addLists[i].clear();
		}
		for (i=0; i<(int)tree->ipts.size(); i++) {
			for (j=0; j<6; j++) {
				if (tree->neighbors[i][j] != -1) {
					addLists[tree->neighbors[i][j]].push_back(i);
				}
			}
		}

		bool *added = new bool[tree->pts.size()];
		for (i=0; i<(int)tree->pts.size(); i++) added[i] = false;
		for (i=0; i<(int)tree->pts.size(); i++) {
			if (accepted[i]) {
				for (j=0; j<(int)addLists[i].size(); j++) {
					unsigned int k = addLists[i][j];
					if (!accepted[k] && added[k] == false) {
						added[k] = true;
						addToFMMHeap(k, heap, accepted, newPhi,1);
					}
				}
			}
		}
		delete [] added;

    while (heap.size() > 0) {
      e = heap.pop();
      u = e.u;
			if (accepted[u] && fabs(newPhi[u]) < fabs(e.phi)) continue;
      newPhi[u] = e.phi;
      accepted[u] = true;

      for (unsigned int i=0; i<addLists[u].size(); i++) {
        int v = addLists[u][i];
        if (!accepted[v]) addToFMMHeap(v, heap, accepted, newPhi,0);
      }
    }

    doDelete = false;

    acc = accepted;
    for (i=0; i<(int)tree->phi.size(); i++, acc++) if (!(*acc)) doDelete = true;
    if (doDelete) {
      bool *deleteCell = new bool[tree->cells.size()];

      acc = deleteCell;
      for (i=0; i<(int)tree->cells.size(); i++, acc++) *acc = false;
      for (i=tree->cells.size()-1; i >= 0; i--) {
        c = &(tree->cells[i]);
        doDelete = true;
        if (c->child[0] != -1) {
          for (j=0; j<8; j++) if (!deleteCell[c->child[j]]) doDelete = false;
          if (doDelete) {
						tree->cells[c->child[0]].parent = -1;
						tree->cells[c->child[1]].parent = -1;
						tree->cells[c->child[2]].parent = -1;
						tree->cells[c->child[3]].parent = -1;
						tree->cells[c->child[4]].parent = -1;
						tree->cells[c->child[5]].parent = -1;
						tree->cells[c->child[6]].parent = -1;
						tree->cells[c->child[7]].parent = -1;
            c->child[0] = c->child[1] = c->child[2] = c->child[3] =
              c->child[4] = c->child[5] = c->child[6] = c->child[7] = -1;
						tree->neighbors[c->vertices[0]][0] = c->vertices[4];
						tree->neighbors[c->vertices[0]][2] = c->vertices[2];
						tree->neighbors[c->vertices[0]][4] = c->vertices[1];

						tree->neighbors[c->vertices[1]][0] = c->vertices[5];
						tree->neighbors[c->vertices[1]][2] = c->vertices[3];
						tree->neighbors[c->vertices[1]][5] = c->vertices[0];
						
						tree->neighbors[c->vertices[2]][0] = c->vertices[6];
						tree->neighbors[c->vertices[2]][3] = c->vertices[0];
						tree->neighbors[c->vertices[2]][4] = c->vertices[3];
						
						tree->neighbors[c->vertices[3]][0] = c->vertices[7];
						tree->neighbors[c->vertices[3]][3] = c->vertices[1];
						tree->neighbors[c->vertices[3]][5] = c->vertices[2];
						
						tree->neighbors[c->vertices[4]][1] = c->vertices[0];
						tree->neighbors[c->vertices[4]][2] = c->vertices[6];
						tree->neighbors[c->vertices[4]][4] = c->vertices[5];
						
						tree->neighbors[c->vertices[5]][1] = c->vertices[1];
						tree->neighbors[c->vertices[5]][2] = c->vertices[7];
						tree->neighbors[c->vertices[5]][5] = c->vertices[4];
						
						tree->neighbors[c->vertices[6]][1] = c->vertices[2];
						tree->neighbors[c->vertices[6]][3] = c->vertices[4];
						tree->neighbors[c->vertices[6]][4] = c->vertices[7];

						tree->neighbors[c->vertices[7]][1] = c->vertices[3];
						tree->neighbors[c->vertices[7]][3] = c->vertices[5];
						tree->neighbors[c->vertices[7]][5] = c->vertices[6];
          }
        }
        doDelete = true;
        for (j=0; j<8; j++) if (accepted[c->vertices[j]]) doDelete = false;
        if (doDelete) {
          deleteCell[i] = true;
        }
      }
    } else done = true;
		iter++;
	}
						
	delete [] accepted;

	tree->phi = newPhi;
}

void SlcSurface::setMaxFunc(double (*maxFunc)(const double *x, void *data), void *data) {
	this->maxFunc = maxFunc;
	this->maxFuncData = data;
}

void SlcSurface::setMinFunc(double (*minFunc)(const double *x, void *data), void *data) {
	this->minFunc = minFunc;
	this->minFuncData = data;
}

void SlcSurface::rebuildTree() {
	std::vector<SlVector3>::iterator x;
	std::vector<double>::iterator p;
	DtTree *newTree = new DtTree();
	ConcentricTripleSplitCriterion splitCriterion;
	splitCriterion.tree = tree;
	newTree->buildTree(tree->lc, tree->uc, tree->max_level, splitCriterion);

	newTree->phi.resize(newTree->pts.size());
	for (x=newTree->pts.begin(),p=newTree->phi.begin(); 
			 x!=newTree->pts.end(); x++,p++) {
		(*p) = eval(*x);
	}
	tree = newTree;
}			

