//-------------------------------------------------------------------
//-------------------------------------------------------------------
//
// Staggered Grid
//   -- Header File
//  
// Primary Author: Adam Bargteil (adamb@cs.utah.edu)
//
// This class implements the basic staggered grid used in typical
// Eulerian fluid simulations.
//
//-------------------------------------------------------------------
//-------------------------------------------------------------------
//
// Copyright (c) 2008-2009, Utah State Board of Regents.  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are
// met:
//
//   Redistributions of source code must retain the above copyright
//  notice, this list of conditions and the following disclaimer.
//
//  Redistributions in binary form must reproduce the above copyright
//  notice, this list of conditions and the following disclaimer in the
//  documentation and/or other materials provided with the distribution.
//
//  Redistributions in binary form as an executable program, or part of
//  an executable program must, when run for the first time by a given
//  user, prominently display the above copyright notice, this list of
//  conditions and the following disclaimer.
//
//  Neither the name of the University of California, Berkeley nor the
//  names of its contributors may be used to endorse or promote products
//  derived from this software without specific prior written
//  permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
//-------------------------------------------------------------------
//-------------------------------------------------------------------

#ifndef STAGGEREDGRID_H
#define STAGGEREDGRID_H

#include <vector>
#include "slVector.H"
#include "slcSurface.H"
#include "lsSurface.H"
#include "slMatrix.H"
#include "slArray.H"

class Int3DoublePair {
public:
	Int3DoublePair() : i(0), d(0.0) {};
	Int3DoublePair(SlInt3 i, double d) {this->i = i; this->d = d;};
	Int3DoublePair(const Int3DoublePair &id) {i=id.i; d=id.d;};
	SlInt3 i;
	double d;
};

bool Int3DoublePairCompare(const Int3DoublePair &p1, const Int3DoublePair &p2);

class StaggeredGrid {
public:
	SlArray3D<double> u,v,w,exx,eyy,ezz,exy,exz,eyz;
	SlArray3D<unsigned char> cellLabels, cornerLabels;
	unsigned int nx, ny, nz, nynz;
	double h,halfh;
	SlVector3 lc, uc;
	int band;
	bool allocate(unsigned int nx, unsigned int ny, unsigned int nz, double h, SlVector3 lc, SlVector3 uc, unsigned int band);

	SlVector3 interpVel(const SlVector3 &x) const;
	double interpxface(const SlVector3 &x, const SlArray3D<double> &f) const;
	double interpyface(const SlVector3 &x, const SlArray3D<double> &f) const;
	double interpzface(const SlVector3 &x, const SlArray3D<double> &f) const;
	double interpvertex(const SlVector3 &x, const SlArray3D<double> &f) const;
	double interpcenter(const SlVector3 &x, const SlArray3D<double> &f) const;
	SlVector3 interpcenter(const SlVector3 &x, const SlArray3D<double> &f1, 
						   const SlArray3D<double> &f2, const SlArray3D<double> &f3) const;
	double interpxyedge(const SlVector3 &x, const SlArray3D<double> &f) const;
	double interpxzedge(const SlVector3 &x, const SlArray3D<double> &f) const;
	double interpyzedge(const SlVector3 &x, const SlArray3D<double> &f) const;
	bool advect(double dt);
	bool advectFlip(double dt);
	bool applyLaplacian(const SlArray3D<double> &x, SlArray3D<double> &y) const;
	bool applyLaplacian(const double *x, double *y) const;
	inline bool clipToGrid(SlVector3 &x) const;
	inline bool boundary(unsigned int i, unsigned int j, unsigned int k) const;
	inline bool boundaryx(unsigned int i, unsigned int j, unsigned int k) const;
	inline bool boundaryy(unsigned int i, unsigned int j, unsigned int k) const;
	inline bool boundaryz(unsigned int i, unsigned int j, unsigned int k) const;
	bool computePressure();
	bool setBoundaryVelocity();
	bool setLabels(const SlcSurface &s);
	bool setLabels(const LsSurface &s);
	bool applyPrecon(SlArray3D<double> &r, SlArray3D<double> &z);
	bool applyA(SlArray3D<double> &x, SlArray3D<double> &b);

	bool extrapolate(const SlcSurface &s);
	bool extrapolate(const LsSurface &s);

	bool advectStrain(double dt);
	bool applyElasticForces(double dt, double mu_e);
	bool updateStrain(double dt, double yieldStress, double flowRate);

	bool advectLevelSet(double dt, LsSurface &s);

	// these functions change the behavior of the simulator
	bool elastic(); // turns on/off elastic force computations
	bool rk2(); // turns on/off rk2 advection
	bool bfecc(); // turns on/off bfecc advection
	bool levelset(); // turns on/off using a levelset for surface representation
	bool slc(); // turns on/off using slc for surface representation
	bool flip(unsigned int nFlipParticlesPerCell, double flipRatio); // turns on/off flip

	// FLIP stuff
	SlVector3 *flipPos, *flipVel, *nFlipPos, *bfFlipPos;
	unsigned int nFlipParticles, nFlipParticlesPerCell;
	bool advectFlip();
	bool setLabelsFlip();
	double flipRatio;

	static const unsigned char AIR = 1;
	static const unsigned char LIQUID = 2;
	static const unsigned char SURFACE = 4;
	static const unsigned char OBSTACLE = 8;
	static const unsigned char FLUID = 6;

	friend std::istream &operator>>(std::istream &strm, StaggeredGrid &g);
	friend std::ostream &operator<<(std::ostream &strm, const StaggeredGrid &g);

private:
	bool advectHelper(double dt, const SlArray3D<double> &u, 
										const SlArray3D<double> &v, const SlArray3D<double> &w,
										SlArray3D<double> &nu, SlArray3D<double> &nv,
										SlArray3D<double> &nw);
	bool advectStrainHelper(double dt, const SlArray3D<double> &exx, const SlArray3D<double> &eyy,
													const SlArray3D<double> &ezz, const SlArray3D<double> &exy,
													const SlArray3D<double> &exz, const SlArray3D<double> &eyz,
													const SlArray3D<double> &nexx, const SlArray3D<double> &neyy,
													const SlArray3D<double> &nezz, const SlArray3D<double> &nexy,
													const SlArray3D<double> &nexz, const SlArray3D<double> &neyz);
	bool advectLsHelper(double dt, const SlArray3D<double> &ls, 
											SlArray3D<double> &nls);
	bool advectParticlesHelper(double dt, const SlVector3 *pos, SlVector3 *npos);
	bool flipParticleContribute(SlArray3D<double> &u, SlArray3D<double> &fu,
															double w0, double w1, double w2, 
															int i, int j, int k, double v);
	bool extrapolateHelper(std::vector<Int3DoublePair> &sortedList, SlArray3D<double> &x);
	bool extrapolateHelper(std::vector<Int3DoublePair> &sortedList, SlArray3D<double> &x,
												 SlArray3D<double> &y, SlArray3D<double> &z);
	SlArray3D<unsigned char> laplacian;
	SlArray3D<double> precon, q, r, z, p, s;
	SlArray3D<double> nu,nv,nw,nexx,neyy,nezz,nexy,nexz,neyz,phiMap,nls;
	SlArray3D<double> bfu, bfv, bfw, fu, fv, fw, bfexx, bfeyy, bfezz, bfexy, bfexz, bfeyz, bfls;
	unsigned int flags;
	static const unsigned int ELASTIC = 1;
	static const unsigned int RK2 = 2;
	static const unsigned int BFECC = 4;
	static const unsigned int SLC = 8;
	static const unsigned int LS = 16;
	static const unsigned int FLIP = 32;
};


inline bool StaggeredGrid::clipToGrid(SlVector3 &x) const {
	static const double FEPS = 1e-6;
	if (x[0] <= lc[0]+halfh) x[0] = lc[0]+halfh+FEPS;
	if (x[0] >= uc[0]-halfh) x[0] = uc[0]-halfh-FEPS;
	if (x[1] <= lc[1]+halfh) x[1] = lc[1]+halfh+FEPS;
	if (x[1] >= uc[1]-halfh) x[1] = uc[1]-halfh-FEPS;
	if (x[2] <= lc[2]+halfh) x[2] = lc[2]+halfh+FEPS;
	if (x[2] >= uc[2]-halfh) x[2] = uc[2]-halfh-FEPS;
	return true;
};

inline bool StaggeredGrid::boundary(unsigned int i, unsigned int j, unsigned int k) const {
	if (i == 0 || i == nx-1 || j == 0 || j == ny-1 || k == 0 || k == nz-1)
		return true;
	return false;
};

inline bool StaggeredGrid::boundaryx(unsigned int , unsigned int , unsigned int ) const {
	return false;
};
inline bool StaggeredGrid::boundaryy(unsigned int , unsigned int , unsigned int ) const {
	return false;
};
inline bool StaggeredGrid::boundaryz(unsigned int , unsigned int , unsigned int ) const {
	return false;
};

#endif
