//-------------------------------------------------------------------
//-------------------------------------------------------------------
//
// Staggered Grid
//   -- Code File
//  
// Primary Author: Adam Bargteil (adamb@cs.utah.edu)
//
// This class implements the basic staggered grid used in typical
// Eulerian fluid simulations.
//
//-------------------------------------------------------------------
//-------------------------------------------------------------------
//
// Copyright (c) 2008-2009, Utah State Board of Regents.  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are
// met:
//
//   Redistributions of source code must retain the above copyright
//  notice, this list of conditions and the following disclaimer.
//
//  Redistributions in binary form must reproduce the above copyright
//  notice, this list of conditions and the following disclaimer in the
//  documentation and/or other materials provided with the distribution.
//
//  Redistributions in binary form as an executable program, or part of
//  an executable program must, when run for the first time by a given
//  user, prominently display the above copyright notice, this list of
//  conditions and the following disclaimer.
//
//  Neither the name of the University of California, Berkeley nor the
//  names of its contributors may be used to endorse or promote products
//  derived from this software without specific prior written
//  permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
//-------------------------------------------------------------------
//-------------------------------------------------------------------

#define EIGEN
#define GREEN

#include "staggeredGrid.H"
//#include <petscksp.h>
#include <fstream>
#include "slMatrix.H"
#include <algorithm>
#ifdef GSL
#include <gsl/gsl_matrix_double.h>
#include <gsl/gsl_linalg.h>
#include <gsl/gsl_vector_double.h>
#include <gsl/gsl_blas.h>
#endif
#ifdef __APPLE__
#include <Accelerate/Accelerate.h>
#elif linux
extern "C" {
#include <atlas/cblas.h>
}
#endif

using std::cout;
using std::cerr;
using std::endl;
//using std::isnan;
//using std::isinf;
using std::sort;

static const double EPS = 1e-12;

bool Int3DoublePairCompare(const Int3DoublePair &p1, const Int3DoublePair &p2) {
	return (p1.d < p2.d);
};

bool StaggeredGrid::allocate(unsigned int nx, unsigned int ny, unsigned int nz, double h, SlVector3 lc, SlVector3 uc, unsigned int band){
	this->nx = nx;
	this->ny = ny;
	this->nz = nz;
	this->nynz = ny*nz;
	this->h = h;
	this->halfh = h/2.0;
	this->lc = lc;
	this->uc = uc;
	this->band = band;
	u.allocate(nx+1, ny, nz);
	v.allocate(nx, ny+1, nz);
	w.allocate(nx, ny, nz+1);
	p.allocate(nx, ny, nz);
	nu.allocate(nx+1, ny, nz);
	nv.allocate(nx, ny+1, nz);
	nw.allocate(nx, ny, nz+1);
	cornerLabels.allocate(nx+1, ny+1, nz+1);
	cellLabels.allocate(nx, ny, nz);
	laplacian.allocate(nx, ny, nz);
	precon.allocate(nx, ny, nz);
	q.allocate(nx, ny, nz);
	r.allocate(nx, ny, nz);
	z.allocate(nx, ny, nz);
	s.allocate(nx, ny, nz);
	phiMap.allocate(nx,ny,nz);
	flags = 0;
	
	return true;
}

bool StaggeredGrid::elastic() {
	if (!(flags & ELASTIC)) {
		exx.allocate(nx, ny, nz);
		eyy.allocate(nx, ny, nz);
		ezz.allocate(nx, ny, nz);
		exy.allocate(nx+1, ny+1, nz);
		exz.allocate(nx+1, ny, nz+1);
		eyz.allocate(nx, ny+1, nz+1);
		nexx.allocate(nx, ny, nz);
		neyy.allocate(nx, ny, nz);
		nezz.allocate(nx, ny, nz);
		nexy.allocate(nx+1, ny+1, nz);
		nexz.allocate(nx+1, ny, nz+1);
		neyz.allocate(nx, ny+1, nz+1);
		flags |= ELASTIC;
		if (flags & BFECC) {
			bfexx.allocate(nx, ny, nz);
			bfeyy.allocate(nx, ny, nz);
			bfezz.allocate(nx, ny, nz);
			bfexy.allocate(nx+1, ny+1, nz);
			bfexz.allocate(nx+1, ny, nz+1);
			bfeyz.allocate(nx, ny+1, nz+1);
		}
	} else {
		flags &= ~ELASTIC;
	}
	return true;
}

bool StaggeredGrid::bfecc() {
	if (!(flags & BFECC)) {
		bfu.allocate(nx+1, ny, nz);
		bfv.allocate(nx, ny+1, nz);
		bfw.allocate(nx, ny, nz+1);
		if (flags & ELASTIC) {
			bfexx.allocate(nx, ny, nz);
			bfeyy.allocate(nx, ny, nz);
			bfezz.allocate(nx, ny, nz);
			bfexy.allocate(nx+1, ny+1, nz);
			bfexz.allocate(nx+1, ny, nz+1);
			bfeyz.allocate(nx, ny+1, nz+1);
		}
		if (flags & LS) {
			bfls.allocate(nx+1, ny+1, nz+1);
		}
		flags |= ELASTIC;
	} else {
		flags &= ~ELASTIC;
	}
	return true;
}

bool StaggeredGrid::levelset() {
	if (!(flags & LS)) {
		flags |= LS;
		nls.allocate(nx+1, ny+1, nz+1);
		if (flags & BFECC) {
			bfls.allocate(nx+1, ny+1, nz+1);
		}
	} else {
		flags &= ~LS;
	}
	return true;
}

bool StaggeredGrid::slc() {
	if (!(flags & SLC)) {
		flags |= SLC;
	} else {
		flags &= ~SLC;
	}
	return true;
}

bool StaggeredGrid::flip(unsigned int nFlipParticlesPerCell, double flipRatio) {
	if (!(flags & FLIP)) {
		flags |= FLIP;

		int nFluidCells = 0;
		for (unsigned int i=0; i<nx; i++) {
			for (unsigned int j=0; j<ny; j++) {
				for (unsigned int k=0; k<nz; k++) {
					if (cellLabels(i,j,k) & FLUID) nFluidCells++;
				}
			}
		}
		
		this->flipRatio = flipRatio;
		this->nFlipParticlesPerCell = nFlipParticlesPerCell;
		nFlipParticles = nFluidCells*nFlipParticlesPerCell;
		flipPos = new SlVector3[nFlipParticles];
		flipVel = new SlVector3[nFlipParticles];
		nFlipPos = new SlVector3[nFlipParticles];
		bfFlipPos = new SlVector3[nFlipParticles];
		fu.allocate(nx+1, ny, nz);
		fv.allocate(nx, ny+1, nz);
		fw.allocate(nx, ny, nz+1);

		SlVector3 *p = flipPos;
		SlVector3 *v = flipVel;
		for (unsigned int i=0; i<nx; i++) {
			for (unsigned int j=0; j<ny; j++) {
				for (unsigned int k=0; k<nz; k++) {
					SlVector3 corner = lc+SlVector3(i*h,j*h,k*h);
					for (unsigned int l=0; l<nFlipParticlesPerCell; l++, p++, v++) {
						*p = corner + SlVector3(((double)rand()) / RAND_MAX,
																		((double)rand()) / RAND_MAX,
																		((double)rand()) / RAND_MAX); //not ideal
						*v = interpVel(*p);
					}
				}
			}
		}
		
	} else {
		flags &= ~FLIP;
		delete [] flipPos;
		delete [] flipVel;
		delete [] nFlipPos;
		delete [] bfFlipPos;
	}
	return true;
}

inline double trilinearInterp(const SlArray3D<double> &x, int i, int j, int k,
															double w0, double w1, double w2) {
	return (1.0-w0)*(1.0-w1)*(1.0-w2)*x(i,j,k)+
		(1.0-w0)*(1.0-w1)*(w2)*x(i,j,k+1)+
		(1.0-w0)*(w1)*(1.0-w2)*x(i,j+1,k)+
		(1.0-w0)*(w1)*(w2)*x(i,j+1,k+1)+
		(w0)*(1.0-w1)*(1.0-w2)*x(i+1,j,k)+
		(w0)*(1.0-w1)*(w2)*x(i+1,j,k+1)+
		(w0)*(w1)*(1.0-w2)*x(i+1,j+1,k)+
		(w0)*(w1)*(w2)*x(i+1,j+1,k+1);
}


double StaggeredGrid::interpxface(const SlVector3 &x, const SlArray3D<double> &f) const {
	SlVector3 y(x-lc);
	int i,j,k;
	y[1] -= halfh;
	y[2] -= halfh;
	i = (int)floor(y[0] / h);
	j = (int)floor(y[1] / h);
	k = (int)floor(y[2] / h);
	double w0, w1, w2;
	w0 = (y[0] - i*h)/h;
	w1 = (y[1] - j*h)/h;
	w2 = (y[2] - k*h)/h;

	return trilinearInterp(f, i, j, k, w0, w1, w2);
}

double StaggeredGrid::interpyface(const SlVector3 &x, const SlArray3D<double> &f) const {
	SlVector3 y(x-lc);
	int i,j,k;
	y[0] -= halfh;
	y[2] -= halfh;
	i = (int)floor(y[0] / h);
	j = (int)floor(y[1] / h);
	k = (int)floor(y[2] / h);
	double w0, w1, w2;
	w0 = (y[0] - i*h)/h;
	w1 = (y[1] - j*h)/h;
	w2 = (y[2] - k*h)/h;

	return trilinearInterp(f, i, j, k, w0, w1, w2);
}

double StaggeredGrid::interpzface(const SlVector3 &x, const SlArray3D<double> &f) const {
	SlVector3 y(x-lc);
	int i,j,k;
	y[0] -= halfh;
	y[1] -= halfh;
	i = (int)floor(y[0] / h);
	j = (int)floor(y[1] / h);
	k = (int)floor(y[2] / h);
	double w0, w1, w2;
	w0 = (y[0] - i*h)/h;
	w1 = (y[1] - j*h)/h;
	w2 = (y[2] - k*h)/h;

	return trilinearInterp(f, i, j, k, w0, w1, w2);
}

double StaggeredGrid::interpvertex(const SlVector3 &x, const SlArray3D<double> &f) const {
	SlVector3 y(x-lc);
	int i,j,k;
	i = (int)floor(y[0] / h);
	j = (int)floor(y[1] / h);
	k = (int)floor(y[2] / h);
	double w0, w1, w2;
	w0 = (y[0] - i*h)/h;
	w1 = (y[1] - j*h)/h;
	w2 = (y[2] - k*h)/h;

	return trilinearInterp(f, i, j, k, w0, w1, w2);
}

double StaggeredGrid::interpcenter(const SlVector3 &x, const SlArray3D<double> &f) const {
	SlVector3 y(x-lc);
	int i,j,k;
	y[0] -= halfh;
	y[1] -= halfh;
	y[2] -= halfh;
	i = (int)floor(y[0] / h);
	j = (int)floor(y[1] / h);
	k = (int)floor(y[2] / h);
	double w0, w1, w2;
	w0 = (y[0] - i*h)/h;
	w1 = (y[1] - j*h)/h;
	w2 = (y[2] - k*h)/h;

	return trilinearInterp(f, i, j, k, w0, w1, w2);
}

SlVector3 StaggeredGrid::interpcenter(const SlVector3 &x, const SlArray3D<double> &f1,
																			const SlArray3D<double> &f2, 
																			const SlArray3D<double> &f3) const {
	SlVector3 y(x-lc);
	int i,j,k;
	y[0] -= halfh;
	y[1] -= halfh;
	y[2] -= halfh;
	i = (int)floor(y[0] / h);
	j = (int)floor(y[1] / h);
	k = (int)floor(y[2] / h);
	double w0, w1, w2;
	w0 = (y[0] - i*h)/h;
	w1 = (y[1] - j*h)/h;
	w2 = (y[2] - k*h)/h;

	SlVector3 ret;

	ret[0] = trilinearInterp(f1, i, j, k, w0, w1, w2);
	ret[1] = trilinearInterp(f2, i, j, k, w0, w1, w2);
	ret[2] = trilinearInterp(f3, i, j, k, w0, w1, w2);
	return ret;
}

double StaggeredGrid::interpxyedge(const SlVector3 &x, const SlArray3D<double> &f) const {
	SlVector3 y(x-lc);
	int i,j,k;
	y[2] -= halfh;
	i = (int)floor(y[0] / h);
	j = (int)floor(y[1] / h);
	k = (int)floor(y[2] / h);
	double w0, w1, w2;
	w0 = (y[0] - i*h)/h;
	w1 = (y[1] - j*h)/h;
	w2 = (y[2] - k*h)/h;

	return trilinearInterp(f, i, j, k, w0, w1, w2);
}

double StaggeredGrid::interpxzedge(const SlVector3 &x, const SlArray3D<double> &f) const {
	SlVector3 y(x-lc);
	int i,j,k;
	y[1] -= halfh;
	i = (int)floor(y[0] / h);
	j = (int)floor(y[1] / h);
	k = (int)floor(y[2] / h);
	double w0, w1, w2;
	w0 = (y[0] - i*h)/h;
	w1 = (y[1] - j*h)/h;
	w2 = (y[2] - k*h)/h;

	return trilinearInterp(f, i, j, k, w0, w1, w2);
}

double StaggeredGrid::interpyzedge(const SlVector3 &x, const SlArray3D<double> &f) const {
	SlVector3 y(x-lc);
	int i,j,k;
	y[0] -= halfh;
	i = (int)floor(y[0] / h);
	j = (int)floor(y[1] / h);
	k = (int)floor(y[2] / h);
	double w0, w1, w2;
	w0 = (y[0] - i*h)/h;
	w1 = (y[1] - j*h)/h;
	w2 = (y[2] - k*h)/h;

	return trilinearInterp(eyz, i, j, k, w0, w1, w2);
}

SlVector3 StaggeredGrid::interpVel(const SlVector3 &x) const {
	return SlVector3(interpxface(x,u), interpyface(x,v), interpzface(x,w));
}

bool StaggeredGrid::setBoundaryVelocity() {
	unsigned int i,j,k;

	// set boundary conditions
	for (j=1; j<ny; j++) {
		for (k=1; k<nz; k++) {
			u(0,j,k) = u(1,j,k) = 0.0;
			v(0,j,k) = v(1,j,k);
			w(0,j,k) = w(1,j,k);
			u(nx,j,k) = u(nx-1,j,k) = 0.0;
			v(nx-1,j,k) = v(nx-2,j,k);
			w(nx-1,j,k) = w(nx-2,j,k);
		}
	}
	for (i=1; i<nx; i++) {
		for (k=1; k<nz; k++) {
			u(i,0,k) = u(i,1,k);
			v(i,0,k) = v(i,1,k) = 0.0;
			w(i,0,k) = w(i,1,k);
			u(i,ny-1,k) = u(i,ny-2,k);
			v(i,ny,k) = v(i,ny-1,k) = 0.0;
			w(i,ny-1,k) = w(i,ny-2,k);
		}
	}
	for (i=1; i<nx; i++) {
		for (j=1; j<ny; j++) {
			u(i,j,0) = u(i,j,1);
			v(i,j,0) = v(i,j,1);
			w(i,j,0) = w(i,j,1) = 0.0;
			u(i,j,nz-1) = u(i,j,nz-2);
			v(i,j,nz-1) = v(i,j,nz-2);
			w(i,j,nz) = w(i,j,nz-1) = 0.0;
		}
	}
	return true;
}

bool StaggeredGrid::advectHelper(double dt, const SlArray3D<double> &u, 
																 const SlArray3D<double> &v, const SlArray3D<double> &w,
																 SlArray3D<double> &nu, SlArray3D<double> &nv,
																 SlArray3D<double> &nw) {
	unsigned int i,j,k;
	SlVector3 x, y, v1, v2;
	for (i=1; i<nx; i++) {
		for (j=1; j<ny-1; j++) {
			for (k=1; k<nz-1; k++) {
				if (boundary(i-1,j,k) || boundary(i,j,k) || boundaryx(i,j,k) || 
						(cellLabels(i,j,k) == AIR && cellLabels(i-1,j,k) == AIR)) {
					nu(i,j,k) = u(i,j,k);
					continue;
				}
				v1.set(u(i,j,k), 
							 0.25*(v(i,j,k)+v(i-1,j,k)+v(i,j+1,k)+v(i-1,j+1,k)),
							 0.25*(w(i,j,k)+w(i-1,j,k)+w(i,j,k+1)+w(i-1,j,k+1)));
				x.set(i*h,(j+0.5)*h, (k+0.5)*h);
				x += lc;
				y = x - v1*dt;
				clipToGrid(y);
				if (flags & RK2) {
					v2 = interpVel(y);
					y = x - 0.5*(v1+v2)*dt;
					clipToGrid(y);
				} 
				nu(i,j,k) = interpxface(y, u);
			}
		}
	}
	for (i=1; i<nx-1; i++) {
		for (j=1; j<ny; j++) {
			for (k=1; k<nz-1; k++) {
				if (boundary(i,j-1,k) || boundary(i,j,k) || boundaryy(i,j,k) ||
						(cellLabels(i,j,k) == AIR && cellLabels(i,j-1,k) == AIR)) {
					nv(i,j,k) = v(i,j,k);
					continue;
					}
				v1.set(0.25*(u(i,j,k)+u(i,j-1,k)+u(i+1,j,k)+u(i+1,j-1,k)), 
							 v(i,j,k),
							 0.25*(w(i,j,k)+w(i,j-1,k)+w(i,j,k+1)+w(i,j-1,k+1)));
				x.set((i+0.5)*h,(j)*h, (k+0.5)*h);
				x += lc;
				y = x - v1*dt;
				clipToGrid(y);
				if (flags & RK2) {
					v2 = interpVel(y);
					y = x - 0.5*(v1+v2)*dt;
					clipToGrid(y);
				}
				nv(i,j,k) = interpyface(y,v);
			}
		}
	}
	for (i=1; i<nx-1; i++) {
		for (j=1; j<ny-1; j++) {
			for (k=1; k<nz; k++) {
				if (boundary(i,j,k-1) || boundary(i,j,k) || boundaryz(i,j,k) ||
						(cellLabels(i,j,k) == AIR && cellLabels(i,j,k-1) == AIR)) {
					nw(i,j,k) = w(i,j,k);
					continue;
				}
				v1.set(0.25*(u(i,j,k)+u(i,j,k-1)+u(i+1,j,k)+u(i+1,j,k-1)), 
							 0.25*(v(i,j,k)+v(i,j,k-1)+v(i,j+1,k)+v(i,j+1,k-1)),
							 w(i,j,k));
				x.set((i+0.5)*h,(j+0.5)*h, k*h);
				x += lc;
				y = x - v1*dt;
				clipToGrid(y);
				if (flags & RK2) {
					v2 = interpVel(y);
					y = x - 0.5*(v1+v2)*dt;
					clipToGrid(y);
				}
				nw(i,j,k) = interpzface(y,w);
			}
		}
	}
	return true;
}

bool StaggeredGrid::advect(double dt) {
	unsigned int i,j,k;
	double *a, *b, *c;

	// set boundary conditions
	setBoundaryVelocity();

	if (flags & BFECC) {
		advectHelper(dt, u, v, w, bfu, bfv, bfw);
		advectHelper(-dt, bfu, bfv, bfw, nu, nv, nw);
		
		a = &(bfu(0,0,0)); b = &(u(0,0,0)); c = &(nu(0,0,0));
		for (i=0; i<u.nx()*u.ny()*u.nz(); i++, a++, b++, c++) {
			(*a) = 1.5*(*b) + 0.5*(*c);
		}
		a = &(bfv(0,0,0)); b = &(v(0,0,0)); c = &(nv(0,0,0));
		for (i=0; i<v.nx()*v.ny()*v.nz(); i++, a++, b++, c++) {
			(*a) = 1.5*(*b) + 0.5*(*c);
		}
		a = &(bfw(0,0,0)); b = &(w(0,0,0)); c = &(nw(0,0,0));
		for (i=0; i<w.nx()*w.ny()*w.nz(); i++, a++, b++, c++) {
			(*a) = 1.5*(*b) + 0.5*(*c);
		}

		advectHelper(dt, bfu, bfv, bfw, nu, nv, nw);
	} else {
		advectHelper(dt, u, v, w, nu, nv, nw);
	}

	// copy new values to grid
	for (i=0; i<nx+1; i++) {
		for (j=0; j<ny; j++) {
			for (k=0; k<nz; k++) {
				if (i == 0 || i == 1 || i == nx || i == nx-1) {
					u(i,j,k) = 0.0; 
				} else {
					u(i,j,k) = nu(i,j,k);
				}
			}
		}
	}

	for (i=0; i<nx; i++) {
		for (j=0; j<ny+1; j++) {
			for (k=0; k<nz; k++) {
				if (j == 0 || j == 1 || j == ny || j == ny-1) {
					v(i,j,k) = 0.0; 
				} else {
					v(i,j,k) = nv(i,j,k);
				}
			}
		}
	}

	for (i=0; i<nx; i++) {
		for (j=0; j<ny; j++) {
			for (k=0; k<nz+1; k++) {
				if (k == 0 || k == 1 || k == nz || k == nz-1) {
					w(i,j,k) = 0.0; 
				} else {
					w(i,j,k) = nw(i,j,k);
				}
			}
		}
	}
	return true;
}

bool StaggeredGrid::advectStrainHelper(double dt, const SlArray3D<double> &exx, const SlArray3D<double> &eyy,
																			 const SlArray3D<double> &ezz, const SlArray3D<double> &exy,
																			 const SlArray3D<double> &exz, const SlArray3D<double> &eyz,
																			 const SlArray3D<double> &nexx, const SlArray3D<double> &neyy,
																			 const SlArray3D<double> &nezz, const SlArray3D<double> &nexy,
																			 const SlArray3D<double> &nexz, const SlArray3D<double> &neyz) {
	unsigned int i,j,k;
	SlVector3 x, y, v1, v2, exxyyzz;

	for (i=1; i<nx-1; i++) {
		for (j=1; j<ny-1; j++) {
			for (k=1; k<nz-1; k++) {
				if (boundary(i,j,k) || cellLabels(i,j,k) == AIR) {
					nexx(i,j,k) = exx(i,j,k);
					neyy(i,j,k) = eyy(i,j,k);
					nezz(i,j,k) = ezz(i,j,k);
					continue;
				}
				v1.set(0.5*(u(i,j,k)+u(i+1,j,k)),
							 0.5*(v(i,j,k)+v(i,j+1,k)),
							 0.5*(w(i,j,k)+w(i,j,k+1)));
				x.set((i+0.5)*h,(j+0.5)*h, (k+0.5)*h);
				x += lc;
				y = x - v1*dt;
				clipToGrid(y);
				if (flags & RK2) {
					v2 = interpVel(y);
					y = x - 0.5*(v1+v2)*dt;
					clipToGrid(y);
				}
				exxyyzz = interpcenter(y,exx,eyy,ezz);
				nexx(i,j,k) = exxyyzz[0];
				neyy(i,j,k) = exxyyzz[1];
				nezz(i,j,k) = exxyyzz[2];
			}
		}
	}

	for (i=1; i<nx; i++) {
		for (j=1; j<ny; j++) {
			for (k=1; k<nz-1; k++) {
				if (boundary(i-1,j-1,k) || boundary(i-1,j,k) || boundary(i,j-1,k) || boundary(i,j,k) || 
						boundaryx(i,j-1,k) || boundaryx(i,j,k) || boundaryy(i-1,j,k) || boundaryy(i,j,k) ||
						(cellLabels(i-1,j-1,k) == AIR && cellLabels(i-1,j,k) == AIR && 
						 cellLabels(i,j-1,k) == AIR && cellLabels(i,j,k) == AIR)) {
					nexy(i,j,k) = exy(i,j,k);
					continue;
				}
				v1.set(0.5*(u(i,j-1,k)+u(i,j,k)), 
							 0.5*(v(i-1,j,k)+v(i,j,k)),
							 0.125*(w(i-1,j-1,k)+w(i-1,j,k)+w(i,j-1,k)+w(i,j,k)+
											w(i-1,j-1,k+1)+w(i-1,j,k+1)+w(i,j-1,k+1)+w(i,j,k+1)));
				x.set((i)*h,(j)*h, (k+0.5)*h);
				x += lc;
				y = x - v1*dt;
				clipToGrid(y);
				if (flags & RK2) {
					v2 = interpVel(y);
					y = x - 0.5*(v1+v2)*dt;
					clipToGrid(y);
				}
				nexy(i,j,k) = interpxyedge(y,exy);
			}
		}
	}

	for (i=1; i<nx; i++) {
		for (j=1; j<ny-1; j++) {
			for (k=1; k<nz; k++) {
				if (boundary(i-1,j,k-1) || boundary(i-1,j,k) || boundary(i,j,k-1) || boundary(i,j,k) || 
						boundaryx(i,j,k-1) || boundaryx(i,j,k) || boundaryz(i-1,j,k) || boundaryz(i,j,k) ||
						(cellLabels(i-1,j,k-1) == AIR && cellLabels(i-1,j,k) == AIR && 
						 cellLabels(i,j,k-1) == AIR && cellLabels(i,j,k) == AIR)) {
					nexz(i,j,k) = exz(i,j,k);
					continue;
				}
				v1.set(0.5*(u(i,j,k-1)+u(i,j,k)), 
							 0.125*(v(i-1,j,k-1)+v(i-1,j,k)+v(i,j,k-1)+v(i,j,k)+
											v(i-1,j+1,k-1)+v(i-1,j+1,k)+v(i,j+1,k-1)+v(i,j+1,k)),
							 0.5*(w(i-1,j,k)+w(i,j,k)));
				x.set((i)*h,(j+0.5)*h, (k)*h);
				x += lc;
				y = x - v1*dt;
				clipToGrid(y);
				if (flags & RK2) {
					v2 = interpVel(y);
					y = x - 0.5*(v1+v2)*dt;
					clipToGrid(y);
				}
				nexz(i,j,k) = interpxzedge(y,exz);
			}
		}
	}

	for (i=1; i<nx-1; i++) {
		for (j=1; j<ny; j++) {
			for (k=1; k<nz; k++) {
				if (boundary(i,j-1,k-1) || boundary(i,j-1,k) || boundary(i,j,k-1) || boundary(i,j,k) || 
						boundaryy(i,j,k-1) || boundaryy(i,j,k) || boundaryz(i,j-1,k) || boundaryz(i,j,k) ||
						(cellLabels(i,j-1,k-1) == AIR && cellLabels(i,j-1,k) == AIR && 
						 cellLabels(i,j,k-1) == AIR && cellLabels(i,j,k) == AIR)) {
					neyz(i,j,k) = eyz(i,j,k);
					continue;
				}
				v1.set(0.125*(u(i,j-1,k-1)+u(i,j-1,k)+u(i,j,k-1)+u(i,j,k)+
											u(i+1,j-1,k-1)+u(i+1,j-1,k)+u(i+1,j,k-1)+u(i+1,j,k)), 
							 0.5*(v(i,j,k-1)+v(i,j,k)),
							 0.5*(w(i,j-1,k)+w(i,j,k)));
							 
				x.set((i+0.5)*h,(j)*h, (k)*h);
				x += lc;
				y = x - v1*dt;
				clipToGrid(y);
				if (flags & RK2) {
					v2 = interpVel(y);
					y = x - 0.5*(v1+v2)*dt;
					clipToGrid(y);
				}
				neyz(i,j,k) = interpyzedge(y,eyz);
			}
		}
	}
	return true;
}

bool StaggeredGrid::advectStrain(double dt) {
	if (!(flags & ELASTIC)) return false;
	unsigned int i,j,k;
	double *a, *b, *c;

	// set boundary conditions
	setBoundaryVelocity();

	if (flags & BFECC) {
		advectStrainHelper(dt, exx, eyy, ezz, exy, exz, eyz, 
											 bfexx, bfeyy, bfezz, bfexy, bfexz, bfeyz);
		advectStrainHelper(-dt, bfexx, bfeyy, bfezz, bfexy, bfexz, bfeyz, 
											 nexx, neyy, nezz, nexy, nexz, neyz);
		
		a = &(bfexx(0,0,0)); b = &(exx(0,0,0)); c = &(nexx(0,0,0));
		for (i=0; i<exx.nx()*exx.ny()*exx.nz(); i++, a++, b++, c++) {
			(*a) = 1.5*(*b) + 0.5*(*c);
		}
		a = &(bfeyy(0,0,0)); b = &(eyy(0,0,0)); c = &(neyy(0,0,0));
		for (i=0; i<eyy.nx()*eyy.ny()*eyy.nz(); i++, a++, b++, c++) {
			(*a) = 1.5*(*b) + 0.5*(*c);
		}
		a = &(bfezz(0,0,0)); b = &(ezz(0,0,0)); c = &(nezz(0,0,0));
		for (i=0; i<ezz.nx()*ezz.ny()*ezz.nz(); i++, a++, b++, c++) {
			(*a) = 1.5*(*b) + 0.5*(*c);
		}
		a = &(bfexy(0,0,0)); b = &(exy(0,0,0)); c = &(nexy(0,0,0));
		for (i=0; i<exy.nx()*exy.ny()*exy.nz(); i++, a++, b++, c++) {
			(*a) = 1.5*(*b) + 0.5*(*c);
		}
		a = &(bfexz(0,0,0)); b = &(exz(0,0,0)); c = &(nexz(0,0,0));
		for (i=0; i<exz.nx()*exz.ny()*exz.nz(); i++, a++, b++, c++) {
			(*a) = 1.5*(*b) + 0.5*(*c);
		}
		a = &(bfeyz(0,0,0)); b = &(eyz(0,0,0)); c = &(neyz(0,0,0));
		for (i=0; i<eyz.nx()*eyz.ny()*eyz.nz(); i++, a++, b++, c++) {
			(*a) = 1.5*(*b) + 0.5*(*c);
		}
		
		advectStrainHelper(dt, bfexx, bfeyy, bfezz, bfexy, bfexz, bfeyz, 
											 nexx, neyy, nezz, nexy, nexz, neyz);
	} else {
		advectHelper(dt, u, v, w, nu, nv, nw);
	}



	// copy new values to grid
	for (i=0; i<nx; i++) {
		for (j=0; j<ny; j++) {
			for (k=0; k<nz; k++) {
				if (i == 0 || i == nx-1 || j == 0 || j == ny-1 || k == 0 || k == nz-1) {
					exx(i,j,k) = 0.0; 
					eyy(i,j,k) = 0.0; 
					ezz(i,j,k) = 0.0; 
				} else {
					exx(i,j,k) = nexx(i,j,k);
					eyy(i,j,k) = neyy(i,j,k);
					ezz(i,j,k) = nezz(i,j,k);
				}
			}
		}
	}

	for (i=0; i<nx+1; i++) {
		for (j=0; j<ny+1; j++) {
			for (k=0; k<nz; k++) {
				if (i == 0 || i == nx || j == 0 || j == ny) {
					exy(i,j,k) = 0.0; 
				} else if (i == 1) {
					exy(i,j,k) = nexy(i+1,j,k);
				}	else if (i == nx-1) {
					exy(i,j,k) = nexy(i-1,j,k);
				} else if (j == 1) {
					exy(i,j,k) = nexy(i,j+1,k);
				} else if (j == ny-1) {
					exy(i,j,k) = nexy(i,j-1,k);
				} else {
					exy(i,j,k) = nexy(i,j,k);
				}
			}
		}
	}

	for (i=0; i<nx+1; i++) {
		for (j=0; j<ny; j++) {
			for (k=0; k<nz+1; k++) {
				if (i == 0 || i == nx || k == 0 || k == nz) {
					exz(i,j,k) = 0.0; 
				} else if (i == 1) {
					exz(i,j,k) = nexz(i+1,j,k);
				}	else if (i == nx-1) {
					exz(i,j,k) = nexz(i-1,j,k);
				} else if (k == 1) {
					exz(i,j,k) = nexz(i,j,k+1);
				} else if (k == nz-1) {
					exz(i,j,k) = nexz(i,j,k-1);
				} else {
					exz(i,j,k) = nexz(i,j,k);
				}
			}
		}
	}

	for (i=0; i<nx; i++) {
		for (j=0; j<ny+1; j++) {
			for (k=0; k<nz+1; k++) {
				if (j == 0 || j == ny || k == 0 || k == nz) {
					eyz(i,j,k) = 0.0; 
				} else if (j == 1) {
					eyz(i,j,k) = neyz(i,j+1,k);
				}	else if (j == ny-1) {
					eyz(i,j,k) = neyz(i,j-1,k);
				} else if (k == 1) {
					eyz(i,j,k) = neyz(i,j,k+1);
				} else if (k == nz-1) {
					eyz(i,j,k) = neyz(i,j,k-1);
				} else {
					eyz(i,j,k) = neyz(i,j,k);
				}
			}
		}
	}
	
	return true;
}

bool StaggeredGrid::advectLsHelper(double dt, const SlArray3D<double> &ls, 
																	 SlArray3D<double> &nls) {
	unsigned int i,j,k;
	SlVector3 x, y, v1, v2;
	double band = this->band * this->h;
	for (i=1; i<nx; i++) {
		for (j=1; j<ny; j++) {
			for (k=1; k<nz; k++) {
				if (fabs(ls(i,j,k)) > band) {
					nls(i,j,k) = ls(i,j,k);
					continue;
				}
				v1.set(0.25*(u(i,j,k)+u(i,j,k-1)+u(i,j-1,k)+u(i,j-1,k-1)), 
							 0.25*(v(i,j,k)+v(i,j,k-1)+v(i-1,j,k)+v(i-1,j,k-1)),
							 0.25*(w(i,j,k)+w(i-1,j,k)+w(i,j,k+1)+w(i-1,j,k+1)));
				x.set(i*h, j*h, k*h);
				x += lc;
				y = x - v1*dt;
				clipToGrid(y);
				if (flags & RK2) {
					v2 = interpVel(y);
					y = x - 0.5*(v1+v2)*dt;
					clipToGrid(y);
				} 
				nls(i,j,k) = interpvertex(y, ls);
			}
		}
	}
	return true;
}

bool StaggeredGrid::advectLevelSet(double dt, LsSurface &ls) {
	unsigned int i,j,k;
	double *a, *b, *c;

	// set boundary conditions
	setBoundaryVelocity();

	if (flags & BFECC) {
		advectLsHelper(dt, ls.phi, bfls);
		advectLsHelper(-dt, bfls, nls);
		
		a = &(bfls(0,0,0)); b = &(ls(0,0,0)); c = &(nls(0,0,0));
		for (i=0; i<ls.phi.nx()*ls.phi.ny()*ls.phi.nz(); i++, a++, b++, c++) {
			(*a) = 1.5*(*b) + 0.5*(*c);
		}
		advectLsHelper(dt, bfls, nls);
	} else {
		advectLsHelper(dt, ls.phi, nls);
	}

	// copy new values to grid
	for (i=0; i<nx+1; i++) {
		for (j=0; j<ny+1; j++) {
			for (k=0; k<nz+1; k++) {
				ls(i,j,k) = nls(i,j,k);
			}
		}
	}
	return true;
}

inline double sqr(double x){return x*x;};

bool StaggeredGrid::computePressure() {
	double sigma_new, sigma, alpha, beta;
	static const double tau = 0.97, safety = 0.25, tol=1e-11;
	int iter=0, max_iter=1000;
	int n=nx*ny*nz;
	
	laplacian = 0;
	precon = 0.0;
	q = 0.0;

	for (unsigned int i=1; i<nx-1; i++) {
		for (unsigned int j=1; j<ny-1; j++) {
			for (unsigned int k=1; k<nz-1; k++) {
				if (cellLabels(i,j,k) & OBSTACLE) continue;
				if (cellLabels(i,j,k) & AIR) {
					if (cellLabels(i+1,j,k) & FLUID) {
						laplacian(i+1,j,k) += 1;
					}
					if (cellLabels(i,j+1,k) & FLUID) {
						laplacian(i,j+1,k) += 1;
					}
					if (cellLabels(i,j,k+1) & FLUID) {
						laplacian(i,j,k+1) += 1;
					}
				} else {
					if (cellLabels(i+1,j,k) & (~OBSTACLE)) {
						laplacian(i,j,k) += 1;
						if (cellLabels(i+1,j,k) & FLUID) {
							laplacian(i+1,j,k) += 1;
							laplacian(i,j,k) |= 8;
						}
					}
					if (cellLabels(i,j+1,k) & (~OBSTACLE)) {
						laplacian(i,j,k) += 1;
						if (cellLabels(i,j+1,k) & FLUID) {
							laplacian(i,j+1,k) += 1;
							laplacian(i,j,k) |= 16;
						}
					}
					if (cellLabels(i,j,k+1) & (~OBSTACLE)) {
						laplacian(i,j,k) += 1;
						if (cellLabels(i,j,k+1) & FLUID) {
							laplacian(i,j,k+1) += 1;
							laplacian(i,j,k) |= 32;
						}
					}
				}
			}
		}
	}

	for (unsigned int i=1; i<nx-1; i++) {
		for (unsigned int j=1; j<ny-1; j++) {
			for (unsigned int k=1; k<nz-1; k++) {
				if (cellLabels(i,j,k) & (AIR | OBSTACLE)) continue;
				double tmp = 0.0;
				switch (laplacian(i-1,j,k) & 56) {
				case 56:
					tmp += 2*sqr(precon(i-1,j,k));
					break;
				case 24:
				case 40:
					tmp += sqr(precon(i-1,j,k));
					break;
				};
				switch (laplacian(i,j-1,k) & 56) {
				case 56:
					tmp += 2*sqr(precon(i,j-1,k));
					break;
				case 24:
				case 48:
					tmp += sqr(precon(i,j-1,k));
					break;
				};
				switch (laplacian(i,j,k-1) & 56) {
				case 56:
					tmp += 2*sqr(precon(i,j,k-1));
					break;
				case 40:
				case 48:
					tmp += sqr(precon(i,j,k-1));
					break;
				};

				double e = (laplacian(i,j,k) & 7) 
					- ((laplacian(i-1,j,k) & 8) ? sqr(precon(i-1,j,k)) : 0.0)
					- ((laplacian(i,j-1,k) & 16) ? sqr(precon(i,j-1,k)) : 0.0)
					- ((laplacian(i,j,k-1) & 32) ? sqr(precon(i,j,k-1)) : 0.0)
					- tau * tmp;
				if (e < safety*(laplacian(i,j,k) & 7)) e = (laplacian(i,j,k) & 7);
				precon(i,j,k) = 1/sqrt(e);
			}
		}
	}

	for (unsigned int i=1; i<nx-1; i++) {
		for (unsigned int j=1; j<ny-1; j++) {
			for (unsigned int k=1; k<nz-1; k++) {
				if (cellLabels(i,j,k) & (AIR | OBSTACLE)) {
					r(i,j,k) = 0.0;
				} else {
					r(i,j,k) = -(u(i+1,j,k) - u(i,j,k) + v(i,j+1,k) - v(i,j,k) + 
											 w(i,j,k+1) - w(i,j,k));
				}
			}
		}
	}

	p = 0.0;
	applyPrecon(r,z);
	cblas_dcopy(n, &(z(0,0,0)), 1, &(s(0,0,0)), 1);
	
	sigma = cblas_ddot(n, &(z(0,0,0)), 1, &(r(0,0,0)), 1);
	
	//if (!(cblas_ddot(n, &(z(0,0,0)), 1, &(s(0,0,0)), 1)<tol))
	if (cblas_dnrm2(n, &(r(0,0,0)), 1) >= tol)
	while (iter++ < max_iter) {
		applyA(s,z);
		alpha = sigma/cblas_ddot(n, &(z(0,0,0)), 1, &(s(0,0,0)), 1);
		cblas_daxpy(n, alpha, &(s(0,0,0)), 1, &(p(0,0,0)), 1);
		cblas_daxpy(n, -alpha, &(z(0,0,0)), 1, &(r(0,0,0)), 1);
		
		if (cblas_dnrm2(n, &(r(0,0,0)), 1) < tol) break;

		applyPrecon(r,z);
		sigma_new = cblas_ddot(n, &(z(0,0,0)), 1, &(r(0,0,0)), 1);
		beta = sigma_new / sigma;

		cblas_daxpy(n, beta, &(s(0,0,0)), 1, &(z(0,0,0)), 1);
		cblas_dcopy(n, &(z(0,0,0)), 1, &(s(0,0,0)), 1);

		sigma = sigma_new;
	}

	for (unsigned int i=1; i<nx-1; i++) {
		for (unsigned int j=1; j<ny-1; j++) {
			for (unsigned int k=1; k<nz-1; k++) {
				if (cellLabels(i,j,k) & (OBSTACLE)) continue;
				if (cellLabels(i-1,j,k) & (~OBSTACLE)) u(i,j,k) -= (p(i,j,k) - p(i-1,j,k));
				if (cellLabels(i,j-1,k) & (~OBSTACLE)) v(i,j,k) -= (p(i,j,k) - p(i,j-1,k));
				if (cellLabels(i,j,k-1) & (~OBSTACLE)) w(i,j,k) -= (p(i,j,k) - p(i,j,k-1));
			}
		}
	}	

	if (flags & FLIP) {
		SlVector3 *pos = flipPos;
		SlVector3 *vel = flipVel;
		for (unsigned int i=0; i<nFlipParticles; i++, pos++, vel++) {
			SlVector3 oVel(interpxface((*pos),fu), interpyface((*pos),fv), interpzface((*pos),fw));
			SlVector3 nVel(interpxface((*pos),u), interpyface((*pos),v), interpzface((*pos),w));
			(*vel) = flipRatio*((*vel)-oVel) + nVel;
		}
	}


	cout<<iter<<": "<<cblas_dnrm2(n, &(r(0,0,0)), 1)<<endl;
	return true;
}

bool StaggeredGrid::applyPrecon(SlArray3D<double> &r, SlArray3D<double> &z) {
	for (unsigned int i=1; i<nx-1; i++) {
		for (unsigned int j=1; j<ny-1; j++) {
			for (unsigned int k=1; k<nz-1; k++) {
				if (cellLabels(i,j,k) &  (AIR | OBSTACLE)) continue;
				double t = r(i,j,k) + 
					((laplacian(i-1,j,k) & 8) ? precon(i-1,j,k)*q(i-1,j,k) : 0) +
					((laplacian(i,j-1,k) & 16) ? precon(i,j-1,k)*q(i,j-1,k) : 0) +
					((laplacian(i,j,k-1) & 32) ? precon(i,j,k-1)*q(i,j,k-1) : 0);
				q(i,j,k) = t*precon(i,j,k);
			}
		}
	}

	unsigned char *l = &(laplacian(nx-2, ny-2, nz-2));
	double *p = &(precon(nx-2, ny-2, nz-2));

	for (unsigned int i=nx-2; i>0; i--) {
		for (unsigned int j=ny-2; j>0; j--) {
			for (unsigned int k=nz-2; k>0; k--, l--, p--) {
				if (cellLabels(i,j,k) & (AIR | OBSTACLE)) continue;
				double t = q(i,j,k) +
					(((*l) & 8) ? (*p)*z(i+1,j,k) : 0) +
					(((*l) & 16) ? (*p)*z(i,j+1,k) : 0) +
					(((*l) & 32) ? (*p)*z(i,j,k+1) : 0);
				z(i,j,k) = t*(*p);
			}
			l -= 2;
			p -= 2;
		}
		l -= 2*nz;
		p -= 2*nz;
	}
	
	return true;
}

bool StaggeredGrid::applyA(SlArray3D<double> &s, SlArray3D<double> &z) {
	unsigned char *l = &(laplacian(1,1,1));
	unsigned char *lminusi = &(laplacian(0,1,1));
	unsigned char *lminusj = &(laplacian(1,0,1));
	unsigned char *lminusk = &(laplacian(1,1,0));
	for (unsigned int i=1; i<nx-1; i++) {
		for (unsigned int j=1; j<ny-1; j++) {
			for (unsigned int k=1; k<nz-1; k++, l++, lminusi++, lminusj++, lminusk++) {
				if (!laplacian(i,j,k)) {z(i,j,k) = 0.0; continue;}
				z(i,j,k) = ((*l) & 7)*s(i,j,k) -
					(((*l) & 8) ? s(i+1,j,k) : 0) -
					(((*l) & 16) ? s(i,j+1,k) : 0) -
					(((*l) & 32) ? s(i,j,k+1) : 0) -
					(((*lminusi) & 8) ? s(i-1,j,k) : 0) -
					(((*lminusj) & 16) ? s(i,j-1,k) : 0) -
					(((*lminusk) & 32) ? s(i,j,k-1) : 0);
			}
			l += 2;
			lminusi += 2;
			lminusj += 2;
			lminusk += 2;
		}
		l += 2*nz;
		lminusi += 2*nz;
		lminusj += 2*nz;
		lminusk += 2*nz;
	}
	return true;
}

bool StaggeredGrid::setLabels(const SlcSurface &s) {
	unsigned int i,j,k;
	double p,x,y,z;
	for (i=1,x=lc[0]+h; i<nx; i++,x+=h) {
		for (j=1,y=lc[1]+h; j<ny; j++,y+=h) {
			for (k=1,z=lc[2]+h; k<nz; k++,z+=h) {
				p = s.eval(SlVector3(x,y,z));
				if (p < 0) {
					cornerLabels(i,j,k) = LIQUID;
				} else {
					cornerLabels(i,j,k) = AIR;
				}
			}
		}
	}
	for (i=0; i<nx; i++) {
		for (j=0; j<ny; j++) {
			for (k=0; k<nz; k++) {
				if (boundary(i,j,k)) {
					cellLabels(i,j,k) = OBSTACLE;
					continue;
				}
				unsigned char label = cornerLabels(i,j,k) & cornerLabels(i,j,k+1) &
					cornerLabels(i,j+1,k) & cornerLabels(i,j+1,k+1) &
					cornerLabels(i+1,j,k) & cornerLabels(i+1,j,k+1) &
					cornerLabels(i+1,j+1,k) & cornerLabels(i+1,j+1,k+1);
				if (label & AIR) {
					cellLabels(i,j,k) = AIR;
				} else if (label & LIQUID) {
					cellLabels(i,j,k) = LIQUID;
				} else {
					cellLabels(i,j,k) = SURFACE;
				}
			}
		}
	}	
	return true;
}

bool StaggeredGrid::setLabels(const LsSurface &s) {
	unsigned int i,j,k;
	for (i=1; i<nx; i++) {
		for (j=1; j<ny; j++) {
			for (k=1; k<nz; k++) {
				if (s(i,j,k) < 0) {
					cornerLabels(i,j,k) = LIQUID;
				} else {
					cornerLabels(i,j,k) = AIR;
				}
			}
		}
	}
	for (i=0; i<nx; i++) {
		for (j=0; j<ny; j++) {
			for (k=0; k<nz; k++) {
				if (boundary(i,j,k)) {
					cellLabels(i,j,k) = OBSTACLE;
					continue;
				}
				unsigned char label = cornerLabels(i,j,k) & cornerLabels(i,j,k+1) &
					cornerLabels(i,j+1,k) & cornerLabels(i,j+1,k+1) &
					cornerLabels(i+1,j,k) & cornerLabels(i+1,j,k+1) &
					cornerLabels(i+1,j+1,k) & cornerLabels(i+1,j+1,k+1);
				if (label & AIR) {
					cellLabels(i,j,k) = AIR;
				} else if (label & LIQUID) {
					cellLabels(i,j,k) = LIQUID;
				} else {
					cellLabels(i,j,k) = SURFACE;
				}
			}
		}
	}	
	return true;
}

bool StaggeredGrid::extrapolateHelper(std::vector<Int3DoublePair> &sortedList, SlArray3D<double> &x) {
	double phi, phi1=0, phi2=0, phi3=0, tphi, x1=0, x2=0, x3=0;
	std::vector<Int3DoublePair>::iterator id;
	bool d1, d2, d3;
	SlInt3 i3;

	sort (sortedList.begin(), sortedList.end(), Int3DoublePairCompare);

	for (id=sortedList.begin(); id != sortedList.end(); id++) {
		phi = id->d;
		d1=d2=d3=false;
		i3.set(id->i[0]-1, id->i[1], id->i[2]);
		tphi = phiMap(i3[0],i3[1],i3[2]);
		if (tphi != -1 && tphi < phi) {
			phi1 = tphi;
			x1 = x(i3[0],i3[1],i3[2]);
			d1 = true;
		}
		i3.set(id->i[0]+1, id->i[1], id->i[2]);
		tphi = phiMap(i3[0],i3[1],i3[2]);
		if (tphi != -1 && tphi < phi) {
			phi1 = tphi;
			x1 = x(i3[0],i3[1],i3[2]);
			d1 = true;
		}

		i3.set(id->i[0], id->i[1]-1, id->i[2]);
		tphi = phiMap(i3[0],i3[1],i3[2]);
		if (tphi != -1 && tphi < phi) {
			phi2 = tphi;
			x2 = x(i3[0],i3[1],i3[2]);
			d2 = true;
		}
		i3.set(id->i[0], id->i[1]+1, id->i[2]);
		tphi = phiMap(i3[0],i3[1],i3[2]);
		if (tphi != -1 && tphi < phi) {
			phi2 = tphi;
			x2 = x(i3[0],i3[1],i3[2]);
			d2 = true;
		}

		i3.set(id->i[0], id->i[1], id->i[2]-1);
		tphi = phiMap(i3[0],i3[1],i3[2]);
		if (tphi != -1 && tphi < phi) {
			phi3 = tphi;
			x3 = x(i3[0],i3[1],i3[2]);
			d3 = true;
		}
		i3.set(id->i[0], id->i[1], id->i[2]+1);
		tphi = phiMap(i3[0],i3[1],i3[2]);
		if (tphi != -1 && tphi < phi) {
			phi3 = tphi;
			x3 = x(i3[0],i3[1],i3[2]);
			d3 = true;
		}
		if (d1 && d2 && d3) {
			x(id->i[0],id->i[1],id->i[2]) = (x1*(phi-phi1)+x2*(phi-phi2)+x3*(phi-phi3))/(3*phi-phi1-phi2-phi3);
		} else if (d1 && d2) {
			x(id->i[0],id->i[1],id->i[2]) = (x1*(phi-phi1)+x2*(phi-phi2))/(2*phi-phi1-phi2);
		} else if (d1 && d3) {
			x(id->i[0],id->i[1],id->i[2]) = (x1*(phi-phi1)+x3*(phi-phi3))/(2*phi-phi1-phi3);
		} else if (d2 && d3) {
			x(id->i[0],id->i[1],id->i[2]) = (x2*(phi-phi2)+x3*(phi-phi3))/(2*phi-phi2-phi3);
		} else if (d1) {
			x(id->i[0],id->i[1],id->i[2]) = x1;
		} else if (d2) {
			x(id->i[0],id->i[1],id->i[2]) = x2;
		} else if (d3) {
			x(id->i[0],id->i[1],id->i[2]) = x3;
		} else {
		}
	}
	return true;
}

bool StaggeredGrid::extrapolateHelper(std::vector<Int3DoublePair> &sortedList, SlArray3D<double> &x,
																			SlArray3D<double> &y, SlArray3D<double> &z) {
	double phi, phi1=0, phi2=0, phi3=0, tphi, x1=0, x2=0, x3=0, y1=0, y2=0, y3=0, z1=0, z2=0, z3=0;
	std::vector<Int3DoublePair>::iterator id;
	bool d1, d2, d3;
	SlInt3 i3;

	sort (sortedList.begin(), sortedList.end(), Int3DoublePairCompare);

	for (id=sortedList.begin(); id != sortedList.end(); id++) {
		phi = id->d;
		d1=d2=d3=false;
		i3.set(id->i[0]-1, id->i[1], id->i[2]);
		tphi = phiMap(i3[0],i3[1],i3[2]);
		if (tphi != -1 && tphi < phi) {
			phi1 = tphi;
			x1 = x(i3[0],i3[1],i3[2]);
			y1 = y(i3[0],i3[1],i3[2]);
			z1 = z(i3[0],i3[1],i3[2]);
			d1 = true;
		}
		i3.set(id->i[0]+1, id->i[1], id->i[2]);
		tphi = phiMap(i3[0],i3[1],i3[2]);
		if (tphi != -1 && tphi < phi) {
			phi1 = tphi;
			x1 = x(i3[0],i3[1],i3[2]);
			y1 = y(i3[0],i3[1],i3[2]);
			z1 = z(i3[0],i3[1],i3[2]);
			d1 = true;
		}

		i3.set(id->i[0], id->i[1]-1, id->i[2]);
		tphi = phiMap(i3[0],i3[1],i3[2]);
		if (tphi != -1 && tphi < phi) {
			phi2 = tphi;
			x2 = x(i3[0],i3[1],i3[2]);
			y2 = y(i3[0],i3[1],i3[2]);
			z2 = z(i3[0],i3[1],i3[2]);
			d2 = true;
		}
		i3.set(id->i[0], id->i[1]+1, id->i[2]);
		tphi = phiMap(i3[0],i3[1],i3[2]);
		if (tphi != -1 && tphi < phi) {
			phi2 = tphi;
			x2 = x(i3[0],i3[1],i3[2]);
			y2 = y(i3[0],i3[1],i3[2]);
			z2 = z(i3[0],i3[1],i3[2]);
			d2 = true;
		}

		i3.set(id->i[0], id->i[1], id->i[2]-1);
		tphi = phiMap(i3[0],i3[1],i3[2]);
		if (tphi != -1 && tphi < phi) {
			phi3 = tphi;
			x3 = x(i3[0],i3[1],i3[2]);
			y3 = y(i3[0],i3[1],i3[2]);
			z3 = z(i3[0],i3[1],i3[2]);
			d3 = true;
		}
		i3.set(id->i[0], id->i[1], id->i[2]+1);
		tphi = phiMap(i3[0],i3[1],i3[2]);
		if (tphi != -1 && tphi < phi) {
			phi3 = tphi;
			x3 = x(i3[0],i3[1],i3[2]);
			y3 = y(i3[0],i3[1],i3[2]);
			z3 = z(i3[0],i3[1],i3[2]);
			d3 = true;
		}
		if (d1 && d2 && d3) {
			x(id->i[0],id->i[1],id->i[2]) = (x1*(phi-phi1)+x2*(phi-phi2)+x3*(phi-phi3))/(3*phi-phi1-phi2-phi3);
			y(id->i[0],id->i[1],id->i[2]) = (y1*(phi-phi1)+y2*(phi-phi2)+y3*(phi-phi3))/(3*phi-phi1-phi2-phi3);
			z(id->i[0],id->i[1],id->i[2]) = (z1*(phi-phi1)+z2*(phi-phi2)+z3*(phi-phi3))/(3*phi-phi1-phi2-phi3);
		} else if (d1 && d2) {
			x(id->i[0],id->i[1],id->i[2]) = (x1*(phi-phi1)+x2*(phi-phi2))/(2*phi-phi1-phi2);
			y(id->i[0],id->i[1],id->i[2]) = (y1*(phi-phi1)+y2*(phi-phi2))/(2*phi-phi1-phi2);
			z(id->i[0],id->i[1],id->i[2]) = (z1*(phi-phi1)+z2*(phi-phi2))/(2*phi-phi1-phi2);
		} else if (d1 && d3) {
			x(id->i[0],id->i[1],id->i[2]) = (x1*(phi-phi1)+x3*(phi-phi3))/(2*phi-phi1-phi3);
			y(id->i[0],id->i[1],id->i[2]) = (y1*(phi-phi1)+y3*(phi-phi3))/(2*phi-phi1-phi3);
			z(id->i[0],id->i[1],id->i[2]) = (z1*(phi-phi1)+z3*(phi-phi3))/(2*phi-phi1-phi3);
		} else if (d2 && d3) {
			x(id->i[0],id->i[1],id->i[2]) = (x2*(phi-phi2)+x3*(phi-phi3))/(2*phi-phi2-phi3);
			y(id->i[0],id->i[1],id->i[2]) = (y2*(phi-phi2)+y3*(phi-phi3))/(2*phi-phi2-phi3);
			z(id->i[0],id->i[1],id->i[2]) = (z2*(phi-phi2)+z3*(phi-phi3))/(2*phi-phi2-phi3);
		} else if (d1) {
			x(id->i[0],id->i[1],id->i[2]) = x1;
			y(id->i[0],id->i[1],id->i[2]) = y1;
			z(id->i[0],id->i[1],id->i[2]) = z1;
		} else if (d2) {
			x(id->i[0],id->i[1],id->i[2]) = x2;
			y(id->i[0],id->i[1],id->i[2]) = y2;
			z(id->i[0],id->i[1],id->i[2]) = z2;
		} else if (d3) {
			x(id->i[0],id->i[1],id->i[2]) = x3;
			y(id->i[0],id->i[1],id->i[2]) = y3;
			z(id->i[0],id->i[1],id->i[2]) = z3;
		} else {
		}
	}
	return true;
}

bool StaggeredGrid::extrapolate(const SlcSurface &s) {
	unsigned int i,j,k;
	double phi,band = this->band*h, x, y, z;
	std::vector<Int3DoublePair> sortedList;

	sortedList.clear();
	phiMap = -1;
	for (i=1,x=lc[0]+1.0*h; i<nx-1; i++,x+=h) {
		for (j=1,y=lc[1]+1.5*h; j<ny-1; j++,y+=h) {
			for (k=1,z=lc[2]+1.5*h; k<nz-1; k++,z+=h) {
				phi = s.eval(SlVector3(x,y,z));
				if (fabs(phi) < band) {
					phiMap(i,j,k) = phi;
					if (phi >= 0.0 && (cellLabels(i,j,k) & cellLabels(i-1,j,k) & AIR)) {
						sortedList.push_back(Int3DoublePair(SlInt3(i,j,k),phi));
					}
				}
			}
		}
	}
	extrapolateHelper(sortedList, u);

	sortedList.clear();
	phiMap = -1;
	for (i=1,x=lc[0]+1.5*h; i<nx-1; i++,x+=h) {
		for (j=1,y=lc[1]+1.0*h; j<ny-1; j++,y+=h) {
			for (k=1,z=lc[2]+1.5*h; k<nz-1; k++,z+=h) {
				phi = s.eval(SlVector3(x,y,z),false);
				if (fabs(phi) < band) {
					phiMap(i,j,k) = phi;
					if (phi >= 0.0 && (cellLabels(i,j,k) & cellLabels(i,j-1,k) & AIR)) {
						sortedList.push_back(Int3DoublePair(SlInt3(i,j,k),phi));
					}
				}
			}
		}
	}
	extrapolateHelper(sortedList, v);

	sortedList.clear();
	phiMap = -1;
	for (i=1,x=lc[0]+1.5*h; i<nx-1; i++,x+=h) {
		for (j=1,y=lc[1]+1.5*h; j<ny-1; j++,y+=h) {
			for (k=1,z=lc[2]+1.0*h; k<nz-1; k++,z+=h) {
				phi = s.eval(SlVector3(x,y,z));
				if (fabs(phi) < band) {
					phiMap(i,j,k) = phi;
					if (phi >= 0.0 && (cellLabels(i,j,k) & cellLabels(i,j,k-1) & AIR)) {
						sortedList.push_back(Int3DoublePair(SlInt3(i,j,k),phi));
					}
				}
			}
		}
	}
	//extrapolateHelper(sortedList, w);

	if (!(flags & ELASTIC)) return true;


	sortedList.clear();
	phiMap = -1;
	for (i=1,x=lc[0]+1.5*h; i<nx-1; i++,x+=h) {
		for (j=1,y=lc[1]+1.5*h; j<ny-1; j++,y+=h) {
			for (k=1,z=lc[2]+1.5*h; k<nz-1; k++,z+=h) {
				phi = s.eval(SlVector3(x,y,z));
				if (fabs(phi) < band) {
					phiMap(i,j,k) = phi;
					if (phi >= 0.0 && (cellLabels(i,j,k) & AIR)) {
						sortedList.push_back(Int3DoublePair(SlInt3(i,j,k),phi));
					}
				}
			}
		}
	}
	extrapolateHelper(sortedList, exx, eyy, ezz);

	sortedList.clear();
	phiMap = -1;
	for (i=1,x=lc[0]+1.0*h; i<nx-1; i++,x+=h) {
		for (j=1,y=lc[1]+1.0*h; j<ny-1; j++,y+=h) {
			for (k=1,z=lc[2]+1.5*h; k<nz-1; k++,z+=h) {
				phi = s.eval(SlVector3(x,y,z));
				if (fabs(phi) < band) {
					phiMap(i,j,k) = phi;
					if (phi >= 0.0 && (cellLabels(i,j,k) & cellLabels(i,j-1,k) & 
														 cellLabels(i-1,j,k) & cellLabels(i-1,j-1,k) & AIR)) {
						sortedList.push_back(Int3DoublePair(SlInt3(i,j,k),phi));
					}
				}
			}
		}
	}
	extrapolateHelper(sortedList, exy);

	sortedList.clear();
	phiMap = -1;
	for (i=1,x=lc[0]+1.0*h; i<nx-1; i++,x+=h) {
		for (j=1,y=lc[1]+1.5*h; j<ny-1; j++,y+=h) {
			for (k=1,z=lc[2]+1.0*h; k<nz-1; k++,z+=h) {
				phi = s.eval(SlVector3(x,y,z));
				if (fabs(phi) < band) {
					phiMap(i,j,k) = phi;
					if (phi >= 0.0 && (cellLabels(i,j,k) & cellLabels(i,j,k-1) &
														 cellLabels(i-1,j,k) & cellLabels(i-1,j,k-1) & AIR)) {
						sortedList.push_back(Int3DoublePair(SlInt3(i,j,k),phi));
					}
				}
			}
		}
	}
	extrapolateHelper(sortedList, exz);

	sortedList.clear();
	phiMap = -1;
	for (i=1,x=lc[0]+1.5*h; i<nx-1; i++,x+=h) {
		for (j=1,y=lc[1]+1.0*h; j<ny-1; j++,y+=h) {
			for (k=1,z=lc[2]+1.0*h; k<nz-1; k++,z+=h) {
				phi = s.eval(SlVector3(x,y,z));
				if (fabs(phi) < band) {
					phiMap(i,j,k) = phi;
					if (phi >= 0.0 && (cellLabels(i,j,k) & cellLabels(i,j,k-1) &
														 cellLabels(i,j-1,k) & cellLabels(i,j-1,k-1) & AIR)) {
 						sortedList.push_back(Int3DoublePair(SlInt3(i,j,k),phi));
					}
				}
			}
		}
	}
	extrapolateHelper(sortedList, eyz);

	return true;
}


bool StaggeredGrid::extrapolate(const LsSurface &s) {
	unsigned int i,j,k;
	double phi,band = this->band*h;
	std::vector<Int3DoublePair> sortedList;

	sortedList.clear();
	phiMap = -1;
	for (i=1; i<nx-1; i++) {
		for (j=1; j<ny-1; j++) {
			for (k=1; k<nz-1; k++) {
				phi = 0.25*(s(i,j,k) + s(i,j,k+1) + s(i,j+1,k) + s(i,j+1,k+1));
				if (fabs(phi) < band) {
					phiMap(i,j,k) = phi;
					if (phi >= 0.0 && (cellLabels(i,j,k) & cellLabels(i-1,j,k) & AIR)) {
						sortedList.push_back(Int3DoublePair(SlInt3(i,j,k),phi));
					}
				}
			}
		}
	}
	extrapolateHelper(sortedList, u);

	sortedList.clear();
	phiMap = -1;
	for (i=1; i<nx-1; i++) {
		for (j=1; j<ny-1; j++) {
			for (k=1; k<nz-1; k++) {
				phi = 0.25*(s(i,j,k) + s(i,j,k+1) + s(i+1,j,k) + s(i+1,j,k+1));
				if (fabs(phi) < band) {
					phiMap(i,j,k) = phi;
					if (phi >= 0.0 && (cellLabels(i,j,k) & cellLabels(i,j-1,k) & AIR)) {
						sortedList.push_back(Int3DoublePair(SlInt3(i,j,k),phi));
					}
				}
			}
		}
	}
	extrapolateHelper(sortedList, v);

	sortedList.clear();
	phiMap = -1;
	for (i=1; i<nx-1; i++) {
		for (j=1; j<ny-1; j++) {
			for (k=1; k<nz-1; k++) {
				phi = 0.25*(s(i,j,k) + s(i,j+1,k) + s(i+1,j,k) + s(i+1,j+1,k));
				if (fabs(phi) < band) {
					phiMap(i,j,k) = phi;
					if (phi >= 0.0 && (cellLabels(i,j,k) & cellLabels(i,j,k-1) & AIR)) {
						sortedList.push_back(Int3DoublePair(SlInt3(i,j,k),phi));
					}
				}
			}
		}
	}
	extrapolateHelper(sortedList, w);

	if (!(flags & ELASTIC)) return true;


	sortedList.clear();
	phiMap = -1;
	for (i=1; i<nx-1; i++) {
		for (j=1; j<ny-1; j++) {
			for (k=1; k<nz-1; k++) {
				phi = 0.125*(s(i,j,k) + s(i,j,k+1) + s(i,j+1,k) + s(i,j+1,k+1) +
										 s(i+1,j,k) + s(i+1,j,k+1) + s(i+1,j+1,k) + s(i+1,j+1,k+1));
				if (fabs(phi) < band) {
					phiMap(i,j,k) = phi;
					if (phi >= 0.0 && (cellLabels(i,j,k) & AIR)) {
						sortedList.push_back(Int3DoublePair(SlInt3(i,j,k),phi));
					}
				}
			}
		}
	}
	extrapolateHelper(sortedList, exx, eyy, ezz);

	sortedList.clear();
	phiMap = -1;
	for (i=1; i<nx-1; i++) {
		for (j=1; j<ny-1; j++) {
			for (k=1; k<nz-1; k++) {
				phi = 0.5*(s(i,j,k) + s(i,j,k+1));
				if (fabs(phi) < band) {
					phiMap(i,j,k) = phi;
					if (phi >= 0.0 && (cellLabels(i,j,k) & cellLabels(i,j-1,k) & 
														 cellLabels(i-1,j,k) & cellLabels(i-1,j-1,k) & AIR)) {
						sortedList.push_back(Int3DoublePair(SlInt3(i,j,k),phi));
					}
				}
			}
		}
	}
	extrapolateHelper(sortedList, exy);

	sortedList.clear();
	phiMap = -1;
	for (i=1; i<nx-1; i++) {
		for (j=1; j<ny-1; j++) {
			for (k=1; k<nz-1; k++) {
				phi = 0.5*(s(i,j,k) + s(i,j+1,k));
				if (fabs(phi) < band) {
					phiMap(i,j,k) = phi;
					if (phi >= 0.0 && (cellLabels(i,j,k) & cellLabels(i,j,k-1) &
														 cellLabels(i-1,j,k) & cellLabels(i-1,j,k-1) & AIR)) {
						sortedList.push_back(Int3DoublePair(SlInt3(i,j,k),phi));
					}
				}
			}
		}
	}
	extrapolateHelper(sortedList, exz);

	sortedList.clear();
	phiMap = -1;
	for (i=1; i<nx-1; i++) {
		for (j=1; j<ny-1; j++) {
			for (k=1; k<nz-1; k++) {
				phi = 0.5*(s(i,j,k) + s(i+1,j,k));
				if (fabs(phi) < band) {
					phiMap(i,j,k) = phi;
					if (phi >= 0.0 && (cellLabels(i,j,k) & cellLabels(i,j,k-1) &
														 cellLabels(i,j-1,k) & cellLabels(i,j-1,k-1) & AIR)) {
 						sortedList.push_back(Int3DoublePair(SlInt3(i,j,k),phi));
					}
				}
			}
		}
	}
	extrapolateHelper(sortedList, eyz);

	return true;
}


bool StaggeredGrid::applyElasticForces(double dt, double mu_e) {
	if (!(flags & ELASTIC)) return false;
	unsigned int i,j,k;
	for (i=1; i<nx; i++) {
		for (j=1; j<ny-1; j++) {
			for (k=1; k<nz-1; k++) {
				if (boundary(i-1,j,k) || boundary(i,j,k) || boundaryx(i,j,k) || 
						(cellLabels(i,j,k) == AIR && cellLabels(i-1,j,k) == AIR)) {
					//(cellLabels(i,j,k) == SURFACE && cellLabels(i-1,j,k) == SURFACE)) {
					continue;
				}
				u(i,j,k) += mu_e * dt * (exx(i,j,k) - exx(i-1,j,k));

				if (!boundary(i-1,j-1,k) && !boundary(i,j-1,k) && 
						!boundaryx(i,j-1,k) && !boundaryy(i-1,j,k) && !boundaryy(i,j,k)) {
					u(i,j,k) += mu_e * dt * (exy(i,j+1,k) - exy(i,j,k));
				}

				if (!boundary(i-1,j,k-1) && !boundary(i,j,k-1) && 
						!boundaryx(i,j,k-1) && !boundaryz(i-1,j,k) && !boundaryz(i,j,k)) {
					u(i,j,k) += mu_e * dt * (exz(i,j,k+1) - exz(i,j,k));
				}
			}
		}
	}
	for (i=1; i<nx-1; i++) {
		for (j=1; j<ny; j++) {
			for (k=1; k<nz-1; k++) {
				if (boundary(i,j-1,k) || boundary(i,j,k) || boundaryy(i,j,k) ||
						(cellLabels(i,j,k) == AIR && cellLabels(i,j-1,k) == AIR)) {
					//(cellLabels(i,j,k) == SURFACE && cellLabels(i,j-1,k) == SURFACE)) {
					continue;
				}
				v(i,j,k) += mu_e * dt * (eyy(i,j,k) - eyy(i,j-1,k));

				if (!boundary(i-1,j-1,k) && !boundary(i-1,j,k) && 
						!boundaryy(i-1,j,k) && !boundaryx(i,j-1,k) && !boundaryx(i,j,k)) {
					v(i,j,k) += mu_e * dt * (exy(i+1,j,k) - exy(i,j,k));
				}

				if (!boundary(i,j-1,k-1) && !boundary(i,j,k-1) && 
						!boundaryy(i,j,k-1) && !boundaryz(i,j-1,k) && !boundaryz(i,j,k)) {
					v(i,j,k) += mu_e * dt * (eyz(i,j,k+1) - eyz(i,j,k));
				}
			}
		}
	}

	for (i=1; i<nx-1; i++) {
		for (j=1; j<ny-1; j++) {
			for (k=1; k<nz; k++) {
				if (boundary(i,j,k-1) || boundary(i,j,k) || boundaryz(i,j,k) ||
						(cellLabels(i,j,k) == AIR && cellLabels(i,j,k-1) == AIR)) {
					//(cellLabels(i,j,k) == SURFACE && cellLabels(i,j,k-1) == SURFACE)) {
					continue;
				}
				w(i,j,k) += mu_e * dt * (ezz(i,j,k) - ezz(i,j,k-1));

				if (!boundary(i-1,j,k-1) && !boundary(i-1,j,k) && 
						!boundaryz(i-1,j,k) && !boundaryx(i,j,k-1) && !boundaryx(i,j,k-1)) {
					w(i,j,k) += mu_e * dt * (exz(i+1,j,k) - exz(i,j,k));
				}

				if (!boundary(i,j-1,k-1) && !boundary(i,j-1,k) && 
						!boundaryz(i,j-1,k) && !boundaryy(i,j,k-1) && !boundaryy(i,j,k-1)) {
					w(i,j,k) += mu_e * dt * (eyz(i,j+1,k) - eyz(i,j,k));
				}
			}
		}
	}
	return true;
}

inline double computeDecay(SlMatrix3x3 &e, double yieldStress, double flowRate) {
	double t = trace(e)/3.0;
	e(0,0) -= t;
	e(1,1) -= t;
	e(2,2) -= t;
	
	double n = sqrt(e(0,0)*e(0,0) + e(1,1)*e(1,1) + e(2,2)*e(2,2) +
									2*e(0,1)*e(0,1) + 2*e(0,2)*e(0,2) + 2*e(1,2)*e(1,2));
	
	double decay = 0;
	if (n > EPS) {
		decay = flowRate * std::max<double>(0.0, n-yieldStress)/n;
	} else {
		decay = flowRate;
	}
	return decay;
}

bool StaggeredGrid::updateStrain(double dt, double yieldStress, double flowRate) {
	if (!(flags & ELASTIC)) return false;
	unsigned int i,j,k;

	for (i=1; i<nx-1; i++) {
		for (j=1; j<ny-1; j++) {
			for (k=1; k<nz-1; k++) {
				if (boundary(i,j,k) || cellLabels(i,j,k) == AIR) {
					continue;
				}
				if (!boundaryx(i,j,k) && !boundaryx(i+1,j,k)) nexx(i,j,k) = exx(i,j,k) + dt * (u(i+1,j,k) - u(i,j,k)) / h;
				if (!boundaryy(i,j,k) && !boundaryy(i,j+1,k)) neyy(i,j,k) = eyy(i,j,k) + dt * (v(i,j+1,k) - v(i,j,k)) / h;
				if (!boundaryz(i,j,k) && !boundaryz(i,j,k+1)) nezz(i,j,k) = ezz(i,j,k) + dt * (w(i,j,k+1) - w(i,j,k)) / h;
			}
		}
	}

	for (i=1; i<nx; i++) {
		for (j=1; j<ny; j++) {
			for (k=1; k<nz-1; k++) {
				if (boundary(i-1,j-1,k) || boundary(i-1,j,k) || boundary(i,j-1,k) || boundary(i,j,k) || 
						boundaryx(i,j-1,k) || boundaryx(i,j,k) || boundaryy(i-1,j,k) || boundaryy(i,j,k) ||
						(cellLabels(i-1,j-1,k) == AIR && cellLabels(i-1,j,k) == AIR && 
						 cellLabels(i,j-1,k) == AIR && cellLabels(i,j,k) == AIR)) {
					continue;
				}
				nexy(i,j,k) = exy(i,j,k) + 0.5 * dt * (u(i,j,k) - u(i,j-1,k) + v(i,j,k) - v(i-1,j,k)) / h;
			}
		}
	}

	for (i=1; i<nx; i++) {
		for (j=1; j<ny-1; j++) {
			for (k=1; k<nz; k++) {
				if (boundary(i-1,j,k-1) || boundary(i-1,j,k) || boundary(i,j,k-1) || boundary(i,j,k) || 
						boundaryx(i,j,k-1) || boundaryx(i,j,k) || boundaryz(i-1,j,k) || boundaryz(i,j,k) ||
						(cellLabels(i-1,j,k-1) == AIR && cellLabels(i-1,j,k) == AIR && 
						 cellLabels(i,j,k-1) == AIR && cellLabels(i,j,k) == AIR)) {
					continue;
				}
				nexz(i,j,k) = exz(i,j,k) + 0.5 * dt * (u(i,j,k) - u(i,j,k-1) + w(i,j,k) - w(i-1,j,k)) / h;
			}
		}
	}

	for (i=1; i<nx-1; i++) {
		for (j=1; j<ny; j++) {
			for (k=1; k<nz; k++) {
				if (boundary(i,j-1,k-1) || boundary(i,j-1,k) || boundary(i,j,k-1) || boundary(i,j,k) || 
						boundaryy(i,j,k-1) || boundaryy(i,j,k) || boundaryz(i,j-1,k) || boundaryz(i,j,k) ||
						(cellLabels(i,j-1,k-1) == AIR && cellLabels(i,j-1,k) == AIR && 
						 cellLabels(i,j,k-1) == AIR && cellLabels(i,j,k) == AIR)) {
					continue;
				}
				neyz(i,j,k) = eyz(i,j,k) + 0.5 * dt * (v(i,j,k) - v(i,j,k-1) + w(i,j,k) - w(i,j-1,k)) / h;
			}
		}
	}

	SlMatrix3x3 e;

	for (i=1; i<nx-1; i++) {
		for (j=1; j<ny-1; j++) {
			for (k=1; k<nz-1; k++) {
				if (boundary(i,j,k) || cellLabels(i,j,k) == AIR) {
					continue;
				}
				
				e(0,0) = nexx(i,j,k);
				e(1,1) = nexx(i,j,k);
				e(2,2) = nezz(i,j,k);

				e(0,1) = e(1,0) = 0.25*(nexy(i,j,k) + nexy(i,j+1,k) +
																nexy(i+1,j,k) + nexy(i+1,j+1,k));
				e(0,2) = e(2,0) = 0.25*(nexz(i,j,k) + nexz(i,j,k+1) +
																nexz(i+1,j,k) + nexz(i+1,j,k+1));
				e(1,2) = e(2,1) = 0.25*(neyz(i,j,k) + neyz(i,j,k+1) +
																neyz(i,j+1,k) + neyz(i,j+1,k+1));

				double decay = computeDecay(e, yieldStress, flowRate);

				exx(i,j,k) = nexx(i,j,k) - std::min<double>(dt*decay,1.0)*e(0,0);
				eyy(i,j,k) = neyy(i,j,k) - std::min<double>(dt*decay,1.0)*e(1,1);
				ezz(i,j,k) = nezz(i,j,k) - std::min<double>(dt*decay,1.0)*e(2,2);
			}
		}
	}

	for (i=1; i<nx; i++) {
		for (j=1; j<ny; j++) {
			for (k=1; k<nz-1; k++) {
				if (boundary(i-1,j-1,k) || boundary(i-1,j,k) || boundary(i,j-1,k) || boundary(i,j,k) || 
						boundaryx(i,j-1,k) || boundaryx(i,j,k) || boundaryy(i-1,j,k) || boundaryy(i,j,k) ||
						(cellLabels(i-1,j-1,k) == AIR && cellLabels(i-1,j,k) == AIR && 
						 cellLabels(i,j-1,k) == AIR && cellLabels(i,j,k) == AIR)) {
					continue;
				}
				e(0,0) = 0.25*(nexx(i-1,j-1,k) + nexx(i-1,j,k) + nexx(i,j-1,k) + nexx(i,j,k));
				e(1,1) = 0.25*(neyy(i-1,j-1,k) + neyy(i-1,j,k) + neyy(i,j-1,k) + neyy(i,j,k));
				e(2,2) = 0.25*(nezz(i-1,j-1,k) + nezz(i-1,j,k) + nezz(i,j-1,k) + nezz(i,j,k));

				e(0,1) = e(1,0) = nexy(i,j,k);
				e(0,2) = e(2,0) = 0.25*(nexz(i,j-1,k) + nexz(i,j-1,k+1) + nexz(i,j,k) + nexz(i,j,k+1));
				e(1,2) = e(2,1) = 0.25*(neyz(i-1,j,k) + neyz(i-1,j,k+1) + neyz(i,j,k) + neyz(i,j,k+1));

				double decay = computeDecay(e, yieldStress, flowRate);

				exy(i,j,k) = nexy(i,j,k) - std::min<double>(dt*decay, 1.0)*e(0,1);
			}
		}
	}

	for (i=1; i<nx; i++) {
		for (j=1; j<ny-1; j++) {
			for (k=1; k<nz; k++) {
				if (boundary(i-1,j,k-1) || boundary(i-1,j,k) || boundary(i,j,k-1) || boundary(i,j,k) || 
						boundaryx(i,j,k-1) || boundaryx(i,j,k) || boundaryz(i-1,j,k) || boundaryz(i,j,k) ||
						(cellLabels(i-1,j,k-1) == AIR && cellLabels(i-1,j,k) == AIR && 
						 cellLabels(i,j,k-1) == AIR && cellLabels(i,j,k) == AIR)) {
					continue;
				}
				e(0,0) = 0.25*(nexx(i-1,j,k-1) + nexx(i-1,j,k) + nexx(i,j,k-1) + nexx(i,j,k));
				e(1,1) = 0.25*(neyy(i-1,j,k-1) + neyy(i-1,j,k) + neyy(i,j,k-1) + neyy(i,j,k));
				e(2,2) = 0.25*(nezz(i-1,j,k-1) + nezz(i-1,j,k) + nezz(i,j,k-1) + nezz(i,j,k));

				e(0,1) = e(1,0) = 0.25*(nexy(i,j,k-1) + nexy(i,j+1,k-1) + nexy(i,j,k) + nexy(i,j+1,k));
				e(0,2) = e(2,0) = nexz(i,j,k);
				e(1,2) = e(2,1) = 0.25*(neyz(i-1,j,k) + neyz(i-1,j+1,k) + neyz(i,j,k) + neyz(i,j+1,k));

				double decay = computeDecay(e, yieldStress, flowRate);

				exz(i,j,k) = nexz(i,j,k) - std::min<double>(dt*decay, 1.0)*e(0,2);
			}
		}
	}

	for (i=1; i<nx-1; i++) {
		for (j=1; j<ny; j++) {
			for (k=1; k<nz; k++) {
				if (boundary(i,j-1,k-1) || boundary(i,j-1,k) || boundary(i,j,k-1) || boundary(i,j,k) || 
						boundaryy(i,j,k-1) || boundaryy(i,j,k) || boundaryz(i,j-1,k) || boundaryz(i,j,k) ||
						(cellLabels(i,j-1,k-1) == AIR && cellLabels(i,j-1,k) == AIR && 
						 cellLabels(i,j,k-1) == AIR && cellLabels(i,j,k) == AIR)) {
					continue;
				}
				e(0,0) = 0.25*(nexx(i-1,j-1,k) + nexx(i-1,j,k) + nexx(i,j-1,k) + nexx(i,j,k));
				e(1,1) = 0.25*(neyy(i-1,j-1,k) + neyy(i-1,j,k) + neyy(i,j-1,k) + neyy(i,j,k));
				e(2,2) = 0.25*(nezz(i-1,j-1,k) + nezz(i-1,j,k) + nezz(i,j-1,k) + nezz(i,j,k));

				e(0,1) = e(1,0) = 0.25*(nexy(i,j,k-1) + nexy(i+1,j,k-1) + nexy(i,j,k) + nexy(i+1,j,k));
				e(0,2) = e(2,0) = 0.25*(nexz(i,j-1,k) + nexz(i+1,j-1,k) + nexz(i,j,k) + nexz(i+1,j,k));
				e(1,2) = e(2,1) = neyz(i,j,k);

				double decay = computeDecay(e, yieldStress, flowRate);

				eyz(i,j,k) = neyz(i,j,k) - std::min<double>(dt*decay, 1.0)*e(1,2);
			}
		}
	}
	return true;
}

bool StaggeredGrid::advectParticlesHelper(double dt, const SlVector3 *pos, SlVector3 *npos) {
	const SlVector3 *o = pos;
	SlVector3 *n = npos;
	if (flags & RK2) {
		for (unsigned int i=0; i<nFlipParticles; i++, o++, n++) {
			SlVector3 v1 = interpVel(*o);
			SlVector3 y = (*o) + dt*v1;
			clipToGrid(y);
			SlVector3 v2 = interpVel(y);
			(*n) = (*o) + 0.5*(v1+v2)*dt;
			clipToGrid(*n);
		}
	} else {
		for (unsigned int i=0; i<nFlipParticles; i++, o++, n++) {
			(*n) = (*o) + dt*interpVel((*o));
			clipToGrid(*n);
		}
	}
	return true;
}

bool StaggeredGrid::flipParticleContribute(SlArray3D<double> &u, SlArray3D<double> &fu,
																					 double w0, double w1, double w2, 
																					 int i, int j, int k, double v) {
	double w = (1.0-w0)*(1.0-w1)*(1.0-w2);
	u(i,j,k) += w*v;
	fu(i,j,k) += w;
	
	w = (1.0-w0)*(1.0-w1)*(w2);
	u(i,j,k+1) += w*v;
	fu(i,j,k+1) += w;
	
	w = (1.0-w0)*w1*(1.0-w2);
	u(i,j+1,k) += w*v;
	fu(i,j+1,k) += w;
	
	w = (1.0-w0)*w1*w2;
	u(i,j+1,k+1) += w*v;
	fu(i,j+1,k+1) += w;
	
	w = w0*(1.0-w1)*(1.0-w2);
	u(i+1,j,k) += w*v;
	fu(i+1,j,k) += w;
	
	w = w0*(1.0-w1)*w2;
	u(i+1,j,k+1) += w*v;
	fu(i+1,j,k+1) += w;
	
	w = w0*w1*(1.0-w2);
	u(i+1,j+1,k) += w*v;
	fu(i+1,j+1,k) += w;
	
	w = w0*w1*w2;
	u(i+1,j+1,k+1) += w*v;
	fu(i+1,j+1,k+1) += w;

	return true;
}

bool StaggeredGrid::advectFlip(double dt) {
	if (flags & BFECC) {
		advectParticlesHelper(dt, flipPos, bfFlipPos);
		advectParticlesHelper(-dt, bfFlipPos, nFlipPos);
		
		SlVector3 *a = bfFlipPos, *b = flipPos, *c = nFlipPos;
		for (unsigned int i=0; i<nFlipParticles; i++, a++, b++, c++) {
			(*a) = 1.5*(*b) + 0.5*(*c);
		}
		advectParticlesHelper(dt, bfFlipPos, nFlipPos);
	} else {
		advectParticlesHelper(dt, flipPos, nFlipPos);
	}

	unsigned int i,j,k;
	for (i=0; i<nx+1; i++) {
		for (j=0; j<ny; j++) {
			for (k=0; k<nz; k++) {
				u(i,j,k) = 0.0; 
				fu(i,j,k) = 0.0;
			}
		}
	}
	for (i=0; i<nx; i++) {
		for (j=0; j<ny+1; j++) {
			for (k=0; k<nz; k++) {
				v(i,j,k) = 0.0; 
				fv(i,j,k) = 0.0;
			}
		}
	}
	for (i=0; i<nx; i++) {
		for (j=0; j<ny; j++) {
			for (k=0; k<nz+1; k++) {
				w(i,j,k) = 0.0; 
				fw(i,j,k) = 0.0;
			}
		}
	}

	for (i=0; i<nx; i++) {
		for (j=0; j<ny; j++) {
			for (k=0; k<nz; k++) {
				if (boundary(i,j,k)) {
					cellLabels(i,j,k) = OBSTACLE;
					continue;
				}
				cellLabels(i,j,k) = AIR;
			}
		}
	}

	SlVector3 *pos = flipPos;
	SlVector3 *vel = flipVel;
	for (unsigned int i=0; i<nFlipParticles; i++, pos++, vel++) {
		SlVector3 y((*pos)-lc);
		int i,j,k;
		double w0, w1, w2;

		i = (int)floor(y[0] / h);
		j = (int)floor(y[1] / h);
		k = (int)floor(y[2] / h);
		cellLabels(i,j,k) = LIQUID;

		y[1] -= halfh;
		y[2] -= halfh;
		i = (int)floor(y[0] / h);
		j = (int)floor(y[1] / h);
		k = (int)floor(y[2] / h);
		w0 = (y[0] - i*h)/h;
		w1 = (y[1] - j*h)/h;
		w2 = (y[2] - k*h)/h;
		flipParticleContribute(u, fu, w0, w1, w2, i, j, k, (*vel)[0]);
		
		y[0] -= halfh;
		y[1] += halfh;
		i = (int)floor(y[0] / h);
		j = (int)floor(y[1] / h);
		k = (int)floor(y[2] / h);
		w0 = (y[0] - i*h)/h;
		w1 = (y[1] - j*h)/h;
		w2 = (y[2] - k*h)/h;
		flipParticleContribute(v, fv, w0, w1, w2, i, j, k, (*vel)[1]);
		
		y[1] -= halfh;
		y[2] += halfh;
		i = (int)floor(y[0] / h);
		j = (int)floor(y[1] / h);
		k = (int)floor(y[2] / h);
		w0 = (y[0] - i*h)/h;
		w1 = (y[1] - j*h)/h;
		w2 = (y[2] - k*h)/h;
		flipParticleContribute(w, fw, w0, w1, w2, i, j, k, (*vel)[2]);
	}
	
	for (i=0; i<nx+1; i++) {
		for (j=0; j<ny; j++) {
			for (k=0; k<nz; k++) {
				if (i == 0 || i == 1 || i == nx || i == nx-1) {
					u(i,j,k) = 0.0; 
				} else {
					if (fu(i,j,k) > EPS) u(i,j,k) /= fu(i,j,k);
				}
				fu(i,j,k) = u(i,j,k);
			}
		}
	}
	for (i=0; i<nx; i++) {
		for (j=0; j<ny+1; j++) {
			for (k=0; k<nz; k++) {
				if (j == 0 || j == 1 || j == ny || j == ny-1) {
					v(i,j,k) = 0.0; 
				} else {
					if (fv(i,j,k) > EPS) v(i,j,k) /= fv(i,j,k);
				}
				fv(i,j,k) = v(i,j,k);
			}
		}
	}
	for (i=0; i<nx; i++) {
		for (j=0; j<ny; j++) {
			for (k=0; k<nz+1; k++) {
				if (k == 0 || k == 1 || k == nz || k == nz-1) {
					w(i,j,k) = 0.0; 
				} else {
					if (fw(i,j,k) > EPS) w(i,j,k) /= fw(i,j,k);
				}
				fw(i,j,k) = w(i,j,k);
			}
		}
	}	
	for (i=0; i<nx; i++) {
		for (j=0; j<ny; j++) {
			for (k=0; k<nz; k++) {
				if (cellLabels(i,j,k) == FLUID) {
					if (cellLabels(i-1,j,k) == AIR ||
							cellLabels(i+1,j,k) == AIR ||
							cellLabels(i,j-1,k) == AIR ||
							cellLabels(i,j+1,k) == AIR ||
							cellLabels(i,j,k-1) == AIR ||
							cellLabels(i,j,k+1) == AIR) {
						cellLabels(i,j,k) = SURFACE;
					}
				}
			}
		}
	}
	return true;
}

std::istream &operator>>(std::istream &strm, StaggeredGrid &g) {
	std::ios::fmtflags orgFlags = strm.setf(std::ios::skipws);

	eatStr("[", strm);
	strm >> g.nx;
	eatStr(",", strm);
	strm >> g.ny;
	eatStr(",", strm);
	strm >> g.nz;
	eatStr("]", strm);
	g.nynz = g.ny*g.nz;
	eatStr(",", strm);
	strm >> g.h;
	g.halfh = g.h / 2.0;
	eatStr(",", strm);
	strm >> g.lc;
	eatStr("x", strm);
	strm >> g.uc;
	eatStr(",", strm);
	strm >> g.band;

	eatStr(",", strm);
	strm >> g.u;

	eatStr(",", strm);
	strm >> g.v;

	eatStr(",", strm);
	strm >> g.w;

	eatStr(",", strm);
	strm >> g.exx;

	eatStr(",", strm);
	strm >> g.eyy;

	eatStr(",", strm);
	strm >> g.ezz;

	eatStr(",", strm);
	strm >> g.exy;

	eatStr(",", strm);
	strm >> g.exz;

	eatStr(",", strm);
	strm >> g.eyz;

	g.p.allocate(g.nx, g.ny, g.nz);
	g.nu.allocate(g.nx+1, g.ny, g.nz);
	g.nv.allocate(g.nx, g.ny+1, g.nz);
	g.nw.allocate(g.nx, g.ny, g.nz+1);
	g.nexx.allocate(g.nx, g.ny, g.nz);
	g.neyy.allocate(g.nx, g.ny, g.nz);
	g.nezz.allocate(g.nx, g.ny, g.nz);
	g.nexy.allocate(g.nx+1, g.ny+1, g.nz);
	g.nexz.allocate(g.nx+1, g.ny, g.nz+1);
	g.neyz.allocate(g.nx, g.ny+1, g.nz+1);
	g.cornerLabels.allocate(g.nx+1, g.ny+1, g.nz+1);
	g.cellLabels.allocate(g.nx, g.ny, g.nz);
	g.laplacian.allocate(g.nx, g.ny, g.nz);
	g.phiMap.allocate(g.nx, g.ny, g.nz);

	strm.flags(orgFlags);
	return strm;
}

std::ostream &operator<<(std::ostream &strm, const StaggeredGrid &g) {
	strm << "[" << g.nx << "," << g.ny << "," << g.nz << "],";
	strm << g.h << "," << g.lc << "x" << g.uc << "," << g.band << ",\n";
	strm << g.u << "," << g.v << "," << g.w << ",";
	strm << g.exx << "," << g.eyy << "," << g.ezz << ",";
	strm << g.exy << "," << g.exz << "," << g.eyz << ",";
	strm << "\n";

	return strm;
}

